/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
package base;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.util.ArrayList;
import java.util.List;

import main.FunSiPJar;

import org.apache.log4j.Logger;

import base.cluster.ClusterClassificationAction;


import util.classificationFeatures.ClassificationFeature;
import util.io.IOUtils;

/**
 * This class represents the different actions etc. defined in a single
 * configuration file. After being parsed, all these actions are launched
 * by the workflow parameters also defined in the configuration file.
 * @author Michiel Van Bel
 */
public class FunSiPWorkflow implements Runnable{

	/**
	 * The different types of available multithreading:
	 * a) NO_MULTITHREADING : no multithreading at all
	 * b) SINGLE_PC_MULTITHREADING : launches each different ClassificationAction in a seperate thread,
	 * 									and waits until they're all finished;
	 * c) CLUSTER_MULTITHREADING : attempts to launch each ClassificationAction in a seperate job for the
	 * 								cluster (standard qsub-command). The utilizes a busy loop to poll
	 * 								once each minute to check whether all the posted jobs are finished.
	 * 
	 * @author Michiel Van Bel
	 */
	public static enum MULTITHREADING{NO_MULTITHREADING,SINGLE_PC_MULTITHREADING,CLUSTER_MULTITHREADING};
	
	//logging facility, name of the parameterfile, list with classificationactions to be executed, threading indicator
	private Logger logger 						= null;
	private String fileName 					= null;
	private List<ClassificationAction> caList 	= new ArrayList<ClassificationAction>();
	private List<List<ClassificationAction>> ca2List = null; 
	private MULTITHREADING threading 			= MULTITHREADING.NO_MULTITHREADING;
	private int mt_finished_count				= 0;
	private int mt_count						= -1;
	private boolean mt_stop						= true;
	
	//merging facility for classification results 
	private boolean mergeResults				= false;
	private List<List<String>> inputMergeFiles	= new ArrayList<List<String>>();
	private List<String> outputMergeFiles		= new ArrayList<String>();
	
	//optimalization variables
	private boolean optimizer					= false;
	private int optimizer_start					= 0;
	private int optimizer_stop					= 100;
	private int optimizer_step					= 10;
	
	/**
	 * The various possible purposes.
	 * @author Michiel Van Bel
	 */
	public enum PURPOSE{/**The classification actions will try to score 
	an indicated functional site.*/FUNCTIONAL_SITE_PREDICTION,/**The classification actions
	will try to identify and score possible AS events*/AS_PREDICTION};
	
	private PURPOSE purpose 					= PURPOSE.FUNCTIONAL_SITE_PREDICTION;
	
	private boolean toGenomeViewOutput			= false;
	
	
	
	/**
	 * Constructor, does nothing really but assigning the provided
	 * method parameters to local variables.
	 * @param logger The logging facility
	 * @param fileName The name of the configurationfile by which this workflow
	 * 					is created. Can be null if this method is invoked by other means.
	 * 	
	 */
	public FunSiPWorkflow(Logger logger, String fileName){
		this.logger = logger;
		this.fileName  = fileName;
	}
	
	
	
	
	/**
	 * This method is used when additional fastafiles that need to be evaluated are
	 * provided on the commandline by the user. This method creates lists of files
	 * and to-merge-files and adds these to the workflow.
	 * @param files The files that need to be evaluated.
	 */
	public void setExplicitFiles(File[] files){
		
		if(files != null && files.length!=0){
			
			/*
			 * inputmergefiles consists of a list of lists of 2 strings each
			 * outputmergfile consists of a list of same size as inputmergefiles
			 */
			inputMergeFiles = new ArrayList<List<String>>();
			outputMergeFiles = new ArrayList<String>();
			
			/*
			 * get all the paths to the files that need to be evaluated
			 */
			String [] filePaths = new String[files.length];
			for(int i=0;i<files.length;i++){
				filePaths[i]= files[i].getPath();
			}
			
			
			/*
			 * Adds the list of files that need to be evaluated (and
			 * the corresponding output-strings) to each classificationaction
			 * 
			 * Each outputfilepath is unique, because it consists of both
			 * the name of the classificationaction and the name of the file
			 * that needs to be evaluated
			 */
			for(int i=0;i<caList.size();i++){
				ClassificationAction action = caList.get(i);
				String n = action.getClassification_name();
				String[] outFilePaths = new String[filePaths.length];
				for(int j=0;j<files.length;j++){
					outFilePaths[j] = n+"_"+files[j].getName()+".classif";
				}
				action.setTestfiles(filePaths);
				action.setPredictionOutputFiles(outFilePaths);
			}

			
			/*
			 * Creation of the lists with files that need to be merged.
			 * It needs to adapt to the fact that each classificationaction
			 * might have a different outputdirectory
			 */
			List<String[]> outFiles = new ArrayList<String[]>();
			List<String> outputDirs = new ArrayList<String>();
			for(int i=0;i<caList.size();i++){ //all ca's should have same amount of testfiles...
				ClassificationAction action = caList.get(i);
				String outputDir = action.getOutputdirectory();
				outputDir = sanitizeOutputDir(outputDir);
				outputDirs.add(outputDir);
				String[] outActionFiles = action.getPredictionOutputFiles();
				String[] newActionFiles = new String[outActionFiles.length];
				for(int j=0;j<outActionFiles.length;j++){
					String temp = outputDir+outActionFiles[j];
					if(temp.indexOf(" ")!=-1)
						temp = "\""+temp+"\"";
					newActionFiles[j] = temp;
				}
				
				outFiles.add(newActionFiles);
			}
			
			/*
			 * Finally, addition of all the to-be-merged files and their
			 * respective destiny's to this workflow-object. 
			 */
			String [] ex = outFiles.get(0);
			for(int i=0;i<ex.length;i++){
				List<String> tempMerge  = new ArrayList<String>();
				String outDir = null;
				for(int j=0;j<outFiles.size();j++){
					String[] f = outFiles.get(j);
					tempMerge.add(f[i]);
					outDir = outputDirs.get(j); //so we have at least one outputdir, being the last
				}
				
				String mergefile = outDir+files[i].getName()+".FunSiP_AD";
				if(mergefile.indexOf(" ")!=-1)
					mergefile = "\""+mergefile+"\"";
				outputMergeFiles.add(mergefile);
				inputMergeFiles.add(tempMerge);
				
			}
			
		}
		
	/*	for(int i=0;i<inputMergeFiles.size();i++){
			System.out.println(outputMergeFile.get(i));
			System.out.println("===============");
			List<String> imf = inputMergeFiles.get(i);
			for(int j=0;j<imf.size();j++)
				System.out.println(imf.get(j));
		}*/
	}
	
	private String sanitizeOutputDir(String s){
		String temp = s;
		if(temp.startsWith("."))
			temp = temp.substring(1,temp.length());
		if(temp.startsWith("/")||temp.startsWith("\\"))
			temp = temp.substring(1,temp.length());
		
		if(FunSiPJar.RUN_FROM_JAR){
			File f = new File(FunSiPJar.USER_DIR+temp);
			String ret = f.getAbsolutePath();
			if(ret.indexOf("/")==-1)
				ret+="\\";
			else
				ret+="/";
			ret.trim();
			return ret;
		}
		else
			return s;
	}
	
	
	
	/**
	 * Inherited because the workflow must be run as a thread. 
	 * The method does three different things :
	 * 	a) it checks whether the workflow should run in optimization mode,
	 * 		and -if true- changes the classificationactions to adapt to this.
	 *  b) it switches to the different methods according to the type of multithreading
	 *     that should be used
	 *  c) it checks whether the results should be merged, and acts accordingly
	 */
	public void run(){
		logger.debug("Starting workflow for configuration file "+fileName);
		
		if(optimizer){
			logger.debug("Optimization detected : converting all classificationfiles");
			optimizeClassifications();
		}
		else{
			ca2List = new ArrayList<List<ClassificationAction>>();
			ca2List.add(caList);
		}
		
		
		switch(threading){
		case NO_MULTITHREADING : no_multithreading();break;
		case SINGLE_PC_MULTITHREADING : single_pc_multithreading();break;
		case CLUSTER_MULTITHREADING : cluster_multithreading();break;
		default : 
			//shouldnt happen, bu you never know.
			logger.fatal("Unknown/unimplemented type of selected threading : "+threading);
			return;
		}
		
		//if necessary, merge the output
		if(mergeResults){
			logger.debug("Merging classification files.");
			mergeResults();
		}
		else{
			logger.debug("Skipping merge of classification files");
		}
		
		logger.debug("Ending workflow for configuration file "+fileName);
	}
	

	

	/*
	 * The following 2 methods create a new list of ClassificationActions that are
	 * to be executed by the workflow. This new list of classificationactions is used
	 * for performing the optimalization-procedure for finding the best combination
	 * of features with their respective best up/down stream ranges.
	 * The creation of the new list of classificationactions is done by taking every
	 * classificationfeature defined in the classificationactions, generating all the
	 * possible up/downstream combinations (is a O(n^2) operation) and placing each one
	 * of this combinations in a different classificationaction, which is then added to
	 * the list of classsificationactions (replacing the original list).
	 */
	private void optimizeClassifications(){
		
		ca2List = new ArrayList<List<ClassificationAction>>();
		//this loop attempts to create a lot of new actions for optimization by
		//considering each classificationFeature alone and interpreting the given
		//up- and downranges and mins and max's for range interpolation.
		int numActions = 0;
		for(int i=0;i<caList.size();i++){
			ClassificationAction action = caList.get(i);	
			List<ClassificationAction> newActions = new ArrayList<ClassificationAction>();
			
			if(action.getCrossvalidation_fold()==-1){
				action.setCrossvalidation_fold(10);
				action.setComplexityCrossvalidation(true);
			}
			List<ClassificationFeature> cfList = action.getClassificationFeatures();
			for(int j=0;j<cfList.size();j++){
				ClassificationFeature cf = cfList.get(j);
				List<ClassificationAction> generatedActions = generate_actions(cf,action);
				newActions.addAll(generatedActions);
			}
			numActions+=newActions.size();
			ca2List.add(newActions);
		}
		//caList = newActions;
		logger.info("# of actions after optimization expansion : "+numActions);
	}
	
	private List<ClassificationAction> generate_actions(ClassificationFeature cf,ClassificationAction action){
		List<ClassificationAction> result = new ArrayList<ClassificationAction>();
		
		for(int i=optimizer_start;i<=optimizer_stop;i+=optimizer_step){
			for(int j=optimizer_start;j<=optimizer_stop;j+=optimizer_step){
				ClassificationFeature feat = cf.make_copy();
				feat.setUpDownRanges(i, j);
				List<ClassificationFeature> newCfList = new ArrayList<ClassificationFeature>();
				newCfList.add(feat);
				ClassificationAction newAction = action.make_copy();
				newAction.setClassificationFeatures(newCfList);
				result.add(newAction);
			}
		}
		return result;
	}
	
	
	
	
	/*
	 * This method merges the results of classificationprocedures performed on a 
	 * fasta-file. This needs to be done in order for the program to be compatible
	 * with the SpliceMachine output, and because each of the different actions
	 * present a different classification (such as acceptor recognition, donor
	 * recognition, etc...).
	 */
	private void mergeResults(){
		logger.debug("Merging prediction files");
		for(int i=0;i<inputMergeFiles.size();i++){
			List<String> list = inputMergeFiles.get(i);
			File[] inputFiles = new File[list.size()];
			for(int j=0;j<list.size();j++){
				String te = list.get(j);
				if(te.startsWith("\"")&&te.endsWith("\""))
					te = te.substring(1,te.length()-1);			
				inputFiles[j] = new File(te);
			}
			String me = outputMergeFiles.get(i);
			if(me.startsWith("\"")&&me.endsWith("\""))
				me = me.substring(1,me.length()-1);
			
			File output = new File(me);
			try{
				IOUtils.mergeFiles(output,inputFiles);
			}
			catch(Exception exc){
				logger.error(exc);
			}
		}
		logger.debug("Merge of prediction files complete");
		
		if(toGenomeViewOutput){
			logger.debug("Converting standard output to genomeview output");
			for(int i=0;i<outputMergeFiles.size();i++){
				File file = new File(outputMergeFiles.get(i));
				convertToGenomeViewOutput(file);
			}
			logger.debug("Output converion complete");
		}
	}
	
	public void convertToGenomeViewOutput(File file){
		List<String> originalStrings = new ArrayList<String>();
		List<String> newStrings = new ArrayList<String>();
		try{
			BufferedReader reader = new BufferedReader(new FileReader(file));
			String s = reader.readLine();
			while(s!=null){
				originalStrings.add(s);
				s = reader.readLine();
			}
			reader.close();
			
			for(int i=0;i<originalStrings.size();i++){
				String seq = originalStrings.get(i);
				String[] split = seq.split(" ");
				StringBuffer b = new StringBuffer();
				b.append((i+1)+"\t");
				int site = Integer.parseInt(split[0]);
				if(split[1].equals("donor")){
					b.append((site)+"\t");
					b.append((site+1)+"\t");
					b.append(1+"\t");
				}
				else if(split[1].equals("donor_rev")){
					b.append(site+"\t");
					b.append((site-1)+"\t");
					b.append(-1+"\t");
				}
				else if(split[1].equals("acceptor")){
					b.append((site-2)+"\t");
					b.append((site-1)+"\t");
					b.append(1+"\t");
				}
				else if(split[1].equals("acceptor_rev")){
					b.append((site+1)+"\t");
					b.append((site)+"\t");
					b.append(-1+"\t");
				}
				b.append(split[1]+"\t");
				b.append("FunSiP"+"\t");
				b.append(split[2]);
				newStrings.add(b.toString());
			}
			
			BufferedWriter writer = new BufferedWriter(new FileWriter(file,false));
			for(int i=0;i<newStrings.size();i++){
				writer.write(newStrings.get(i)+"\n");
			}
			writer.close();
		}
		catch(Exception exc){
			logger.error("Exception occured in class "+this.getClass()+" in method convertToGenomeViewOutput()");
			logger.error(exc);
			return;
		}
	}
	
	
	

	
	/*
	 * Perform each of the classificationactions in a non-multithreading 
	 * way. This is done by serially running each classification action.
	 * 
	 * To avoid problems with the waitForSync()-method used for multithreading,
	 * the start_classification-method from the ClassificationAction is 
	 * directly invoked. 
	 */
	private void no_multithreading(){

		for(int up=0;up<ca2List.size();up++){
			List<ClassificationAction> caList = ca2List.get(up);
			
			List<OptimizationScore> scores = new ArrayList<OptimizationScore>();
			for(int i=0;i<caList.size();i++){
				logger.debug("Classification action "+(i+1));
				ClassificationAction ca = caList.get(i);
				ca.start_classification();
				if(optimizer){
					double[] res = getBestOptimizationResult(ca.getCrossValidationTabFile());
					ca.getCrossValidationTabFile().delete();
					ClassificationFeature feature = ca.getClassificationFeatures().get(0);
					int [] up_down = feature.getUpDownRanges();
					String name = feature.toNoRangeString();
					if(res!=null && res.length==2){
						OptimizationScore score = new OptimizationScore(up_down[0],up_down[1],res[1],res[0],name,i);
						scores.add(score);
					}
				}
			}
			if(optimizer){
				try{
					List<List<OptimizationScore>> splitScores = splitAccordingToFeature(scores);
					logger.info("Splitin scores according to feature type. Detected # : "+splitScores.size());
					List<OptimizationScore> bestScores = extractBestScores(splitScores);
					logger.info("Extracting best scores for each feature. #Feature types :"+bestScores.size());
					bestScores = sortScores(bestScores);
					logger.info("Sorting the best scores");
					doFinalOptimization(bestScores,up);
					logger.info("Final optimilization done");
				}
				catch(Exception exc){
					logger.error("Exception occured in class "+this.getClass()+" in method no_multithreading()");
					logger.error(exc);
				}
			}
		}
	}
	
	/*
	 * This method is used to relocate the outputfiles of a classification
	 * to the original outputdirectory. This is needed to ensure that the merging
	 * procedure might go forth as planned.
	 */
	private void copy_output_files_for_merging(List<ClassificationAction> mt_classification_actions){
		for(int i=0;i<mt_classification_actions.size();i++){
			ClassificationAction ca = mt_classification_actions.get(i);
			File outputDirectory = new File(ca.getOutputdirectory());
			File parentDirectory = outputDirectory.getParentFile();  //is the original outputdirectory
			
			String[] predictionOutputFiles  = ca.getPredictionOutputFiles();
			for(int j=0;j<predictionOutputFiles.length;j++){
				File current_output_file = new File(outputDirectory.getPath()+"/"+predictionOutputFiles[j]);
				File new_output_file = new File(parentDirectory.getPath()+"/"+predictionOutputFiles[j]);
				
				//start copying from current to new
				try{
					BufferedReader bReader = new BufferedReader(new FileReader(current_output_file));
					BufferedWriter bWriter = new BufferedWriter(new FileWriter(new_output_file));
					String s = bReader.readLine();
					while(s!=null){
						bWriter.write(s+"\n");
						s = bReader.readLine();
					}
					bReader.close();
					bWriter.close();
				}
				catch(Exception exc){
					logger.error("Error while copying the output prediction files");
					logger.error("Exception occured in class "+this.getClass()+" in method copy_output_files_for_merging()");
					logger.error(exc);
				}
			}
				
		}
	}
	
	/*
	 * Perform each of the classificationactions in a local
	 * multithreading environment. This is done by starting all the 
	 * classificationactions in a different thread (so no direct use
	 * of a threadpool of any sort). This thread then waits until all
	 * other spawned threads are finished.
	 */
	private void single_pc_multithreading(){
		try{
			for(int up = 0; up<ca2List.size();up++){
				List<ClassificationAction> caList = ca2List.get(up);
				
				mt_finished_count = 0;
				for(int i=0;i<caList.size();i++){
					ClassificationAction ca = caList.get(i);

					//insert new outputdirectories to prevent filename-clashes
					ca.setOutputdirectory(ca.getOutputdirectory()+"output_"+i+"/");
					Thread t = new Thread(ca);
					Thread.sleep(100);
					t.start();
				}

				mt_count = caList.size();

				while(mt_stop){
					Thread.sleep(1000);	//busy loop while waiting for all actions to complete
				}	
				mt_stop = true;
				mt_count = 0;
				mt_finished_count = 0;

				//apply the necessary operations to accomodate for merging of outputfiles
				if(inputMergeFiles.size()!=0)
					copy_output_files_for_merging(caList);

				if(optimizer){
					List<OptimizationScore> scores = new ArrayList<OptimizationScore>();
					//get all the crossvalidation data from the classificationActions
					for(int i=0;i<caList.size();i++){
						ClassificationAction ca = caList.get(i);
						double[] res = getBestOptimizationResult(ca.getCrossValidationTabFile());

						//deletion of both tab-file and directory, then rename outputdir to old one 
						ca.getCrossValidationTabFile().delete();
						File outputDir = new File(ca.getOutputdirectory());
						String parName = outputDir.getParent();
						if(!parName.endsWith("/"))
							parName = parName+"/";
						ca.setOutputdirectory(parName);
						deleteFilesAndDirectory(outputDir);


						ClassificationFeature feature = ca.getClassificationFeatures().get(0);
						int [] up_down = feature.getUpDownRanges();
						String name = feature.toNoRangeString();
						OptimizationScore score = new OptimizationScore(up_down[0],up_down[1],res[1],res[0],name,i);
						scores.add(score);
					}
					List<List<OptimizationScore>> splitScores = splitAccordingToFeature(scores);
					List<OptimizationScore> bestScores = extractBestScores(splitScores);
					logger.debug("Optimizer : Best scores extracted");
					bestScores = sortScores(bestScores);
					doFinalOptimization(bestScores,up);
					logger.info("Final optimilization done");
					Thread.sleep(2000);
				}
			}
		}
		catch(Exception exc){
			logger.error("Exception occured in class "+this.getClass()+" in method single_pc_multithreading()");
			logger.error(exc);
			exc.printStackTrace();
			
		}
		
	}
	
	/*
	 * Deletes all the files in the temporary outputdirectories,
	 * or marks them for deletion on JVM-exit.
	 */
	private void deleteFilesAndDirectory(File dir){
		File[] files = dir.listFiles();
		for(File f : files){
			//lower level dir contains roc-curves, delete them also
			if(f.isDirectory()){
				File[] rocFiles = f.listFiles();
				for(File rf : rocFiles){
					rf.delete();
					rf.deleteOnExit();
				}
			}
			f.delete();
			f.deleteOnExit();
		}
		dir.delete();
		dir.deleteOnExit();
	}
	
	/**
	 * This method is used for synchronization purposes between the different
	 * spawned threads. Every thread thats is finished, is halted by this
	 * method (acting as a barrier). Once all the threads have stopped, workflow
	 * is resumed.
	 *
	 */
	public synchronized void waitForSync() throws InterruptedException{
		mt_finished_count++;
		if(mt_count==-1)
			mt_count = caList.size();	//just in case the mt_count is not set!
		if(mt_finished_count==mt_count){
			mt_stop = false;
			this.notifyAll();
		}
		else while(mt_finished_count<mt_count){
			this.wait();
		}
	}
	

	/*
	 * This method tries to launch all the jobs on the cluster.
	 * This is done by launching all the different classificationactions
	 * in a localized small listening thread, which effectively launches 
	 * the jobs on the cluster
	 */
	private void cluster_multithreading(){
		try{
			
			for(int up=0;up<ca2List.size();up++){
				List<ClassificationAction> caList = ca2List.get(up);

				mt_finished_count = 0;
				List<ClassificationAction> clusterCaList = new ArrayList<ClassificationAction>();

				
				/*
				 * Create a clusterclassification action for each classificationaction
				 * Launch each clusterclassificationaction on the cluster, and wait untill
				 * all of them have finished.
				 */
				for(int i=0;i<caList.size();i++){
					ClassificationAction ca = caList.get(i);

					String[] testFiles = ca.getTestfiles();
					String[] outputFiles = ca.getPredictionOutputFiles();

					if(testFiles == null || testFiles.length==0){	//e.g. optimalisation.
						ca.setOutputdirectory(ca.getOutputdirectory()+"output_"+i+"/");
						clusterCaList.add(ca);
						ClusterClassificationAction cca = new ClusterClassificationAction(ca,this,""+i,logger);
						Thread t = new Thread(cca);
						t.start();
						Thread.sleep(1000);	//otherwise it my be that the clustersoftware cannot coop.
					}
					else{
						for(int j=0;j<testFiles.length;j++){
							ClassificationAction copy = ca.make_copy();
							String[] newTestFiles = {testFiles[j]};
							String[] newOutputFiles = {outputFiles[j]};
							copy.setTestfiles(newTestFiles);
							copy.setPredictionOutputFiles(newOutputFiles);

							String index = i+"_"+j;

							copy.setOutputdirectory(ca.getOutputdirectory()+"output_"+index+"/");
							clusterCaList.add(copy);

							ClusterClassificationAction cca = new ClusterClassificationAction(copy,this,index,logger);
							Thread t = new Thread(cca);
							t.start();					
							Thread.sleep(1000); //in order to avoid timing-clashes with qsub command
						}
					}
				}

				mt_count = clusterCaList.size();

				while(mt_stop){
					Thread.sleep(10000);	//busy loop while waiting for all actions to complete
				}	

				mt_finished_count = 0;
				mt_count = 0;
				mt_stop = true;
				
				//apply the necessary operations to accomodate for merging of outputfiles
				if(inputMergeFiles.size()!=0)
					copy_output_files_for_merging(clusterCaList);

				if(optimizer){
					List<OptimizationScore> scores = new ArrayList<OptimizationScore>();
					//get all the crossvalidation data from the classificationActions
					for(int i=0;i<caList.size();i++){
						ClassificationAction ca = caList.get(i);
						double[] res = getBestOptimizationResult(ca.getCrossValidationTabFile());

						//deletion of both tab-file and directory, then rename outputdir to old one 
						ca.getCrossValidationTabFile().delete();
						File outputDir = new File(ca.getOutputdirectory());
						String parName = outputDir.getParent();
						if(!parName.endsWith("/"))
							parName = parName+"/";
						ca.setOutputdirectory(parName);
						deleteFilesAndDirectory(outputDir);


						ClassificationFeature feature = ca.getClassificationFeatures().get(0);
						int [] up_down = feature.getUpDownRanges();
						String name = feature.toNoRangeString();
						OptimizationScore score = new OptimizationScore(up_down[0],up_down[1],res[1],res[0],name,i);
						scores.add(score);
					}
					List<List<OptimizationScore>> splitScores = splitAccordingToFeature(scores);
					List<OptimizationScore> bestScores = extractBestScores(splitScores);
					logger.debug("Optimizer : Best scores extracted");
					bestScores = sortScores(bestScores);
					doFinalOptimization(bestScores,up);
					logger.info("Final optimilization done");
				}
			}
		}
		catch(Exception exc){
			logger.error("Exception occured in class "+this.getClass()+" in method cluster_multithreading()");
			logger.error(exc);
		}
	}
	
	/*
	 * This method tries to launch all the jobs on the cluster.
	 * This is done by launching all the different classificationactions
	 * in a localized small listening thread, which effectively launches 
	 * the jobs on the cluster
	 *
	private void cluster_multithreading(){
		try{
			mt_finished_count = 0;
			for(int i=0;i<caList.size();i++){
				ClassificationAction ca = caList.get(i);
				
				//insert new outputdirectories to prevent filename-clashes
				ca.setOutputdirectory(ca.getOutputdirectory()+"output_"+i+"/");
				ClusterClassificationAction cca = new ClusterClassificationAction(ca,this,i,logger);
				
				Thread t = new Thread(cca);
				t.start();
			}
			while(mt_stop){
				Thread.sleep(10000);	//busy loop while waiting for all actions to complete
			}	
			
			//apply the necessary operations to accomodate for merging of outputfiles
			if(inputMergeFiles.size()!=0)
				copy_output_files_for_merging();
			
			if(optimizer){
				List<OptimizationScore> scores = new ArrayList<OptimizationScore>();
				//get all the crossvalidation data from the classificationActions
				for(int i=0;i<caList.size();i++){
					ClassificationAction ca = caList.get(i);
					double[] res = getBestOptimizationResult(ca.getCrossValidationTabFile());
					
					//deletion of both tab-file and directory, then rename outputdir to old one 
					ca.getCrossValidationTabFile().delete();
					File outputDir = new File(ca.getOutputdirectory());
					String parName = outputDir.getParent();
					if(!parName.endsWith("/"))
						parName = parName+"/";
					ca.setOutputdirectory(parName);
					deleteFilesAndDirectory(outputDir);
					
					
					ClassificationFeature feature = ca.getClassificationFeatures().get(0);
					int [] up_down = feature.getUpDownRanges();
					String name = feature.toNoRangeString();
					OptimizationScore score = new OptimizationScore(up_down[0],up_down[1],res[1],res[0],name,i);
					scores.add(score);
				}
				List<List<OptimizationScore>> splitScores = splitAccordingToFeature(scores);
				List<OptimizationScore> bestScores = extractBestScores(splitScores);
				logger.debug("Optimizer : Best scores extracted");
				bestScores = sortScores(bestScores);
				doFinalOptimization(bestScores);
				logger.info("Final optimilization done");
			}
		}
		catch(Exception exc){
			logger.error(exc);
		}
	}*/
	
	
	/*
	 * This method starts the optimalization process of the workflow by 
	 * dynamically altering all the ClassificationActions and the contained
	 * ClassificationFeatures within. Indeed, for every type of ClassificationFeature,
	 * all the ranges (defined by variables in the config-file) are created and placed
	 * within a single ClassificationAction (each one does a complexity-crossvalidation).
	 * 
	 * Then, the best classification-results are extracted, and placed in a list. The list
	 * is then divided in sublists so every sublist contains only data from a single 
	 * ClassificationFeature (e.g. positional, length 1). From these sublists, again the best
	 * results are extracted. This results in a list which contains classificationfeatures 
	 * from distinct types only, with the best up-down-ranges and the best complexity's.
	 * 
	 * This list is then sorted in descending order, according to the crossvalidationresults
	 * defined by the F1-measure.
	 * 
	 * Then, the final optimalization is performed by steadily increasing the list of final
	 * classificationfeatures and performing crossvalidation on each increased list. If the
	 * result is better then the previous one, the classificationfeature is held in the list,
	 * otherwise it is removed. 
	 * 
	 * We acknowledge that this is a rather heuristic greedy approach, but a brute force - go
	 * through all possible combinations - approach is in no way achievable with current (and
	 * distant future) generations of processing power.
	 */
	private void doFinalOptimization(List<OptimizationScore> scores, int index){
		try{
			List<ClassificationAction> caList = ca2List.get(index);
			
			String outputFileName = caList.get(0).getOutputdirectory()+caList.get(0).getClassification_name()+"_optimalization.txt";
			BufferedWriter writer = new BufferedWriter(new FileWriter(new File(outputFileName)));
			writeScores(writer,scores,"All features");
			
			List<OptimizationScore> finalList = new ArrayList<OptimizationScore>();
			String messg = "Selected features";
			if(scores!=null && scores.size()>=1){
				OptimizationScore firstScore = scores.get(0);
				finalList.add(firstScore);
				double bestScore = firstScore.f1_score;
				double bestComplexity = 0.0;
				List<ClassificationFeature> best_features = new ArrayList<ClassificationFeature>();
				List<ClassificationFeature> test_features = new ArrayList<ClassificationFeature>();
				best_features.addAll(caList.get(firstScore.caListIndex).getClassificationFeatures());
				test_features.addAll(caList.get(firstScore.caListIndex).getClassificationFeatures());
				
				for(int i=1;i<scores.size();i++){
					OptimizationScore score = scores.get(i);
					ClassificationAction ca = caList.get(score.caListIndex);
					List<ClassificationFeature> cfList = ca.getClassificationFeatures();
					test_features.addAll(cfList);
					ca.setClassificationFeatures(test_features);
					ca.start_classification();
					double[] res = getBestOptimizationResult(ca.getCrossValidationTabFile());
					if(res[0]>bestScore){
						bestScore = res[0];
						bestComplexity = res[1];
						best_features.clear();
						best_features.addAll(test_features);
						score.f1_score = bestScore;
						finalList.add(score);
					}
					else{
						test_features.clear();
						test_features.addAll(best_features);
					}
				}
				messg+=" -> complexity : "+bestComplexity;
			}
			
			writeScores(writer,finalList,messg);
			writer.close();
		}
		catch(Exception exc){
			logger.error("Exception occured in class "+this.getClass()+" in method doFinalOptimization()");
			logger.error(exc);
		}
	}
	
	/*
	 * Used for writing the results of the optimalization
	 */
	private void writeScores(BufferedWriter writer,List<OptimizationScore> scores,String titel)throws IOException{
		writer.write(titel+"\n");
		writer.write("--------------------------------\n\n");
		for(int i=0;i<scores.size();i++){
			OptimizationScore score = scores.get(i);
			writer.write(score.classification_feat+" "+score.variable_one+" "+score.variable_two+" -->"+score.f1_score+"\n");
		}
		writer.write("\n\n\n\n");
	}
	
	/*
	 * This method extracts the best complexity for the best F1-measure
	 * from a given complexityCrossvalidation. 
	 * Care should be taken to let this method conform to the file-layout.
	 * Therefore we have chosen to locate the right tab dynamically but searching
	 * in the header for the F1-measure string.
	 */
	private double[] getBestOptimizationResult(File file){
		double[] result = new double[2];
		try{
			BufferedReader reader = new BufferedReader(new FileReader(file));
			String s = reader.readLine();	//first line only contains the header
			String[] split = s.split("\t");
			int f1_location = 0;
			int complexity_location = 0;
			for(int i=0;i<split.length;i++){
				if(split[i].equals("F1_MEASURE"))
					f1_location = i;
				if(split[i].equals("SVM_complex"))
					complexity_location = i;
			}
			
			double best_f1 = 0.0;
			double best_complexity = 0.0;
			s = reader.readLine();
			while(s!=null){
				String[] values = s.split("\t");
				double f1_measure = Double.parseDouble(values[f1_location]);
				double complexity = Double.parseDouble(values[complexity_location]);
				if(f1_measure>best_f1){
					best_f1 = f1_measure;
					best_complexity = complexity;
				}
				s = reader.readLine();
			}
			
			result[0] = best_f1;
			result[1] = best_complexity;
			reader.close();
		}
		catch(Exception exc){
			logger.error("Exception occured in class "+this.getClass()+" in method getBestOptimizationResult()");
			logger.error(exc);
			exc.printStackTrace();
		}
		return result;
	}
	
	/*
	 * This method splits the list of OptimizationScores in a list of sublists, 
	 * in order to facilitate the optimization.
	 * Each sublist consists of classificationfeatures with the same type (name+length, e.g. 
	 * Positional 1), but with different upstream and downstream ranges.
	 * 
	 * Method should be foolprove.
	 */
	private List<List<OptimizationScore>> splitAccordingToFeature (List<OptimizationScore> list){
		List<List<OptimizationScore>> result = new ArrayList<List<OptimizationScore>>();
		List<String> classification_types = new ArrayList<String>();
		
		//locate all the different types, use a list instead of a set because we need the order
		for(int i=0;i<list.size();i++){
			OptimizationScore score = list.get(i);
			String name = score.classification_feat;
			if(!classification_types.contains(name))
				classification_types.add(name);
		}
		
		//make the necessary sublists
		for(int i=0;i<classification_types.size();i++){
			result.add(new ArrayList<OptimizationScore>());
		}
		
		//put all optimizationscores in the correct sublists
		for(int i=0;i<list.size();i++){
			OptimizationScore score = list.get(i);
			if(score!=null){
				String type = score.classification_feat;
				int index = classification_types.indexOf(type); //index cannot be -1, due to above procedure
				result.get(index).add(score);
			}
		}
		
		return result;
	}
	
	/*
	 * This method extracts for each sublist the best score, and all these scores
	 * with a unique type are then added to the resulting list.
	 */
	private List<OptimizationScore> extractBestScores(List<List<OptimizationScore>> scores){
		List<OptimizationScore> result = new ArrayList<OptimizationScore>();
		if(scores==null || scores.size()==0)
			return result;
		for(int i=0;i<scores.size();i++){
			List<OptimizationScore> sublist = scores.get(i);
			
			if(sublist!=null && sublist.size()>0){
				
				OptimizationScore best_score = sublist.get(0);
				double best_f1_score = best_score.f1_score;
			
				//sublist contains extact types, just different ranges
				for(int j=0;j<sublist.size();j++){
					OptimizationScore score = sublist.get(j);
					double f1_score = score.f1_score;
					if(f1_score>best_f1_score){
						best_f1_score = f1_score;
						best_score = score;
					}
				}
			
				//add best range optimizationscore to final list.
				result.add(best_score);
			}
		}
		return result;
	}

	/**
	 * Dummy bubblesort algorithm (normally list only consists of 2-10 items) 
	 * for the OptimizationScore type.
	 * @param list List that needs to be sorted (descending order)
	 * @return Sorted list
	 */
	public List<OptimizationScore> sortScores(List<OptimizationScore> list){
		List<OptimizationScore> sortedList = new ArrayList<OptimizationScore>();
		OptimizationScore[] score_array = new OptimizationScore[list.size()];
		for(int i=0;i<list.size();i++)
			score_array[i] = list.get(i);
		
		for(int i=0;i<score_array.length;i++){
			for(int j=i+1;j<score_array.length;j++){
				if(score_array[i].f1_score<score_array[j].f1_score){
					OptimizationScore temp = score_array[i];
					score_array[i] = score_array[j];
					score_array[j] = temp;
				}
			}
		}
		
		for(int i=0;i<score_array.length;i++)
			sortedList.add(score_array[i]);
		
		return sortedList;
	}
	
	
	
	/*
	 *************************************************
	 * Bunch of simple setters and getters
	 *************************************************
	 */
	public void addClassificationAction(ClassificationAction ca){
		if(ca!=null && caList!=null)
			caList.add(ca);
	}
	
	public List<ClassificationAction> getClassificationActions(){
		return caList;
	}
	
	public void setClassificationActions(List<ClassificationAction> caList){
		this.caList = caList;
	}
	
	public boolean setThreading(String threading_style) {
		boolean found = false;
		threading_style = threading_style.toUpperCase();
		for(MULTITHREADING mt : MULTITHREADING.values()){
			if(threading_style.equals(mt.toString())){
				threading = mt;
				found  = true;
				break;
			}
		}
		return found;
	}
	

	public PURPOSE getPurpose() {
		return purpose;
	}

	public void setPurpose(PURPOSE purpose) {
		this.purpose = purpose;
	}
	
	public boolean setPurpose(String s){
		boolean found = false;
		s = s.toUpperCase();
		for(PURPOSE p : PURPOSE.values()){
			if(s.equals(p.toString())){
				this.purpose = p;
				found = true;
				break;
			}
		}
		return found;
	}
	
	public String getLoggingStyle(){
		return logger.getLevel().toString();
	}
	
	public MULTITHREADING getThreading(){
		return threading;
	}
	
	public List<List<String>> getInputMergeFiles() {
		return inputMergeFiles;
	}
	public void setInputMergeFiles(List<List<String>> inputMergeFiles) {
		this.inputMergeFiles = inputMergeFiles;
	}
	public boolean isMergeResults() {
		return mergeResults;
	}
	public void setMergeResults(boolean mergeResults) {
		this.mergeResults = mergeResults;
	}
	public List<String> getOutputMergeFile() {
		return outputMergeFiles;
	}
	public void setOutputMergeFile(List<String> outputMergeFile) {
		this.outputMergeFiles = outputMergeFile;
	}
	public void addInputMergeFiles(List<String> inputMerge){
		this.inputMergeFiles.add(inputMerge);
	}
	public void addOutputMergeFile(String outputMerge){
		this.outputMergeFiles.add(outputMerge);
	}
	public boolean isOptimizer() {
		return optimizer;
	}
	public void setOptimizer(boolean optimizer) {
		this.optimizer = optimizer;
	}
	public int getOptimizer_step() {
		return optimizer_step;
	}
	public void setOptimizer_steps(int optimizer_start,int optimizer_stop, int optimizer_step) {
		this.optimizer_start 	= optimizer_start;
		this.optimizer_stop 	= optimizer_stop;
		this.optimizer_step 	= optimizer_step;
	}
	public int[]getOptimizerData(){
		int[] res = {optimizer_start,optimizer_stop,optimizer_step};
		return res;
	}
	public void setFileName(String s){
		this.fileName = s;
	}
	public String getFileName(){
		return this.fileName;
	}
	
	
	
	/*
	 *****************************************************************************************
	 *****************************************************************************************
	 *		Private class used for containing the results of the crossvalidation
	 *		that is performed for optimalization of the classification features
	 *****************************************************************************************
	 *****************************************************************************************
	 */
	private class OptimizationScore implements Comparable{
		
		public OptimizationScore(int var1,int var2,double compl,double f1,String classif, int index){
			this.variable_one 			= var1;
			this.variable_two 			= var2;
			this.complexity 			= compl;
			this.f1_score				= f1;
			this.classification_feat	= classif;
			this.caListIndex			= index;
		}
		
		public int compareTo(Object os) throws ClassCastException{
			if(os instanceof OptimizationScore){
				OptimizationScore o = (OptimizationScore)os;
				int result = 0;
				if(this.f1_score<o.f1_score)
					result = -1;
				else if(this.f1_score==o.f1_score)
					result = 0;
				else	//this.f1_score > o.f1_score
					result = 1;
				
				return result;
			}
			else
				throw new ClassCastException();
		}
		
		public int variable_one 			= 0;
		public int variable_two				= 0;
		public double complexity			= 0;
		public double f1_score				= 0;
		public String classification_feat	= "";
		public int caListIndex				= -1;
	}




}
