/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
package base;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.util.ArrayList;
import java.util.List;

import org.apache.log4j.ConsoleAppender;
import org.apache.log4j.FileAppender;
import org.apache.log4j.Level;
import org.apache.log4j.Logger;
import org.apache.log4j.SimpleLayout;


import util.classificationFeatures.ClassificationFeature;
import util.classificationFeatures.ClassificationFeatureUtils;
import util.conversions.Conversion;
import util.featureSelection.FeatureSelectionStub.FEATURE_SELECTION_PURPOSE;
import util.featureSelection.FeatureSelectionStub.FEATURE_SELECTION_TYPE;

/**
 * This class is used for loading an parsing the parameterfile for the FunSiP
 * project. The parsing of the parameterfile should not take too long,
 * and care is taken that as many faulty inputs as possible are catched
 * by the system before poisining the workflow. 
 * @author Michiel Van Bel
 */
public final class ParameterFileParser {

	//logging facility, used for anouncing goal-completions or failures.
	private Logger logger = null;
	
	/*
	 * when rebuilding a configuration file, the parameters and their values are 
	 * seperated by a number of tabs. Because the parameters may vary in size, we 
	 * define a fixed number of tabs (one tab == 6 chars) that may be occupied by the
	 * parameter names or whitespace. Thus, we have gain a nicely printed new 
	 * configuration file. 
	 */ 
	public static final int MIN_TABS = 8;
	
	
	//Defines in what state the parser is. This is necessary for easy configuration
	//file parsing.
	private enum LOAD_OPTIONS{NO_LOAD,LOAD_GENERAL,LOAD_ACTION};
	

	/*
	 * Keys for indicating workflow
	 */
	/**Indicator for top-level option that regulates workflow */
	public static final String KEY_GENERAL_OPTIONS 	= "GENERAL_OPTIONS";
	public static final String KEY_ACTION		   	= "ACTION";
	public static final String KEY_END				= "END";
	
	/*
	 * Keys for indicating options within the GENERAL_OPTIONS frame
	 */
	public static final String KEY_GO_MULTITHREADING		= "MULTITHREADING";
	public static final String KEY_GO_LOGGING_STYLE			= "LOGGING_STYLE";
	public static final String KEY_GO_LOGGING_OUTPUT		= "LOGGING_OUTPUT";
	public static final String KEY_GO_MERGE_PREDICTION		= "MERGE_PREDICTION_OUTPUT";
	public static final String KEY_GO_OPTIMIZER				= "OPTIMIZER";
	public static final String KEY_GO_PURPOSE				= "PURPOSE";
	
	/*
	 * Keys for indicating options within the ACTION frame
	 */
	public static final String KEY_AC_NAME						= "NAME";
	public static final String KEY_AC_CLASSIFICATION_NAME		= "CLASSIFICATION_NAME";
	public static final String KEY_AC_CLASSIFICATION_FEATURE	= "CLASSIFICATION_FEATURE";
	public static final String KEY_AC_OUTPUT					= "OUTPUT_DIRECTORY";
	public static final String KEY_AC_COMMAND					= "COMMAND";
	public static final String KEY_AC_MODEL_FILE				= "MODEL_FILE";
	public static final String KEY_AC_CLASSIFIER_TYPE			= "CLASSIFIER_TYPE";
	public static final String KEY_AC_CLASSIFIER_OPTIONS		= "CLASSIFIER_OPTIONS";
	public static final String KEY_AC_POSITIVE_TRAINING_FILE	= "POSITIVE_TRAINING_FILE";
	public static final String KEY_AC_NEGATIVE_TRAINING_FILE	= "NEGATIVE_TRAINING_FILE";
	public static final String KEY_AC_POSITIVE_TRAINING_FILE_SS = "POSITIVE_TRAINING_FILE_SECSTRUCT";
	public static final String KEY_AC_NEGATIVE_TRAINING_FILE_SS = "NEGATIVE_TRAINING_FILE_SECSTRUCT";
	public static final String KEY_AC_FUNCSITE_POSITION			= "SPLICESITE_POSITION";
	public static final String KEY_AC_POSITIVE_TRAINING_AMOUNT	= "POSITIVE_TRAINING_AMOUNT";
	public static final String KEY_AC_NEGATIVE_TRAINING_AMOUNT	= "NEGATIVE_TRAINING_AMOUNT";
	public static final String KEY_AC_FEATURES_OUTPUT			= "FEATURES_OUTPUT_FILE";
	public static final String KEY_AC_TEST_FILES				= "TEST_FILES";
	public static final String KEY_AC_OUTPUT_PREDICTION_FILES	= "OUTPUT_PREDICTION_FILES";
	public static final String KEY_AC_FILE_TYPE					= "FILE_TYPE";
	public static final String KEY_AC_STRAND					= "STRAND";
	public static final String KEY_AC_CLASSIFICATION_PATTERN	= "CLASSIFICATION_PATTERN";
	public static final String KEY_AC_CLASSIFICATION_PATTERN_LOC= "CLASSIFICATION_PATTERN_LOCATION";
	public static final String KEY_AC_CLASSIFICATION_PATTERN_OUT= "CLASSIFICATION_PATTERN_OUTPUT";
	public static final String KEY_AC_CROSSVALIDATION			= "CROSSVALIDATION";
	public static final String KEY_AC_COMPlEX_CROSSVALIDATION	= "COMPLEXITY_CROSSVALIDATION";
	public static final String KEY_AC_MAX_CROSSVALIDATION		= "MAXIMUM_CROSSVALIDATION";
	public static final String KEY_AC_FEATURE_SELECTION_TYPE	= "FEATURE_SELECTION_TYPE";
	public static final String KEY_AC_FEATURE_SELECTION_PURPOSE = "FEATURE_SELECTION_PURPOSE";
	public static final String KEY_AC_FEATURE_SELECTION_INPUT	= "FEATURE_SELECTION_INPUT_FILE";
	public static final String KEY_AC_FEATURE_SELECTION_OUTPUT	= "FEATURE_SELECTION_OUTPUT_FILE";
	public static final String KEY_AC_FEATURE_SELECTION_MAX		= "FEATURE_SELECTION_MAX_FEATURES";
	
	
	public static final String[] POSSIBLE_KEYS = {
		KEY_GENERAL_OPTIONS,KEY_ACTION,KEY_END,
		
		KEY_GO_MULTITHREADING,KEY_GO_LOGGING_STYLE,KEY_GO_LOGGING_OUTPUT,
		KEY_GO_MERGE_PREDICTION,KEY_GO_OPTIMIZER,KEY_GO_PURPOSE,
		
		KEY_AC_NAME,KEY_AC_CLASSIFICATION_NAME,KEY_AC_CLASSIFICATION_FEATURE,KEY_AC_OUTPUT,
		KEY_AC_COMMAND,KEY_AC_MODEL_FILE,KEY_AC_CLASSIFIER_TYPE,KEY_AC_CLASSIFIER_OPTIONS,
		KEY_AC_POSITIVE_TRAINING_FILE,KEY_AC_NEGATIVE_TRAINING_FILE,
		KEY_AC_POSITIVE_TRAINING_FILE_SS,KEY_AC_NEGATIVE_TRAINING_FILE_SS,
		KEY_AC_FUNCSITE_POSITION,KEY_AC_POSITIVE_TRAINING_AMOUNT,KEY_AC_NEGATIVE_TRAINING_AMOUNT,
		KEY_AC_FEATURES_OUTPUT,KEY_AC_TEST_FILES,KEY_AC_OUTPUT_PREDICTION_FILES,
		KEY_AC_FILE_TYPE,KEY_AC_STRAND,KEY_AC_CLASSIFICATION_PATTERN,KEY_AC_CLASSIFICATION_PATTERN_LOC,
		KEY_AC_CLASSIFICATION_PATTERN_OUT,KEY_AC_CROSSVALIDATION,KEY_AC_COMPlEX_CROSSVALIDATION,
		KEY_AC_MAX_CROSSVALIDATION,KEY_AC_FEATURE_SELECTION_TYPE,KEY_AC_FEATURE_SELECTION_PURPOSE,
		KEY_AC_FEATURE_SELECTION_INPUT,KEY_AC_FEATURE_SELECTION_OUTPUT,KEY_AC_FEATURE_SELECTION_MAX
	};
	
	
	/**
	 * Constructor for the Parameter file parser.
	 * @param logger The logging facility used by the parser.
	 */
	public ParameterFileParser(Logger logger){
		this.logger = logger;
	}
	
	
	/**
	 * Produce a configline string which has a fixed length,
	 * according to the number of tabs defined. 
	 * @param command The inital string that is to be incorporated into the final string
	 * @return The config line string
	 */
	private static String getConfigString(String command){
		if(command==null)
			return null;
		int len = command.length();
		int curT = len/6;	//curT == current tabs length
		if(len%6!=0)
			curT++;
		StringBuffer buf = new StringBuffer();
		buf.append("\t"+command);
		for(int i=0;i<=(MIN_TABS-curT);i++)
			buf.append("\t");
		return buf.toString();
	}
	
	/**
	 * This method creates a string that contains both a parameter and
	 * its value. The line is produced in the most aesthetically feasible
	 * way. 
	 * @param command The parameter command
	 * @param result The parameter value
	 * @return The resulting string.
	 */
	public static String getConfigString(String command,String result){
		if(command==null||result==null)
			return null;
		int len = command.length();
		int curT = len/6;	//curT == current tabs length
		if(len%6!=0)
			curT++;
		StringBuffer buf = new StringBuffer();
		buf.append("\t"+command);
		for(int i=0;i<=(MIN_TABS-curT);i++)
			buf.append("\t");
		buf.append(result+"\n");
		return buf.toString();
	}
	
	/**
	 * This method creates a string that contains both a parameter and its
	 * values. The lines is produced in the most aesthetically feasible way.
	 * @param command The parameter command.
	 * @param result The parameter values
	 * @return The resulting string
	 */
	public static String getConfigString(String command,String[]result){
		if(command==null||result==null)
			return null;
		int len = command.length();
		int curT = len/6;	//curT == current tabs length
		if(len%6!=0)
			curT++;
		StringBuffer buf = new StringBuffer();
		buf.append("\t"+command);
		for(int i=0;i<=(MIN_TABS-curT);i++)
			buf.append("\t");
		for(int i=0;i<result.length-1;i++)
			buf.append(result[i]+" ");
		buf.append(result[result.length-1]+"\n");
		return buf.toString();
	}
	
	
	/**
	 * This method creates a single string that contains the entire
	 * configuration file. It first defines the data for the workflow, after which
	 * it iterates over the various classification actions that are defined in the
	 * workflow. 
	 * This method can for example by used by interactive programs that change the
	 * content of a configuration file, and want this changes to be kept.
	 * @param workflow The workflow for which a configuration file needs to be defined.
	 * @return The string that consists of the entire content of the configuration
	 * file associated with the workflow. 
	 */
	public static String getConfigurationFileString(FunSiPWorkflow workflow){
		//smashing pumpkins rule!
		StringBuffer result = new StringBuffer();
		if(workflow==null)
			return result.toString();
		//get string for workflow commands
		result.append("# Automatically generated configurationfile \n");
		result.append("# Generated by FunSiP-program version "+main.FunSiP.VERSION +" \n");
		result.append("# --------------------------------------------------------------------- \n");
		result.append(KEY_GENERAL_OPTIONS+"\n");
		result.append(getConfigString(KEY_GO_MULTITHREADING,workflow.getThreading().toString()));
		
		if(workflow.getInputMergeFiles().size()!=0){//we assume workflow is consistent
			List<List<String>>imf = workflow.getInputMergeFiles();
			List<String> outS = workflow.getOutputMergeFile();
			for(int i=0;i<imf.size();i++){
				result.append(getConfigString(KEY_GO_MERGE_PREDICTION));
				List<String> ls = imf.get(i);
				String outString = outS.get(i);
				for(int j=0;j<ls.size();j++){
					result.append(ls.get(j)+" ");
				}
				result.append(outString+"\n");
			}
		}
		
		if(workflow.isOptimizer()){
			int[] optData = workflow.getOptimizerData();
			result.append(getConfigString(KEY_GO_OPTIMIZER)+optData[0]+" "+optData[1]+" "+workflow.getOptimizer_step()+"\n");
		}
		result.append(KEY_END+"\n\n");
		
		
		//get strings for classificationaction commands
		List<ClassificationAction> caList = workflow.getClassificationActions();
		for(int i=0;i<caList.size();i++){
			String st = getConfigurationFileString(caList.get(i));
			result.append(st);
		}
		return result.toString();
	}
	
	
	/**
	 * This method creates the string that is associated with the given classification
	 * action. It is clear that the produced string cannot be used on its own. Instead,
	 * the produced string should be used inside a workflow command string.
	 * @param ca The classification action for which the configuration file string
	 * should be produced.
	 * @return The configuration file string associated with the provided classification
	 * action.
	 */
	public static String getConfigurationFileString(ClassificationAction ca){
		StringBuffer result = new StringBuffer();
		if(ca==null)
			return result.toString();
		result.append(KEY_ACTION+"\n");
		//name
		if(ca.getName()!=null)
			result.append(getConfigString(KEY_AC_NAME,ca.getName()));
		//classification name
		if(ca.getClassification_name()!=null)
			result.append(getConfigString(KEY_AC_CLASSIFICATION_NAME,ca.getClassification_name()));
		//classifier type
		result.append(getConfigString(KEY_AC_CLASSIFICATION_FEATURE,ca.getClassifier_type().toString()));
		//classifier options
		if(ca.getClassifier_options()!=null){
			result.append(getConfigString(KEY_AC_CLASSIFIER_OPTIONS));
			String[] opt = ca.getClassifier_options();
			for(int i=0;i<opt.length-1;i++)
				result.append(opt+" ");
			result.append(opt[opt.length-1]+"\n");
		}
		//classification features
		List<ClassificationFeature>cfList = ca.getClassificationFeatures();
		for(int i=0;i<cfList.size();i++){
			result.append(getConfigString(KEY_AC_CLASSIFICATION_FEATURE,cfList.get(i).toString()));
		}
		//positive training file
		if(ca.getPositive_training_file()!=null)
			result.append(getConfigString(KEY_AC_POSITIVE_TRAINING_FILE,ca.getPositive_training_file()));
		//negative training file
		if(ca.getNegative_training_file()!=null)
			result.append(getConfigString(KEY_AC_NEGATIVE_TRAINING_FILE,ca.getNegative_training_file()));
		//positive training ss file
		if(ca.getPositive_training_file_secstruct()!=null)
			result.append(getConfigString(KEY_AC_POSITIVE_TRAINING_FILE_SS,ca.getPositive_training_file_secstruct()));
		//negative training ss file
		if(ca.getNegative_training_file_secstruct()!=null)
			result.append(getConfigString(KEY_AC_NEGATIVE_TRAINING_FILE_SS,ca.getNegative_training_file_secstruct()));
		//output directory
		if(ca.getOutputdirectory()!=null)
			result.append(getConfigString(KEY_AC_OUTPUT,ca.getOutputdirectory()));
		//splicesite position
		if(ca.getTraining_functional_site_position()!=-1)
			result.append(getConfigString(KEY_AC_FUNCSITE_POSITION,ca.getTraining_functional_site_position()+""));
		//command
		result.append(getConfigString(KEY_AC_COMMAND,ca.getCommand().toString()));
		//model file
		if(ca.getModel_file()!=null)
			result.append(getConfigString(KEY_AC_MODEL_FILE,ca.getModel_file()));
		//features output
		if(ca.getFeatures_output_file()!=null)
			result.append(getConfigString(KEY_AC_FEATURES_OUTPUT,ca.getFeatures_output_file()));
		//test files
		if(ca.getTestfiles()!=null){
			result.append(getConfigString(KEY_AC_TEST_FILES));
			String[]testfiles = ca.getTestfiles();
			for(int i=0;i<testfiles.length-1;i++)
				result.append(testfiles[i]+" ");
			result.append(testfiles[testfiles.length-1]+"\n");
		}
		//output prediction files
		if(ca.getPredictionOutputFiles()!=null){
			result.append(getConfigString(KEY_AC_OUTPUT_PREDICTION_FILES));
			String[] predFiles = ca.getPredictionOutputFiles();
			for(int i=0;i<predFiles.length-1;i++)
				result.append(predFiles[i]+" ");
			result.append(predFiles[predFiles.length-1]+"\n");
		}
		//file type
		result.append(getConfigString(KEY_AC_FILE_TYPE,ca.getTestfiles_filetype().toString()));
		//strand
		result.append(getConfigString(KEY_AC_STRAND,ca.getStrand().toString()));
		//crossvalidation
		if(ca.getCrossvalidation_fold()!=-1)
			result.append(getConfigString(KEY_AC_CROSSVALIDATION,ca.getCrossvalidation_fold()+""));
		//complexity crossvalidation
		if(ca.isComplexityCrossvalidation())
			result.append(getConfigString(KEY_AC_COMPlEX_CROSSVALIDATION,"TRUE"));
		//maximum crossvalidation
		if(ca.getMaximum_crossvalidation()!=-1)
			result.append(getConfigString(KEY_AC_MAX_CROSSVALIDATION,ca.getMaximum_crossvalidation()+""));
		//positive training amount
		if(ca.getPositive_training_amount()!=-1)
			result.append(getConfigString(KEY_AC_POSITIVE_TRAINING_AMOUNT,ca.getPositive_training_amount()+""));
		//negative training amount
		if(ca.getNegative_training_amount()!=-1)
			result.append(getConfigString(KEY_AC_NEGATIVE_TRAINING_AMOUNT,ca.getNegative_training_amount()+""));
		//classification pattern
		if(ca.getPattern()!=null)
			result.append(getConfigString(KEY_AC_CLASSIFICATION_PATTERN,ca.getPattern()));
		//classification pattern location
		if(ca.getPattern_classification_location()!=0)//is different, is added to location
			result.append(getConfigString(KEY_AC_CLASSIFICATION_PATTERN_LOC,ca.getPattern_classification_location()+""));
		//classification pattern output
		if(ca.getPattern_classification_output_f()!=0&& ca.getPattern_classification_output_r()!=0) //see above
			result.append(getConfigString(KEY_AC_CLASSIFICATION_PATTERN_OUT,ca.getPattern_classification_output_f()+" "+ca.getPattern_classification_output_r()));
		//feature selection type -->all fs is dependend on this
		if(ca.getFeatureSelection().getFeature_selection_type()!=FEATURE_SELECTION_TYPE.NO_FEATURE_SELECTION || ca.getFeatureSelection().getFeature_selection_purpose()==FEATURE_SELECTION_PURPOSE.LOAD_FILTER){
			if(ca.getFeatureSelection().getFeature_selection_type()!=FEATURE_SELECTION_TYPE.NO_FEATURE_SELECTION)
				result.append(getConfigString(KEY_AC_FEATURE_SELECTION_TYPE,ca.getFeatureSelection().getFeature_selection_type().toString()));
			result.append(getConfigString(KEY_AC_FEATURE_SELECTION_PURPOSE,ca.getFeatureSelection().getFeature_selection_purpose().toString()));
			if(ca.getFeatureSelection().getFeature_selection_input_file()!=null)
				result.append(getConfigString(KEY_AC_FEATURE_SELECTION_INPUT,ca.getFeatureSelection().getFeature_selection_input_file()));
			if(ca.getFeatureSelection().getFeature_selection_output_file()!=null)
				result.append(getConfigString(KEY_AC_FEATURE_SELECTION_OUTPUT,ca.getFeatureSelection().getFeature_selection_output_file()));
			if(ca.getFeatureSelection().getMaxNumFeatures()!=-1)
				result.append(getConfigString(KEY_AC_FEATURE_SELECTION_MAX,ca.getFeatureSelection().getMaxNumFeatures()+""));
		}
		
		result.append(KEY_END+"\n");
		
		return result.toString();
	}
	
	
	/**
	 * This method creates the configuration file for a single classification-action, so the classification
	 * can be started remotely. This method's primary use is aimed at creating configurationfiles
	 * for jobs that will be started on the cluster. Towards this end, a standard workflow
	 * object will be created, to which the provided classificationAction will be added.
	 * 
	 * Following up to the fact that only one classification-action is defined and will be used, several
	 * of the general_options will be disabled: 
	 *  a) <merge_prediction_output> is useless with less than two classification actions 
	 *  b) <optimizer> should definitily be disabled, since it 
	 * @param action The classification action for which a configuration file should be produced.
	 * @param fileName The name for the file to which the configuration data should be written.
	 * @return The file containing the configuration data.
	 * @throws IOException In case the production of the configurationfile goes wrong.
	 */
	public synchronized static File createConfigurationFile(ClassificationAction action,String fileName)throws IOException{
		
		File file = new File(fileName);
		BufferedWriter writer = new BufferedWriter(new FileWriter(file));
		Logger lo = Logger.getLogger("TemporaryLogger");
		FunSiPWorkflow fsw = new FunSiPWorkflow(lo,fileName);
		fsw.addClassificationAction(action);
		String temp = getConfigurationFileString(fsw);
		writer.write(temp);
		writer.write("\n");
		writer.close();
		return file;
	}
	
	
	
	/**
	 * This method parses a configuration file and sets all the appropiate values 
	 * to the parameters defined in the configuration file. The method tries to
	 * locate the configuration file in different ways, dependent on whether or not
	 * the program is started from a jar-archive or not.
	 * @param fileName The name of the configuration file.
	 * @return The workflow-object with all its options set, and with its
	 * defined classification actions.
	 */
	public FunSiPWorkflow parseFile(String fileName){
		//test first the fileName as it is given...
		FunSiPWorkflow work = parseFile_1(fileName);
		if(work!=null)
			return work;
		
		//if this does not work, try to change the fileName
		logger.debug("Cannot open file, trying to relocate path");
		if(fileName!=null && fileName.startsWith(".")){
			fileName = fileName.substring(1,fileName.length());
			work = parseFile_1(fileName);
		}
		return work;
	}
	
	/**
	 * This method parses the parameterfile and creates a workflow object.
	 * The method itself largely consists of a giant nested if-structure 
	 * to parse the different options.
	 * @param fileName The name of the parameterfile.
	 * @return The workflow object.
	 */
	private FunSiPWorkflow parseFile_1(String fileName){
		logger.debug("Parsing parameterfile...");
		
		//intializing the variables used for parsing
		FunSiPWorkflow workflow = new FunSiPWorkflow(logger,fileName);
		
		try{
		LOAD_OPTIONS load = LOAD_OPTIONS.NO_LOAD;
		ClassificationAction action = null;
		
		InputStream stream  = getClass().getResourceAsStream(fileName);
		BufferedReader reader = null;
		
		if(stream!=null){
			reader = new BufferedReader(new InputStreamReader(stream));
		}
		else{
			reader = new BufferedReader(new FileReader(fileName));
		}

		//now, read every line of the parameterfile
		String line = reader.readLine();
		for(;line!=null;line=reader.readLine()){
			if(line.equals("")||line.startsWith("#")) //remove commentary
				continue;	
			line = line.trim();
		
			//define the different top-level headers
			if(line.equals(KEY_GENERAL_OPTIONS)){
				if(load==LOAD_OPTIONS.NO_LOAD)
					load = LOAD_OPTIONS.LOAD_GENERAL;
				else
					return fatalErrorExit("Starting top-level bloc inside other top-level bloc!");
				continue;
			}
			else if(line.equals(KEY_ACTION)){
				if(load==LOAD_OPTIONS.NO_LOAD){
					load = LOAD_OPTIONS.LOAD_ACTION;
					action = new ClassificationAction(logger,workflow);
				}
				else
					return fatalErrorExit("Starting top-level bloc inside other top-level bloc!");
				continue;
			}
			else if(line.equals(KEY_END)){
				if(load==LOAD_OPTIONS.NO_LOAD){
					return fatalErrorExit("Ending non existing top-level bloc!");
				}
				else if(load==LOAD_OPTIONS.LOAD_ACTION){
					workflow.addClassificationAction(action);
				}
				else if(load==LOAD_OPTIONS.LOAD_GENERAL){
					
				}
				
				load = LOAD_OPTIONS.NO_LOAD;
				continue;
			}
			
			String[] split = line.split("\t+");
			String command = split[0];
			//now sweep over all the other lines
			
			//*******************************************************************/
			//********** OPTIONS OF THE GENERAL_OPTIONS BLOC ********************/
			if(load==LOAD_OPTIONS.LOAD_GENERAL){
				
				//multithreading
				if(command.equals(KEY_GO_MULTITHREADING)){
					if(split.length!=2)
						return fatalErrorExit("Wrong number of parameters for multithreading");
					boolean test = workflow.setThreading(split[1]);
					if(!test)
						return fatalErrorExit("Unknown multithreading option");
					else
						logger.debug("Multithreading set to "+split[1].toUpperCase());
				}
				
				//the purpose of the workflow: functional site/AS/ ... prediction
				else if(command.equals(KEY_GO_PURPOSE)){
					if(split.length!=2)
						return fatalErrorExit("Wrong number of parameters for PURPOSE");
					boolean test = workflow.setPurpose(split[1]);
					if(!test)
						return fatalErrorExit("Unknown purpose option");
					else
						logger.debug("Purpose set to "+split[1].toUpperCase());
				}
				
				//changing the logging style (debug,info,error,...)
				else if(command.equals(KEY_GO_LOGGING_STYLE)){
					if(split.length!=2)
						return fatalErrorExit("Wrong number of parameters for logging_style");
					boolean test = setLoggingStyle(split[1]);
					if(!test)
						return fatalErrorExit("Unknown logging option/problem while parsing logging_style options");
					else
						logger.debug("Logging style changed.");
				}
				
				//changing the logging output (file, console,...)
				else if(command.equals(KEY_GO_LOGGING_OUTPUT)){
					if(split.length<2)
						return fatalErrorExit("Wrong number of parameters for logging_output");
					boolean test = setLoggingOutput(split);
					if(!test)
						return fatalErrorExit("Unknown logging option/problem while parsing logging_output options");
					else
						logger.debug("Logging output changed.");
				}
				
				//indicates that some files should be merged at the end of the workflow
				else if(command.equals(KEY_GO_MERGE_PREDICTION)){
					if(split.length!=2)
						return fatalErrorExit("Wrong number of parameters for merge_prediction_output");
					workflow.setMergeResults(true);
					List<String> inputMergeFiles = new ArrayList<String>();
					String []split2 = split[1].split(" ");
					split2 = recoverSpaceDirectories(split2);
					for(int i=0;i<split2.length-1;i++)
						inputMergeFiles.add(split2[i]);
					workflow.addInputMergeFiles(inputMergeFiles);
					workflow.addOutputMergeFile(split2[split2.length-1]);
					logger.debug("Mergefile defined");
				}
				
				//indicates that the workflow should run in optimization mode
				else if(command.equals(KEY_GO_OPTIMIZER)){
					workflow.setOptimizer(true);
					String[] split2 = split[1].split(" ");
					workflow.setOptimizer_steps(Integer.parseInt(split2[0]),Integer.parseInt(split2[1]),Integer.parseInt(split2[2]));
					logger.debug("Optimization workflow selected");
				}
				
				//unrecognized option inside the GENERAL_OPTIONS bloc
				else{
					return fatalErrorExit("Illegal command inside the GENERAL_OPTIONS top-level bloc :\n" +
							"--> "+line);
				}
			}
			
			
			//**********************************************************/
			//********** OPTIONS OF THE ACTION BLOC ********************/
			else if(load==LOAD_OPTIONS.LOAD_ACTION){
				//giving a name to the workflow
				if(command.equals(KEY_AC_NAME)){
					if(split.length==2)
						action.setName(split[1]);
					else
						return fatalErrorExit("Incorrect number of parameters for ACTION / option = NAME");
				}
				
				//settting the classification name
				else if(command.equals(KEY_AC_CLASSIFICATION_NAME)){
					if(split.length==2)
						action.setClassification_name(split[1]);
					else
						return fatalErrorExit("Incorrect number of parameters for ACTION /option = CLASSIFICATION_NAME");
				}
				
				//setting the classifier type
				else if(command.equals(KEY_AC_CLASSIFIER_TYPE)){
					if(split.length==2){
						boolean test = action.setClassifier_type(split[1]);
						if(!test)
							return fatalErrorExit("Impossible to parse classifier_type : "+split[1]);
					}
					else
						return fatalErrorExit("Incorrect number of parameters for ACTION /option = CLASSIFIER_TYPE");
				}
				
				//setting the classifier options
				else if(command.equals(KEY_AC_CLASSIFIER_OPTIONS)){
					if(split.length==2)
						action.setClassifier_options(split[1].split(" "));
					else
						return fatalErrorExit("Incorrect number of parameters for ACTION /option = CLASSIFIER_OPTIONS");
				}
				
				//adding a classification feature
				else if(command.equals(KEY_AC_CLASSIFICATION_FEATURE)){
					if(split.length==2){
						ClassificationFeature cf = ClassificationFeatureUtils.parseFeatureFromString(split[1]);
						if(cf!=null){
							action.addClassificationFeature(cf);
							Conversion conversion = cf.getConversion();
							if(conversion!=null && !action.getConversionList().contains(conversion))
								action.addConversion(conversion);
						}
						else
							return fatalErrorExit("Impossible to parse classification feature : "+split[1]);
					}
					else
						return fatalErrorExit("Incorrect number of parameters for ACTION /option = CLASSIFICATION_FEATURE");
				}
				
				//adding positive training file
				else if(command.equals(KEY_AC_POSITIVE_TRAINING_FILE)){
					if(split.length==2)
						action.setPositive_training_file(split[1]);
					else
						return fatalErrorExit("Incorrect number of parameters for ACTION /option = POSITIVE_TRAINING_FILE");
				}
				
				//adding negative training file
				else if(command.equals(KEY_AC_NEGATIVE_TRAINING_FILE)){
					if(split.length==2)
						action.setNegative_training_file(split[1]);
					else
						return fatalErrorExit("Incorrect number of parameters for ACTION /option = NEGATIVE_TRAINING_FILE");
				}
				
				//adding positive secondary structure training file
				else if(command.equals(KEY_AC_POSITIVE_TRAINING_FILE_SS)){
					if(split.length==2)
						action.setPositive_training_file_secstruct(split[1]);
					else
						return fatalErrorExit("Incorrect number of parameters for ACTION /option = POSITIVE_TRAINING_FILE_SECSTRUCT");
				}
				
				//adding negative secondary structure training file
				else if(command.equals(KEY_AC_NEGATIVE_TRAINING_FILE_SS)){
					if(split.length==2)
						action.setNegative_training_file_secstruct(split[1]);
					else
						return fatalErrorExit("Incorrect number of parameters for ACTION /option = NEGATIVE_TRAINING_FILE_SECSTRUCT");
				}
				
				//setting outputdirectory
				else if(command.equals(KEY_AC_OUTPUT)){
					if(split.length==2)
						action.setOutputdirectory(split[1]);
					else
						return fatalErrorExit("Incorrect number of parameters for ACTION /option = OUTPUT_DIRECTORY");
				}
				
				//setting the splicesiteposition
				else if(command.equals(KEY_AC_FUNCSITE_POSITION)){
					if(split.length==2)
						action.setTraining_functional_site_position(Integer.parseInt(split[1]));
					else
						return fatalErrorExit("Incorrect number of parameters for ACTION /option = SPLICESITE_POSITION");
				}
				
				//setting the CommandOption
				else if(command.equals(KEY_AC_COMMAND)){
					if(split.length==2){
						boolean test =  action.setCommand(split[1]);
						if(!test)
							return fatalErrorExit("Unknown parameter for option COMMAND : "+split[1]);
					}
					else
						return fatalErrorExit("Incorrect number of parameters for ACTION /option = COMMAND");
				}
				
				//setting the classification output
				else if(command.equals(KEY_AC_MODEL_FILE)){
					if(split.length==2)
						action.setModel_file(split[1]);
					else
						return fatalErrorExit("Incorrect number of parameters for ACTION /option = MODEL_FILE");
				}
				
				//setting the features output file
				else if(command.equals(KEY_AC_FEATURES_OUTPUT)){
					if(split.length==2)
						action.setFeatures_output_file(split[1]);
					else
						return fatalErrorExit("Incorrect number of parameters for ACTION /option = FEATURES_OUTPUT_FILE");
				}
				
				//settting the testfiles
				else if(command.equals(KEY_AC_TEST_FILES)){
					if(split.length!=1){
						String[] temp = new String[split.length-1];
						System.arraycopy(split,1, temp, 0, split.length-1);
						//accomodate for possible wildcards
						temp = extractFileNames(temp);
						action.setTestfiles(temp);
					}
					else
						return fatalErrorExit("Incorrect number of parameters for ACTION /option = TEST_FILES");
				}
				
				//setting the names of the outputfiles containing the predictions
				else if(command.equals(KEY_AC_OUTPUT_PREDICTION_FILES)){
					if(split.length!=1){
						String[] temp = new String[split.length-1];
						System.arraycopy(split,1,temp,0,split.length-1);
						action.setPredictionOutputFiles(temp);
					}
					else
						return fatalErrorExit("Incorrect number of parameters for ACTION /option = OUTPUT_PREDICTION_FILES");
					
				}
				
				//setting the filetype of the testfiles
				else if(command.equals(KEY_AC_FILE_TYPE)){
					if(split.length==2){
						boolean test = action.setTestfiles_filetype(split[1]);
						if(!test)
							return fatalErrorExit("Unknown parameter for option FILE_TYPE : "+split[1]);
					}
					else
						return fatalErrorExit("Incorrect number of parameters for ACTION /option = FILE_TYPE");
				}
				
				//setting the strand
				else if(command.equals(KEY_AC_STRAND)){
					if(split.length==2){
						boolean test = action.setStrand(split[1]);
						if(!test)
							return fatalErrorExit("Unknown parameter for option STRAND : "+split[1]);
					}
					else
						return fatalErrorExit("Incorrect number of parameters for ACTION /option= STRAND");
				}
				
				//setting the crossvalidation option
				else if(command.equals(KEY_AC_CROSSVALIDATION)){
					if(split.length==2)
						action.setCrossvalidation_fold(Integer.parseInt(split[1]));
					else
						return fatalErrorExit("Incorrect number of parameters for ACTION /option = CROSSVALIDATION");
				}
				
				//setting the complexity crossvalidation option
				else if(command.equals(KEY_AC_COMPlEX_CROSSVALIDATION)){
					if(split.length==2){
						if(split[1].toUpperCase().equals("TRUE"))
							action.setComplexityCrossvalidation(true);
						else if(split[1].toUpperCase().equals("FALSE"))
							action.setComplexityCrossvalidation(false);
						else
							return fatalErrorExit("Unknown parameter for option COMPLEXITY_CROSSVALIDATION (true/false) : "+split[1]);
					}
					else
						return fatalErrorExit("Incorrect number of parameters for ACTION /option = COMPLEXITY_CROSSVALIDATION");
				}
				
				//setting the crossvalidation_fixed option
				else if(command.equals(KEY_AC_MAX_CROSSVALIDATION)){
					if(split.length==2)
						action.setMaximum_crossvalidation(Integer.parseInt(split[1]));
					else
						return fatalErrorExit("Incorrect number of parameters for ACTION /option = MAXIMUM_CROSSVALIDATION");
				}
				
				//settting the amount of positive training data
				else if(command.equals(KEY_AC_POSITIVE_TRAINING_AMOUNT)){
					if(split.length==2)
						action.setPositive_training_amount(Integer.parseInt(split[1]));
					else
						return fatalErrorExit("Incorrect number of parameters for ACTION /option = POSITIVE_TRAINING_AMOUNT");
				}
				
				//settting the amount of negative training data
				else if(command.equals(KEY_AC_NEGATIVE_TRAINING_AMOUNT)){
					if(split.length==2)
						action.setNegative_training_amount(Integer.parseInt(split[1]));
					else
						return fatalErrorExit("Incorrect number of parameters for ACTION /option = NEGATIVE_TRAINING_AMOUNT");
				}
				
				//setting the search pattern
				else if(command.equals(KEY_AC_CLASSIFICATION_PATTERN)){
					if(split.length==2)
						action.setPattern(split[1].toUpperCase());
					else
						return fatalErrorExit("Incorrect number of parameters for ACTION /option = CLASSIFICATION_PATTERN");
				}
				
				else if(command.equals(KEY_AC_CLASSIFICATION_PATTERN_LOC)){
					if(split.length==2){
						String[] split2 = split[1].split(" ");
						action.setPattern_classification_location(Integer.parseInt(split2[0]));
					}
					else
						return fatalErrorExit("Incorrect number of parameters for ACTION /option = CLASSIFICATION_PATTERN_LOCATION");
				}
				
				//setting the location of the classification pointer, according to the search pattern
				else if(command.equals(KEY_AC_CLASSIFICATION_PATTERN_OUT)){
					if(split.length==2){
						String[] split2 = split[1].split(" ");
						action.setPattern_classification_output_f(Integer.parseInt(split2[0]));
						action.setPattern_classification_output_r(Integer.parseInt(split2[1]));
					}
					else
						return fatalErrorExit("Incorrect number of parameters for ACTION /option = CLASSIFICATION_PATTERN_OUTPUT");
				}
				
				//setting the feature selection type option
				else if(command.equals(KEY_AC_FEATURE_SELECTION_TYPE)){
					if(split.length==2){
						boolean test = action.getFeatureSelection().setFeature_selection_type(split[1]);
						if(!test)
							return fatalErrorExit("Unknown parameter for option FEATURE_SELECTION_TYPE : "+split[1]);
					}
					else
						return fatalErrorExit("Incorrect number of parameters for ACTION /option = FEATURE_SELECTION_TYPE");
						
				}
				
				//setting the feature selection task option
				else if(command.equals(KEY_AC_FEATURE_SELECTION_PURPOSE)){
					if(split.length==2){
						boolean test = action.getFeatureSelection().setFeature_selection_purpose(split[1]);
						if(!test)
							return fatalErrorExit("Unknown parameter for option FEATURE_SELECTION_PURPOSE : "+split[1]);
					}
					else
						return fatalErrorExit("Incorrect number of parameters for ACTION /option = FEATURE_SELECTION_PURPOSE");
				}
				
				//setting the feature input file
				else if(command.equals(KEY_AC_FEATURE_SELECTION_INPUT)){
					if(split.length==2)
						action.getFeatureSelection().setFeature_selection_input_file(split[1]);
					else
						return fatalErrorExit("Incorrect number of parameters for ACTION /option = FEATURE_SELECTION_INPUT_FILE");
				}
				
				//setting the feature output file
				else if(command.equals(KEY_AC_FEATURE_SELECTION_OUTPUT)){
					if(split.length==2)
						action.getFeatureSelection().setFeature_selection_output_file(split[1]);
					else
						return fatalErrorExit("Incorrect number of parameters for ACTION /option = FEATURE_SELECTION_OUTPUT_FILE");
				}
				
				//setting the feature selection maximum feature number
				else if(command.equals(KEY_AC_FEATURE_SELECTION_MAX)){
					if(split.length==2)
						action.getFeatureSelection().setMaxNumFeatures(Integer.parseInt(split[1]));
					else
						return fatalErrorExit("Incorrect number of parameters for ACTION /option = FEATURE_SELECTION_MAX_FEATURES");
				}
				
				//setting fixed values for the svm_distance->output probability
				/*else if(command.equals("OUTPUT_FIXED_ALPHA")){
					if(split.length==2)
						 action.setOutput_fixed_alpha(Double.parseDouble(split[1]));
					else
						return fatalErrorExit("Incorrect number of parameters for ACTION /option = OUTPUT_FIXED_ALPHA");
					
				}
				
				//settting fixed values for the svm_distance -> output probability
				else if(command.equals("OUTPUT_FIXED_BETA")){
					if(split.length==2)
						action.setOutput_fixed_beta(Double.parseDouble(split[1]));
					else
						return fatalErrorExit("Incorrect number of parameters for ACTION /option = OUTPUT_FIXED_BETA");
				}
				*/
				//unknown option in top-level bloc "ACTION"
				else{
					return fatalErrorExit("Unrecognized option for action : "+split[0]);
				}
			}
			else{
				return fatalErrorExit("Illegal top-level indicator or indicator outside top-level block.\n" +
						"--> "+line+"\n\tPlease make sure that:\n" +
								"\t\ta)A configurationfile is supplied.\n"+
								"\t\tb)The configurationfile is properly constructed.");
			}
			
		}
		reader.close();
		}
		catch(NumberFormatException exc){
			return fatalErrorExit("One of the options should be a number, but could not be converted from a string :\n"+exc);
		}
		catch(IndexOutOfBoundsException exc){
			return fatalErrorExit("One of the options contains too few arguments");
		}
		catch(Exception exc){
			//exc.printStackTrace();
			return fatalErrorExit("Generic Error while parsing parameter file:\n" +exc);
		}
		
		
		return workflow;
	}
	
	
	/**
	 * Method which displays the given string and then exists the program,
	 * due to the fact that one of the parameters could not be parsed.
	 * Because it is still unclear which options are necessary and which
	 * are optional, we opted to halt the program on every parsing error,
	 * to prevent faulty or unclear results.
	 * @param message The message to be displayed.
	 */
	public FunSiPWorkflow fatalErrorExit(String message){
		if(logger!=null){
			logger.error("Error occured while parsing the parameter file:");
			logger.error(message);
			logger.fatal("Unable to continue, ending program.");
		}
		else{
			System.out.println(message);
		}
		return null;
	}
		
	
	/*
	 * Because of the fact that we described in the API for the configuration
	 * files that the entities in the merge_file indicator should be space seperated,
	 * we ran into trouble with the fact that some directories may of course contain spaces :(
	 * 
	 * The idea is then to surround those indicators by "-signs and then refilter and
	 * rearrange everything into the correct state.
	 */
	private String[]recoverSpaceDirectories(String[] t){
		if(t==null)
			return null;
		List<String> result = new ArrayList<String>();
		boolean openAcc = false;
		String openedDir = "";
		for(int i=0;i<t.length;i++){
			String s = t[i];
			if(!openAcc){
				if(!s.startsWith("\""))
					result.add(s);
				else{
					openAcc = true;
					openedDir+=s;
				}
			}
			else{
				if(!s.endsWith("\""))
					openedDir+=" "+s;
				else{
					openAcc = false;
					openedDir+=" "+s;
					String temp = openedDir;
					result.add(temp);
					openedDir = "";
				}
			}
			
		}
		
		String[]ret = new String[result.size()];
		for(int i=0;i<result.size();i++)
			ret[i] = result.get(i);
		
		return ret;
	}
	
	/*
	 * This method attempts to change the output of the loggingoutput,
	 * by redirecting it either to a file (with a third argument containing
	 * the file name) or to the console.
	 */
	private boolean setLoggingOutput(String[] logging_output){
		String style = logging_output[1];
		style = style.toUpperCase();
		if(style.equals("CONSOLE")){
			if(logging_output.length!=2)
				return false;
			ConsoleAppender appender = new ConsoleAppender(new SimpleLayout());
			logger.removeAllAppenders();
			logger.addAppender(appender);
			return true;
		}
		else if(style.equals("FILE")){
			if(logging_output.length!=3)
				return false;
			FileAppender appender = null;
			try{
				 appender = new FileAppender(new SimpleLayout(),logging_output[2]);
			}
			catch(IOException exc){
				return false;
			}
			logger.removeAllAppenders();
			logger.addAppender(appender);
			return true;
		}
		else
			return false;
	}
	
	/*
	 * Private method which attempts to change the logging style. 
	 * Because the change is supplied by a string, a boolean value
	 * is returned in case the string cannot be parsed.
	 */
	private boolean setLoggingStyle(String logging_style){
		String logging_styl = logging_style;
		logging_styl = logging_styl.toUpperCase();
		if(logging_styl.equals("DEBUG")){
			logger.setLevel((Level)Level.DEBUG);
			return true;
		}
		else if(logging_styl.equals("INFO")){
			logger.setLevel((Level)Level.INFO);
			return true;
		}
		else if(logging_styl.equals("ERROR")){
			logger.setLevel((Level)Level.ERROR);
			return true;
		}
		else if(logging_styl.equals("FATAL")){
			logger.setLevel((Level)Level.FATAL);
			return true;
		}
		else if(logging_styl.equals("WARN")){
			logger.setLevel((Level)Level.WARN);
			return true;
		}
		else
			return false;
	}
	
	/*
	 * This method tries to extract all the files from an array of filenames. (This method
	 * is similar to that found in FunSiP.java, for parsing the command line options). 
	 * However, since there is no underlying operating system, the *-wildcard character
	 * for defining multiple files at once does not work direclty. Therefore we have
	 * to do this manually (*sigh*).
	 * 
	 */
	private String[] extractFileNames(String[] args){
		String [] res = null;
		List<File> files = new ArrayList<File>();
		
		for(int i=0;i<args.length;i++){
			String temp = args[i];
			File file = new File(temp);
			String name = file.getName();
			if(name.startsWith("*")){
				String extension = name.substring(name.lastIndexOf("."),name.length());
				File dir = file.getParentFile();
				if(dir==null) 
					break;
				File [] tempfiles = dir.listFiles();
				for(int j=0;j<tempfiles.length;j++){
					if(tempfiles[j].getName().endsWith(extension))
						files.add(tempfiles[j]);
				}
			}
			else{
				files.add(file);
			}
		}
		
		
		res = new String[files.size()];
		for(int i=0;i<files.size();i++){
			res[i] = files.get(i).getPath();
		}
		return res;
	}
}
