/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
package base.cluster;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileOutputStream;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.OutputStream;
import java.io.PrintWriter;

import org.apache.log4j.Logger;

import base.ClassificationAction;
import base.FunSiPWorkflow;
import base.ParameterFileParser;




/**
 * This class is a special wrapup around a normal classificationaction.
 * It is designed to execute a normal classificationAction on a GridEngine cluster node,
 * while the parent FunSiPWorkflow object is residing on the master node. 
 * The upside of this implementation is the fact that this way
 * the cluster can be utilized, without having to resort to techniques such
 * as java RMI. The downside however is the fact that we have to read and write 
 * console-commands to standard output in order to check whether or not the
 * classificationActions running on the cluster nodes have finished or not.
 * 
 * This class should in this stage only be used for speeding up the optimalization
 * procedure, which is necessary before building a classification model or using said
 * classification model to evaluate sequence files.
 * @author Michiel Van Bel
 */
public class ClusterClassificationAction implements Runnable{

	/**
	 * The time a node waits between anouncing its finish (milliseconds).
	 */
	public static final long SLEEP_TIME = 20000;
	
	private ClassificationAction action = null;
	private FunSiPWorkflow parentworkflow = null;
	private String index = "";
	private Logger logger = null;
	
	//the six references to the used files
	//I'll sleep and dream dreams of a nuclear apocalypse.
	private File executionScriptFile	 	= null;
	private File configurationFile		 	= null;
	private File normalClusterOutputFile	= null;
	private File errorClusterOutputFile		= null;
	private File normalClusterJobOutputFile	= null;
	private File errorClusterJobOutputFile  = null;
	
	
	/**
	 * Constructor. Initialization of necessary class variables.
	 * Nonne of the passed arguments should be null.
	 * @param ca The classification action which needs to be executed on a cluster node. 
	 * @param workflow The parent workflow.
	 * @param index The index of the classification action in the parent workflow.
	 * @param logger The logging facility.
	 */
	public ClusterClassificationAction (ClassificationAction ca,FunSiPWorkflow workflow, String index, Logger logger){
		this.action 	= ca;
		this.parentworkflow 	= workflow;
		this.index 		= index;
		this.logger 	= logger;
	}
	
	
	/**
	 * Delete all the temporary files that were used 
	 * for executing the classificationaction on a cluster node.
	 */
	private void deleteFiles(){
		logger.debug("Deleting temporary files used for cluster interaction");
		try{
			if(executionScriptFile!=null)
				executionScriptFile.delete();
			if(configurationFile!=null)
				configurationFile.delete();
			if(normalClusterOutputFile!=null)
				normalClusterOutputFile.delete();
			if(errorClusterOutputFile!=null)
				errorClusterOutputFile.delete();
			if(normalClusterJobOutputFile!=null)
				normalClusterJobOutputFile.delete();
			if(errorClusterJobOutputFile!=null)
				errorClusterJobOutputFile.delete();
			
			if(executionScriptFile!=null)
				executionScriptFile.deleteOnExit();
			if(configurationFile!=null)
				configurationFile.deleteOnExit();
			if(normalClusterOutputFile!=null)
				normalClusterOutputFile.deleteOnExit();
			if(errorClusterOutputFile!=null)
				errorClusterOutputFile.deleteOnExit();
			if(normalClusterJobOutputFile!=null)
				normalClusterJobOutputFile.deleteOnExit();
			if(errorClusterJobOutputFile!=null)
				errorClusterJobOutputFile.deleteOnExit();
		}
		catch(Exception exc){
			logger.error(exc);
		}
	}
	
	
	/**
	 * This method creates the shell script that will be used to start FunSiP
	 * with the configuration file that is especially created for executing this single
	 * classification action. Warning: this is not the file that will be used to
	 * start the job on the cluster. 
	 * 
	 * @param scriptFileName The name of that the shell script file should have.
	 * @param configFileName The name of the configuration file that should be 
	 * launched by SPR.
	 * @return The created shell script file.
	 */
	private File createExecutionScript(String scriptFileName,String configFileName){
		File file = new File(scriptFileName);
		try{
			BufferedWriter writer = new BufferedWriter(new FileWriter(file));
			writer.write("java -Xbootclasspath/p:. -Xbootclasspath/a:./log4j-1.2.14.jar:./weka.jar -Xms512M -Xmx1500M main.FunSiP "+configFileName+"\n");
			writer.close();
		}
		catch(Exception exc){
			logger.error(exc);
		}
		return file;
	}
	
	
	/**
	 * Implementation of the run()-method of the Runnable interface. Necessary
	 * to start this action as a seperate thread.
	 * 
	 * This method will trie to create a job on the cluster, after which it 
	 * will wait until the job is finished. Then, it will notify the 
	 * parent workflow.
	 */
	public void run(){
		
		//indicate the names of the temporary files that will be used by 
		//the program.
		String scriptFileName 		= "./cluster_job_"+index+".sh";
		String configFileName		= "./cluster_job_"+index+"_config.txt";
		String clusterOutFileName	= "./cluster_job_"+index+"_cluster_out.txt";
		String clusterErrFileName	= "./cluster_job_"+index+"_cluster_err.txt";
		String [] command = {"qsub",scriptFileName};
		
		try{
			//create the execution-script (with the appropiate directory from the classificationaction,
			//place the script in the general directory, but keep file-link
			executionScriptFile = createExecutionScript(scriptFileName,configFileName);
		
			//create the configuration-file by invoking method of classificationaction,
			//provide the name and location of the configuration-file to be placed (normally,
			//also put it in top-level directory)
			configurationFile =  ParameterFileParser.createConfigurationFile(action,configFileName);
			normalClusterOutputFile = new File(clusterOutFileName);
			errorClusterOutputFile	= new File(clusterErrFileName);
			logger.debug("Scriptfile and configurationfile for cca_"+index+" created.");
			
			//launch the classification-action, by invoking the local environment,
			//store the number that was given by qsub
			FileOutputStream outStream = new FileOutputStream(normalClusterOutputFile);
			FileOutputStream errStream = new FileOutputStream(errorClusterOutputFile);
			
			//execute the qsub command on the cluster.
			//Also retrieve the associated input stream and output stream.
			Runtime rt = Runtime.getRuntime();
			Process proc = rt.exec(command);
			InputStream out = proc.getInputStream();
			InputStream err = proc.getErrorStream();
			
			//start different processreader threads that take care of the output generated by
			//the qsub command. Then it will block this clusterclassificationaction
			//thread while waiting for the processreader to signal whether or not
			//the job on the cluster has ended.
			ProcessReader outReader = new ProcessReader(out,outStream);
			ProcessReader errReader = new ProcessReader(err,errStream);
			outReader.start();
			errReader.start();
			proc.waitFor();
			
			logger.debug("Job submitted for index :"+index);
			
			//if submit_number < 0
			//We get a number from the clustersoftware when submitting a job. This
			//number is highly important to check whether or not a job is finished.
			int submit_number = getSubmitNumber(normalClusterOutputFile,errorClusterOutputFile);
			if(submit_number==-1)
				throw new Exception("Unable to parse job-number from data : unknown reason");
			else if(submit_number==-2)
				throw new Exception("Unable to parse job-number from data : error-file is not empty");
			else if(submit_number==-3)
				throw new Exception("Unable to parse job-number from data : out-file is not parsable");
			else{
				normalClusterJobOutputFile = new File(scriptFileName+".o"+submit_number);
				errorClusterJobOutputFile = new File(scriptFileName+".e"+submit_number);
			}
		
			logger.debug("job-number extracted : "+submit_number);
			
			//every 20 seconds (SLEEP_TIME), launch qstat -u $USER, filter the first two 
			//lines(rubbish), then split it (I think it's tabs), take the first number and 
			//put it in list of current active or to-do numbers
			boolean job_done = false;
			long elapsed_time = 0;
			while(!job_done){
				Thread.sleep(SLEEP_TIME); //consumes no processor cycles
				job_done = checkJobEnded(submit_number);
				elapsed_time+=SLEEP_TIME;
				if(!job_done){
					logger.debug("Job "+submit_number+" not ended yet. Running time : "+elapsed_time(elapsed_time));
				}
			}
			
			logger.debug("job "+submit_number+" is done.");
		
			//delete the execution script, configuration file, error file and output file
			//then return this method to caller.
			//The deletion is done by invoking delete() and deleteOnExit()
			
			deleteFiles();
			
			//returning to the calling method is however, no trivial task, due to the fact that we are
			//working with different threads. Therefore, ending this thread should be done by invoking
			//the waitForSync-method from the FunSiPWorkflow. In the end this will result in all threads
			//being terminated.
			if(this.parentworkflow!=null)
				this.parentworkflow.waitForSync();
			
		}
		catch(Exception exc){
			logger.error(exc);
			exc.printStackTrace();
		}	
		finally{
			//at last, something that should even be done while an error is thrown,
			//call waitFor Sync
			try{
				if(this.parentworkflow!=null)
					this.parentworkflow.waitForSync();
			}
			catch(Exception exc){
				logger.error(exc);
			}
		}
		
		logger.debug("Ending workflow for cluster classification action "+index);
	}
	
	
	/**
	 * This method creates a string indicating how long (in human readable format:
	 * hours/minutes/seconds) an indicated number of milliseconds takes. 
	 * @param milliseconds The number of milliseconds that should be converted to a 
	 * human readable format.
	 * @return The string in human readable format.
	 */
	private String elapsed_time(long milliseconds){
		long seconds = milliseconds/1000;
		StringBuffer result = new StringBuffer();
		if(seconds/3600!=0){
			long hours = seconds/3600;
			seconds = seconds%3600;
			if(hours==1)
				result.append(hours+" hour, ");
			else
				result.append(hours+" hours, ");
		}
		if(seconds/60!=0){
			long minutes = seconds/60;
			seconds = seconds%60;
			if(minutes==1)
				result.append(minutes+" minute, ");
			else
				result.append(minutes+" minutes, ");
		}
		if(seconds==1)
			result.append(seconds+" second");
		else
			result.append(seconds+" seconds");
		return result.toString();
	}
	
	
	/**
	 * 
	 * Some really ugly method to check whether the job is still running on the cluster
	 * lumping together normal output and error output
	 *  
	 * normal working with the processReader doesn't seem to work... 
	 * solution: immediate redirect to file
	 *
	 * @param jobNumber The number of the job. This number is supplied by the 
	 * gridengine software.
	 * @return A boolean value, indicating whether a job has ended or not. No distinction
	 * is made between a job that has ended in the normal way, or a job that has 
	 * crashed. The only thing checked is whether or not a job is still alive on the 
	 * cluster.
	 * @throws Exception Checking whether a job is still alive on the cluster requires
	 * some heavy fileparsing and other odd ends. These might of course cause exeptions,
	 * which or not caught but returned to the calling method.
	 */
	private boolean checkJobEnded(int jobNumber) throws Exception{
		boolean result = false;
		String [] command = {"qstat","-j",""+jobNumber};
		File checkJobFileOut = new File("./check_job_"+jobNumber+"_out.txt");
		File checkJobFileErr = new File("./check_job_"+jobNumber+"err.txt");
		
		/*
		 * Two streams, that link to the error output and normal output files,
		 * generated by execute command.
		 */
		FileOutputStream out = new FileOutputStream(checkJobFileOut);
		FileOutputStream err = new FileOutputStream(checkJobFileErr);
		Runtime rt = Runtime.getRuntime();
		Process proc = rt.exec(command);
		
		/*
		 * Read the error and outputstreams on the fly.
		 */
		InputStream out_stream = proc.getInputStream();
		InputStream err_stream = proc.getErrorStream();
		ProcessReader out_reader = new ProcessReader(out_stream,out);
		ProcessReader err_reader = new ProcessReader(err_stream,err);
		out_reader.start();
		err_reader.start();
		proc.waitFor();
		
		/*
		 * if there is only one line available in both the error and the 
		 * output (its something like: job 1790289 does not exist), then
		 * the job is no more alive on the cluster. If there are more lines,
		 * the job is definitly still alive on the cluster.
		 */ 
		int out_lines = getNumLines(checkJobFileOut);
		int err_lines = getNumLines(checkJobFileErr);
		int total = out_lines + err_lines;
		if(total==1)
			result = true;
		else{
			result = false;
		}
		return result;
	}
	
	
	
	/**
	 * This method simply tries to return the number of lines in a 
	 * file. Various error checks are performed, in order not to
	 * return impossible values.
	 * @param f The file f, from which the number of lines should be extracted.
	 * @return The number of lines in the file.
	 */
	public int getNumLines(File f){
		int result = 0;
		if(!f.exists())
			return result;
		else{
			try{
				BufferedReader bReader = new BufferedReader(new FileReader(f));
				String line = bReader.readLine();
				while(line!=null){
					result++;
					line = bReader.readLine();
				}
				bReader.close();
				f.delete();
			}
			catch(Exception exc){
				logger.error("Error while determining #lines in file");
				logger.error(exc);
			}
			return result;
		}
	}
	
	
	/**
	 * This method parses the submit-number from the outputfile.
	 * Some serious error checking is going on inside. Normally the
	 * returned number is the number given by the cluster software to this job.
	 * However, the values -1, -2, -3 are also possible:
	 * a)-1 : General exception, impossible to obtain job number.
	 * b)-2 : The error file is not empty 
	 * c)-3 : If an error occured while reading the output en error files.
	 * @param outputFile The file containing standard output from the qsub-command
	 * @param errorFile The file containing error output from the qsub-command
	 * @return The number assigned to this job by the cluster software.
	 */
	public int getSubmitNumber(File outputFile,File errorFile){
		int result = -1;
		BufferedReader outReader = null;
		BufferedReader errReader = null;
		try{
			outReader = new BufferedReader(new FileReader(outputFile));
			errReader = new BufferedReader(new FileReader(errorFile));
			
			boolean error_encounterd = false;
			String line = errReader.readLine();
			while(line!=null){
				line = errReader.readLine();
				if(!line.equals(""))
					error_encounterd=true;
			}
			
			if(error_encounterd)
				return -2;
			
			line = outReader.readLine(); //if no error was encounterd, this line should have result
			if(line!=null){
				String[] split = line.split(" ");
				//third string has number
				result = Integer.parseInt(split[2]);
				return result;
			}
			
		}
		catch(Exception exc){
			logger.error(exc);
			return -3;
		}
		finally{
			try{
				if(outReader!=null)
					outReader.close();
				if(errReader!=null)
					errReader.close();
			}
			catch(Exception exc){
				logger.error(exc);
			}
		}
		return result;
	}
	
	
	
	/**
	 * This private class functions simply as a stub, because there has to 
	 * be a way for the native process to tunnel its output to. If this 
	 * is not done properly, the native process stops and blocks the entire
	 * workflow.  
	 * @author Michiel Van Bel
	 */
	private class ProcessReader extends Thread{
		
		private InputStream in = null;
		private OutputStream os = null;
		
		/**
		 * Constructor
		 * @param in A stream coming from the native process itself. 
		 * @param toFile A stream indicating where to data from the inputstream
		 * should be written to. This stream is linked to a filestream.
		 */
		public ProcessReader(InputStream in,OutputStream toFile){
			this.in = in;
			this.os = toFile;
		}
				
		public void run(){
			try{
				PrintWriter pw = new PrintWriter(os);
				BufferedReader reader = new BufferedReader(new InputStreamReader(in));
				String line = null;
				while((line = reader.readLine()) != null){
					pw.println(line);
				}
				pw.flush();
				pw.close();
			}
			catch(Exception exc){
			}
		}
	}	
}
