/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
package classifier;

import java.io.File;
import java.util.List;

import util.featureSelection.ValPosCombination;
import util.results.CrossValidationResult;
import weka.attributeSelection.AttributeEvaluator;
import weka.core.Instances;

/**
 * This interface defines all the necessary methods that a classifier should
 * conform to, in order to be compatible with FunSiP.
 * @author Michiel Van Bel
 */

public interface Classifier {

	/**
	 * The possible types of data: POSITIVE_DATA indicates that the source is 
	 * positive training data. NEGATIVE_DATA indicates that the source is 
	 * negative training data.  UNCLASSIFIED_DATA indicates that the source
	 * is a DNA-sequence, that is to be evaluated.
	 */
	public enum DATA_TYPE {POSITIVE_DATA, NEGATIVE_DATA, UNCLASSIFIED_DATA};

	
	
	/**
	 * Performs a crossvalidation of trainingfile. The results of this crossvalidation
	 * (number of true positives,false positives,true negatives,false negatives and 
	 * deduced numbers) are then returned.
	 * The normal procedure for crossvalidation can be followed. However, we extended
	 * the notion of it so different numbers of positive/negative examples can be used
	 * during the training phase of each crossvalidation step. The number of positive/
	 * negative examples for the testing phase during each crossvalidation step remains
	 * the same (begin (n-1)*(total_amount)). For further information, see the
	 * Crossvalidation.txt document in the /doc subdirectory.
	 * @param n The fold of the crossvalidation. Frequent numbers are 2,5 and 10
	 * @param maxPosTrain The maximum amount of positive training examples during the 
	 * training phase of the crossvalidation.
	 * @param maxNegTrain The maximum amount of negative training examples during the
	 * training phase of the crossvalidation.
	 * @return The crossvalidationresult object, which contains the results of
	 * the crossvalidation, and the deduced statistics.
	 */
	public CrossValidationResult crossValidate(int n,int maxPosTrain,int maxNegTrain);
	
	
	/**
	 * Creates an array with string values, to be parsed by the implementation
	 * of the classifier. Needed to be place in the classifier-interface to reduce
	 * implementation issues. 
	 * @param fold The fold of the crossvalidation
	 * @param fileIn The file containing the training features. 
	 * @param fileOut The file for output (if applicable).
	 * @return Commandline array with options
	 */
	public String[] prepareCrossvalidationCommand(int fold,String fileIn,String fileOut);
	
	
	/**
	 * Creates an array with string values, to be parsed by the classifier.
	 * @param fileIn The name of the file containing the extracted features.
	 * @param fileOut The name of the file to which the output should be written (if applicable).
	 * @return Commandline array with options
	 */
	public String[] prepareTrainingCommand(String fileIn,String fileOut);
	
	
	/**
	 * This method builds an SVM model file from a file with trainingexamples.
	 * The name of the file containing the extracted features should have been
	 * defined prior to this.
	 */
	public boolean buildClassifier();
	
	
	/**
	 * Merges the featurefiles (one with positive training features, one with negative
	 * training features), in order to make the actual training file. The type
	 * (donor/acceptor) is set by the constructor of the svm's, so it can take
	 * the necessary information from there.
	 * @param tempFilePositive The name of the file with features for positive training
	 * @param tempFileNegative The name of the file with features for negative training
	 * @return The resulting training file.
	 */
	public File mergeFeatureFiles(File tempFilePositive,File tempFileNegative);
	
	
	/**
	 * This method writes the temporary featuredata (being all the features
	 * extracted from 1 sequence, each feature in a different list) to a file.
	 * The type of data is also important, since it is necessary to label the 
	 * data (positive, negative, unknown) according to the type of data.
	 * 
	 * At first all data (all features of all sequences) was kept in memory, but this
	 * very quickly resulted in heap overflow problems, for large datasets.
	 * @param tempFileName The name of the file to which the data should be written.
	 * @param forward_strand Indicates whether or not the data is located on the forward strand. 
	 * @param data The featuredata, put in a nested linked list.
	 * @param dataType The type of data (see enum in this interface)
	 * @return The file with the temporary data.
	 */
	public File writeTemporaryFeatureData(String tempFileName,boolean forward_strand,List<List<Double>> data,DATA_TYPE dataType);
	
	
	/**
	 * This method writes the temporary featuredata (being all the features
	 * extracted from 1 sequence, each feature in a different list) to a file.
	 * The type of data is also important, since it is necessary to label the 
	 * data (positive, negative, unknown) according to the type of data.
	 * 
	 * At first all data (all features of all sequences) was kept in memory, but this
	 * very quickly dissolved heap overflow problems.
	 * @param tempFileName The name of the file to which the data should be written.
	 * @param data The featuredata
	 * @param dataType The type of data (see enum in this interface)
	 * @return The file with the temporary data.
	 */
	public File writeTemporaryFeatureData(String tempFileName,List<List<Double>> data,DATA_TYPE dataType);
	

	
	/**
	 * Creates a string of features for only one featurevector
	 * @param data The featuredata 
	 * @param dataType The datatype (positive,negative,unclassified)
	 * @return The featurestring
	 */
	public String generateFeatureString(List<Double> data,DATA_TYPE dataType);
	
	
	
	/**
	 * Sets the modelfile, and - dependend on the implementation -
	 * there may be an attempt to build the SVM  from this modelfile. 
	 * @return Whether or not loading the classifier succeeded.
	 */
	public boolean loadClassifier();
	
	/**
	 * Sets the modelfile, and - dependend on the implementation -
	 * there may be an attempt to build the SVM  from this modelfile. 
	 * @param modelFile The name of the modelfile
	 * @return Whether or not loading the classifier succeeded.
	 */
	public boolean loadClassifier(String modelFile);
	
	/**
	 * Use the trained (or untrained, the modelfile must be set before though)
	 * SVM to classify data, and write the output to an outputfile
	 * @param testFile The name of the file that contains the extracted features,
		outputdirectory is supposed to be in the filename. 
	 * @param outputFile The name of the outputfile, outputdirectory is
	 * 	supposed to be in the filename.
	 */
	public void classify(String testFile,String outputFile);
	
	/**
	 * Use the trained classifier to classify a single instance of data, defined by
	 * the instance parameter.
	 * @param instance The instance (consisting of extracted features) to be classified.
	 * @return The string indicating the result of the classification.
	 */
	public String classify_single_instance(String instance);
	
	/**
	 * Use the trained classifier to classify a single instance of data in a very fast way,
	 * without having to resort to string parsing procedures (recommanded method for
	 * doing these classifications).
	 * @param features The features that make up the instance that needs to be classified.
	 * @return A value given by the classifier. This is NOT a simple -1/+1 value. Indeed, 
	 * when using SVM's this value indicates the distance to the hyperplane.
	 */
	public Double classify_single_instance_fast(double[]features);
	
	
	/**
	 * Method which produces a string that is similar to that of the Splicemachine
	 * program, according to the provided results. 
	 * @param classification_result The result of the classification, in string format.
	 * @param funsite The location of the functional site in the sequence. 
	 * @param increase An extra increae for the output (see documentation).
	 * @param classification_name The name for this type of functional site.
	 * @return The string in Splicemachine format.
	 */
	public String to_splice_machine_output(String classification_result, int funsite,int increase, String classification_name);
	
	
	/**
	 * Method which produces a string that is similar to the output provided by
	 * Splicemachine (with the provided results).
	 * @param distance A value (distance to hyperplane for SVM's) that is used to give a score to 
	 * a certain functional site. 
	 * @param funsite The location of the fuctional site in the sequence.
	 * @param increase An extra increase for the location of the functional site
	 * in the output (see documentation).
	 * @param classification_name The name for this type of evaluated functional site.
	 * @return The string in Splicemachine format.
	 */
	public String to_splice_machine_output(Double distance, int funsite,int increase, String classification_name);
	
	
	/**
	 * Method which produces a string that can be used by the GenomeView program.
	 * @param id A unique id for the functional site in the sequence 
	 * @param distance  A value (distance to hyperplane for SVM's) that is used to give a score to 
	 * a certain functional site.
	 * @param funsite_start The start of the functional site in the sequence
	 * @param funsite_stop The stop of the functional site in the sequence
	 * @param classification_name The name for this type of evaluated functional site.
	 * @return The string in Splicemachine format.
	 */
	public String to_genomeview_output(int id, Double distance, int funsite_start, int funsite_stop, String classification_name);
	
	/**
	 * Returns the amount of positive and negative examples in a trainingfile.
	 * @param file The trainingfile
	 * @return An array of size 2, with the first number being the amount of
	 * positive training examples and the second number the amount of negative
	 * training examples. 
	 */
	public int[] getPosNegExamplesInFile(File file);
	
	
	/**
	 * @return The name of the file containg the model for the classifier.
	 */
	public String getModelFile();
	
	/**
	 * Changes the model for the classifier by changing the name of the modelfile.
	 * @param svmModelFile The name of the file containg the new model.
	 */
	public void setModelFile(String svmModelFile);
	
	/**
	 * Changes the various options of this classifier.
	 * @param options The new set of options for this classifier.
	 */
	public void setOptions(ClassifierOptions options);
	
	/**
	 * Returns the options set for this classifier
	 * @return The set of options for this classifier.
	 */
	public ClassifierOptions getOptions();
	
	/**
	 * Performs feature selection by evaluating different attributes.
	 * @param sort Whether to sort the resulting valposcombinations according to their values
	 * @param evaluator The evaluator used for performing the evaluation of the attributes
	 * @return A list with the values and the original positions of those values in order
	 * to be able to locate the classificationfeature this attribute belonged to.
	 */
	public List<ValPosCombination> performAttributeEvaluation(boolean sort,AttributeEvaluator evaluator);
	
	
	/**
	 * @return The instances used for training the model of this classifier.
	 */
	public Instances getTrainingFileInstances();
	
	
	/**
	 * After having used featureselection to get a filter, this filter can be used
	 * to change the featurefiles in order to optimize the svms.
	 * 
	 * @param attributeFilter The filter: this is an array with the numbers of the attributes
	 * that MUST be preserved.
	 * @param maxNumFeatures The maximum amount of features to be used by the classifier. 
	 * @param toBeFilteredFile The file containing the various features (set in a classifier dependend
	 * way) which should be filtered by the given attributefilter.
	 */
	public void applyAttributeFilter(List<Integer> attributeFilter, int maxNumFeatures, File toBeFilteredFile);
	
	/**
	 * @return the sigmoid variable A (see documentation about restructuring the output
	 * by use of sigmoid curves)
	 */
	public double getSigmoid_A();
	
	/**
	 * Changes the sigmoid variable A (see documentation about restructuring the output
	 * by use of sigmoid curves)
	 * @param sigmoid_A The new sigmoid variable A
	 */
	public void setSigmoid_A(double sigmoid_A);
	
	/**
	 * @return the sigmoid variable B (see documentation about restructuring the output
	 * by use of sigmoid curves)
	 */
	public double getSigmoid_B();
	
	/**
	 * Changes the sigmoid variable B (see documentation about restructuring the output
	 * by use of sigmoid curves)
	 * @param sigmoid_B The new sigmoid variable B
	 */
	public void setSigmoid_B(double sigmoid_B);
	
	
	/**
	 * @return The file extension to be used by files containing the features 
	 * that will be used for either building the classification model or for evaluation.
	 */
	public String getFileExtension();
}
