/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
package classifier;

import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.text.NumberFormat;
import java.util.ArrayList;
import java.util.List;
import java.util.Locale;
import java.util.Random;
import java.util.zip.ZipEntry;
import java.util.zip.ZipInputStream;
import java.util.zip.ZipOutputStream;

import main.FunSiPJar;

import org.apache.log4j.Logger;

import base.ClassificationAction;


import util.featureSelection.AbstractAttributeEvaluation;
import util.featureSelection.ValPosCombination;
import util.results.CrossValidationResult;
import util.results.SigmoidFunction;
import weka.attributeSelection.AttributeEvaluator;
import weka.classifiers.Evaluation;
import weka.core.Instance;
import weka.core.Instances;
import classifier.wekaSVM.SMO;

/**
 * This class represents the implementation of the classifier interface, suited
 * to be used with the SMO (SVM) implementation of the WEKA(TM) machine learning library.
 * 
 * @author Michiel Van Bel
 */
public class SMO_Weka implements Classifier{

	/*
	 * Some class variables, used for 
	 * internal working only.
	 */
	private Logger logger = null;
	private ClassificationAction ca = null;	
	private SVMOptions svm_options = new SVMOptions();
	private double sigmoid_A = 0.0;
	private double sigmoid_B = 0.0;
	
	private String modelFile = null;
	private NumberFormat format = null;
	private SMO smo = null;
	private Instances instances = null;
	
	/**
	 * Constructor, just initiates the necessary variables.
	 * @param logger The logging facility
	 * @param ca The ClassificationAction to which this classifier belongs (every
	 * classifier belongs to a classification action).
	 */
	public SMO_Weka(Logger logger, ClassificationAction ca){
		this.logger = logger;
		this.ca = ca;
		smo = new SMO();
		smo.turnChecksOff();
		format = NumberFormat.getInstance(Locale.US);
		format.setMaximumFractionDigits(6);
	}
	
	/**
	 * @return The file extension to be used by files containing the features 
	 * that will be used for either building the classification model or for evaluation.
	 */
	public String getFileExtension(){
		return ".arff";
	}
	
	
	/**
	 * Performs a crossvalidation of trainingfile. The results of this crossvalidation
	 * (number of true positives,false positives,true negatives,false negatives and 
	 * deduced numbers) are then returned.
	 * The normal procedure for crossvalidation can be followed. However, we extended
	 * the notion of it so different numbers of positive/negative examples can be used
	 * during the training phase of each crossvalidation step. The number of positive/
	 * negative examples for the testing phase during each crossvalidation step remains
	 * the same (begin (n-1)*(total_amount)). For further information, see the
	 * Crossvalidation.txt document in the /doc subdirectory.
	 * @param n The fold of the crossvalidation. Frequent numbers are 2,5 and 10
	 * @param maxPosTrain The maximum amount of positive training examples during the 
	 * training phase of the crossvalidation.
	 * @param maxNegTrain The maximum amount of negative training examples during the
	 * training phase of the crossvalidation.
	 * @return The crossvalidationresult object, which contains the results of
	 * the crossvalidation, and the deduced statistics.
	 */
	public CrossValidationResult crossValidate(int n,int maxPosTrain,int maxNegTrain){
		CrossValidationResult result = null;
		String fileIn = ca.getOutputdirectory()+ca.getFeatures_output_file();	
		File file = new File(fileIn);
		if(!file.exists())
			return null;
		try{
			FileReader reader = new FileReader(file);
			Instances instances = new Instances(reader);
			instances.setClassIndex(0);
			String[] options = prepareCrossvalidationCommand(n,null,null);
			smo.setOptions(options);
			reader.close();			
			Evaluation eval =  new Evaluation(instances);
			eval.crossValidateModel(smo,instances, n, new Random(1));
			
			result = new CrossValidationResult();
			result.setFalseNegatives((int)eval.numFalseNegatives(1));
			result.setFalsePositives((int)eval.numFalsePositives(1));
			result.setTrueNegatives((int)eval.numTrueNegatives(1));
			result.setTruePositives((int)eval.numTruePositives(1));
			result.setTotalNegatives(result.getTrueNegatives()+result.getFalsePositives());
			result.setTotalPositives(result.getTruePositives()+result.getFalseNegatives());
			
			//do not use result.computeDeductedData() -- instead, use the functions from eval
			result.computeDeductedData();
			result.setAuc(eval.areaUnderROC(1));
		}
		catch(Exception exc){
			logger.error(exc);
		}
		return result;
	}
	
	
	
	/**
	 * Creates an array with string values, to be parsed by the implementation
	 * of the classifier. Needed to be place in the classifier-interface to reduce
	 * implementation issues. 
	 * @param fold The fold of the crossvalidation
	 * @param fileIn The file containing the training features. 
	 * @param fileOut The file for output (if applicable).
	 * @return Commandline array with options
	 */
	public String[] prepareCrossvalidationCommand(int fold,String fileIn,String fileOut){
		String[] options = {"-C",""+svm_options.complexity,"-V",""+fold};
		return options;
	}
	
	/**
	 * Creates an array with string values, to be parsed by the classifier.
	 * @param fileIn The name of the file containing the extracted features.
	 * @param fileOut The name of the file to which the output should be written (if applicable).
	 * @return Commandline array with options
	 */
	public String[] prepareTrainingCommand(String fileIn,String fileOut){
		String[] options = {"-C",""+svm_options.complexity};
		return options;
	}
	
	/**
	 * This method builds an SVM model file from a file with trainingexamples.
	 * The name of the file containing the extracted features should have been
	 * defined prior to this.
	 */
	public boolean buildClassifier(){

		String fileIn = ca.getOutputdirectory()+ca.getFeatures_output_file();
		String fileOut = ca.getOutputdirectory()+ca.getModel_file();
		
		try{
			String [] options = prepareTrainingCommand(fileIn,fileOut);
			smo.setOptions(options);
			FileReader reader = new FileReader(new File(fileIn));
			Instances instances = new Instances(reader);
			instances.setClassIndex(0);
			this.instances = instances;
			
			smo.buildClassifier(instances);
			
			/* 
			 * apparently, we have to build our own model save system,
			 * since it is not included in standard API -- SCREW WEKA
			 */
			saveModel(smo,fileOut);
			
			calculateSigmoids();
			writeExtraModelFile();
			reader.close();
			return true;
		}
		catch(Exception exc){
			logger.error(exc);
			exc.printStackTrace();
			return false;
		}
	}
	
	
	/**
	 * This method saves the SMO java-object to a file in the filesystem.
	 * The file is zipped in order to make sure that no too much space is taken.
	 * @param smo The SMO to be stored
	 * @param fileOut The name of the file in which the SMO should be stored.
	 * @throws Exception Exceptions might occur, but the caller should handle them
	 */
	public void saveModel(SMO smo,String fileOut) throws Exception{
		ZipOutputStream zos = new ZipOutputStream(new BufferedOutputStream(new FileOutputStream(fileOut)));
		ZipEntry entry = new ZipEntry(fileOut);
		zos.putNextEntry(entry);
		ObjectOutputStream out = new ObjectOutputStream(zos);
		out.writeObject(smo);
		out.close();
	}
	
	
	/**
	 * This method loads the SMO java-object from a file. 
	 * @param fileIn The name of the file containing the SMO object
	 * @return The SMO that is extracted
	 * @throws Exception should be handled by the caller
	 */
	public SMO loadModel(String fileIn) throws Exception{
		String loc = fileIn;
		
		InputStream stream  = getClass().getResourceAsStream(loc);
		BufferedInputStream buf = null;
		if(stream!=null)
			buf = new BufferedInputStream(stream);
		else
			buf = new BufferedInputStream(new FileInputStream(loc));
				
		ZipInputStream zis = new ZipInputStream(buf);
		zis.getNextEntry();
		ObjectInputStream in = new ObjectInputStream(zis);
		SMO local_smo = (SMO) in.readObject();
		in.close();
		return local_smo;
	}
	
	
	/**
	 * This method writes an extra model file. This file contains the
	 * sigmoid values that are used for restructuring the output (these
	 * values cannot be saved in the normal way). 
	 */
	private void writeExtraModelFile(){
		try{
			String out = ca.getOutputdirectory()+ca.getModel_file()+".extra";
			BufferedWriter writer = new BufferedWriter(new FileWriter(new File(out)));
			writer.write("SIGMOID_A\t"+sigmoid_A+"\n");
			writer.write("SIGMOID_B\t"+sigmoid_B+"\n");
			writer.close();
		}
		catch(Exception exc){
			logger.error(exc);
		}
	}
	
	/**
	 * This method calculates the sigmoid variables, after training of the
	 * SVM model. The sigmoid variables are computed according to the distances
	 * of the support vectors to the hyperplane. 
	 */
	private void calculateSigmoids(){
		int[] pos_neg = getPosNegExamplesInFile(new File(ca.getOutputdirectory()+ca.getFeatures_output_file()));
		double[] distances = smo.getDistances();
		
		double [] targets = new double[pos_neg[0]+pos_neg[1]];
		for(int i=0;i<pos_neg[0];i++){
			targets[i] = 1.0;
		}
		for(int i=0;i<pos_neg[1];i++){
			targets[pos_neg[0]+i] = -1.0;
		}
		Double[] A_B = new Double[2];
		SigmoidFunction sigmoid = new SigmoidFunction();
		sigmoid.trainSigmoid(distances,targets, pos_neg[1], pos_neg[0], A_B);
		sigmoid_A = A_B[0];
		sigmoid_B = A_B[1];	
	}
	
	
	
	/**
	 * Merges the featurefiles (one with positive training features, one with negative
	 * training features), in order to make the actual training file. The type
	 * (donor/acceptor) is set by the constructor of the svm's, so it can take
	 * the necessary information from there.
	 * @param tempFilePositive The name of the file with features for positive training
	 * @param tempFileNegative The name of the file with features for negative training
	 * @return The resulting training file.
	 */
	public File mergeFeatureFiles(File tempFilePositive,File tempFileNegative){
		try{
			String out = ca.getOutputdirectory()+ca.getFeatures_output_file();
			File file = new File(out);
			BufferedWriter writer = new BufferedWriter(new FileWriter(file));
			BufferedReader posReader = new BufferedReader(new FileReader(tempFilePositive));
			BufferedReader negReader = new BufferedReader(new FileReader(tempFileNegative));
			String s = posReader.readLine();
			//include the header
			while(s!=null){
				writer.write(s+"\n");
				s = posReader.readLine();
			}
			
			//exclude the header
			s = negReader.readLine();
			while(s!=null){
				if(!s.startsWith("@")){
					writer.write(s+"\n");
				}
				s = negReader.readLine();
			}
			writer.flush();
			posReader.close();
			negReader.close();
			writer.close();
			tempFilePositive.delete();
			tempFileNegative.delete();
			return file;
			
		}
		catch(Exception exc){
			logger.error(exc);
			return null;
		}
	}
	
	/**
	 * This method writes the temporary featuredata (being all the features
	 * extracted from 1 sequence, each feature in a different list) to a file.
	 * The type of data is also important, since it is necessary to label the 
	 * data (positive, negative, unknown) according to the type of data.
	 * 
	 * At first all data (all features of all sequences) was kept in memory, but this
	 * very quickly resulted in heap overflow problems, for large datasets.
	 * @param tempFileName The name of the file to which the data should be written.
	 * @param forward_strand Indicates whether or not the data is located on the forward strand. 
	 * @param data The featuredata, put in a nested linked list.
	 * @param dataType The type of data (see enum in this interface)
	 * @return The file with the temporary data.
	 */
	public File writeTemporaryFeatureData(String tempFileName,boolean forward_strand,List<List<Double>> data,DATA_TYPE dataType){
		try{
			String name = null;
			if(forward_strand)
				name = ca.getOutputdirectory()+"forward_"+tempFileName;
			else
				name = ca.getOutputdirectory()+"reverse_"+tempFileName;
			if(!name.endsWith(".arff"))
				name+=".arff";
	
			List<Double> featureList = new ArrayList<Double>();
			for(int j=0;j<data.size();j++){
				List<Double> temp = data.get(j);
				featureList.addAll(temp);
			}
			
			File tempFile = new File(name);
			writeArffData(tempFile,featureList,dataType);
			
			return tempFile;
		}
		catch(Exception exc){
			logger.error(exc);
			return null;
		}
	}
	
	
	/**
	 * Creates the header of a ARFF file (used by the WEKA machine learning library). The
	 * resulting header is then written directly to a Writer-object.
	 * @param writer The writer-object to which the header should be written.
	 * @param data The data of the instances, for which a header must be created.
	 * @throws IOException When writing the header to the writer-object fails.
	 */
	private void createArffHeader(BufferedWriter writer,List<Double> data)throws IOException{
		writer.write("@relation arff_data\n");
		writer.write("@attribute class_attribute_1 {-1,1}\n");
		for(int i=0;i<data.size();i++){
			writer.write("@attribute attribute_"+(i+1)+" numeric\n");
		}
		writer.write("@data\n");
	}
	
	
	
	/**
	 * Creates the header of a ARFF file (used by the WEKA machine learning library). The
	 * resulting header is then written directly to a Writer-object.
	 * @param filter A feature filter.
	 * @param writer The writer object to which the header should be written.
	 * @throws IOException Then writing the header to the writer-object fails.
	 */
	private void createArffHeader(List<Integer> filter, BufferedWriter writer) throws IOException{
		writer.write("@relation arff_data\n");
		writer.write("@attribute class_attribute_1 {-1,1}\n");
		for(int i=1;i<filter.size();i++){
			writer.write("@attribute attribute_"+i+" numeric\n");
		}
		writer.write("@data \n");
	}
	
	
	/**
	 * Write feature data to a file in arff format.
	 * @param file The file to which the data should be written. 
	 * @param data The feature data that should be put in arff format and written to a file.
	 * @param dataType The datatype of the data.
	 * @throws IOException Then writing the arff formatted data to the file fails.
	 */
	private void writeArffData(File file, List<Double> data, DATA_TYPE dataType) throws IOException{
		boolean writeHeader = false;
		if(file.exists()){
			BufferedReader reader = new BufferedReader(new FileReader(file));		
			String s = reader.readLine();
			if(s==null)
				writeHeader = true;
			reader.close();
		}
		else{
			writeHeader = true;
		}
		
		BufferedWriter writer = new BufferedWriter(new FileWriter(file,true));
		if(writeHeader)
			createArffHeader(writer,data);
		
		StringBuffer buffer = new StringBuffer();
		buffer.append("{");
		switch(dataType){
		case POSITIVE_DATA : buffer.append("0 1");break;
		case NEGATIVE_DATA :buffer.append("0 -1"); break;
		case UNCLASSIFIED_DATA : buffer.append("0 ?");break;
		}

		for(int i=0;i<data.size();i++){
			Double d = data.get(i);
			int index= i+1;
			if(d!=0.0 && d!=0){
				buffer.append(","+index+" "+d);
			}
		}
		buffer.append("}\n");
		String  str = new String(buffer);
		writer.write(str);
		writer.flush();
		writer.close();
	}
	
	
	/**
	 * This method writes the temporary featuredata (being all the features
	 * extracted from 1 sequence, each feature in a different list) to a file.
	 * The type of data is also important, since it is necessary to label the 
	 * data (positive, negative, unknown) according to the type of data.
	 * 
	 * At first all data (all features of all sequences) was kept in memory, but this
	 * very quickly dissolved heap overflow problems.
	 * @param tempFileName The name of the file to which the data should be written.
	 * @param data The featuredata
	 * @param dataType The type of data (see enum in this interface)
	 * @return The file with the temporary data.
	 */
	public File writeTemporaryFeatureData(String tempFileName,List<List<Double>> data,DATA_TYPE dataType){
		try{
			String name = ca.getOutputdirectory()+tempFileName;
			if(!name.endsWith(".arff"))
				name+=".arff";
	
			File tempFile = new File(name);
			List<Double> featureList = new ArrayList<Double>();
			for(int j=0;j<data.size();j++){
				List<Double> temp = data.get(j);
				featureList.addAll(temp);
			}
			//System.out.println("--------> training data length : "+featureList.size());
			writeArffData(tempFile,featureList,dataType);
			
			return tempFile;
		}
		catch(Exception exc){
			logger.error(exc);
			return null;
		}
		
	}
	

	
	/**
	 * Creates a string of features for only one featurevector
	 * @param data The featuredata 
	 * @param dataType The datatype (positive,negative,unclassified)
	 * @return The featurestring
	 */
	public String generateFeatureString(List<Double> data,DATA_TYPE dataType){
		StringBuffer buffer = new StringBuffer();
		buffer.append("{");
		switch(dataType){
		case POSITIVE_DATA : buffer.append("0 1");break;
		case NEGATIVE_DATA :buffer.append("0 -1"); break;
		case UNCLASSIFIED_DATA : buffer.append("0 ?");break;
		}
		for(int i=0;i<data.size();i++){
			Double d = data.get(i);
			int index  = i+1;
			if(d!=0.0 && d!=0)
				buffer.append(","+index+" "+d);
		}
		buffer.append("}");
		String  str = new String(buffer);
		return str;
	}
	
	
	
	/**
	 * Sets the modelfile, and - dependend on the implementation -
	 * there may be an attempt to build the SVM  from this modelfile. 
	 * @return Whether or not loading the classifier succeeded.
	 */
	public boolean loadClassifier(){

		try{
			String f = null;
			if(modelFile == null){
				f = ca.getOutputdirectory()+ca.getModel_file()+".extra";
			}
			else{
				f = modelFile+".extra";
			}
			
			if(FunSiPJar.RUN_FROM_JAR){
				if(modelFile.startsWith("./")||modelFile.startsWith(".\\")){
					modelFile = modelFile.substring(1,modelFile.length());
					f = modelFile+".extra";
				}
			}
			
			this.smo = loadModel(modelFile);
			this.instances = this.smo.getTrainingInstances();
			
			
			InputStream stream  = getClass().getResourceAsStream(f);
			InputStreamReader read = null;
			if(stream!=null){
				read = new InputStreamReader(stream);
			}
			else{
				read = new FileReader(new File(f));
			}
			
			
			BufferedReader reader = new BufferedReader(read);
			String s = reader.readLine();
			String[] split = s.split("\t");
			sigmoid_A = Double.parseDouble(split[1]);
			
			s = reader.readLine();
			split = s.split("\t");
			sigmoid_B = Double.parseDouble(split[1]);
			
			reader.close();
			
			return true;
		}
		catch(Exception exc){
			logger.error(exc);
			return false;
		}
	}
	
	/**
	 * Sets the modelfile, and - dependend on the implementation -
	 * there may be an attempt to build the SVM  from this modelfile. 
	 * @param svmFile The name of the modelfile
	 * @return Whether or not loading the classifier succeeded.
	 */
	public boolean loadClassifier(String svmFile){
		modelFile = svmFile;
		return loadClassifier();
	}
	
	/**
	 * Use the trained (or untrained, the modelfile must be set before though)
	 * SVM to classify data, and write the output to an outputfile
	 * @param testFile The name of the file that contains the extracted features,
		outputdirectory is supposed to be in the filename. 
	 * @param outputFile The name of the outputfile, outputdirectory is
	 * 	supposed to be in the filename.
	 */
	public void classify(String testFile,String outputFile){

	}
	
	
	/**
	 * Use the trained classifier to classify a single instance of data, defined by
	 * the instance parameter.
	 * @param instance The instance (consisting of extracted features) to be classified.
	 * @return The string indicating the result of the classification.
	 */
	public String classify_single_instance(String instance){

		try{
			String outputSVMModel = null;
			if(modelFile ==null)
				outputSVMModel = ca.getOutputdirectory()+ca.getModel_file();
			else
				outputSVMModel = modelFile;
			outputSVMModel.charAt(0); //?
			
			
			
			return null;
		}
		catch(Exception exc){
			logger.error(exc);
			return null;
		}
	}
	
	
	/**
	 * Use the trained classifier to classify a single instance of data in a very fast way,
	 * without having to resort to string parsing procedures (recommanded method for
	 * doing these classifications).
	 * @param features The features that make up the instance that needs to be classified.
	 * @return A value given by the classifier. This is NOT a simple -1/+1 value. Indeed, 
	 * when using SVM's this value indicates the distance to the hyperplane.
	 */
	public Double classify_single_instance_fast(double[] features){
				
		Double val = null;
		try{
			
			/*
			 * This checks whether the extracted data and the model are compatible. 
			 * This is necessary to prevent the user interpreting incorrect results, 
			 * because the classifier might do the classification, but the results
			 * would be totally unreliable.
			 */
			if(features.length!=(this.instances.numAttributes()-1)){//in this.instances, the class attribute is already placed.	
				logger.error("Incompatible number of attributes!!");
				logger.error("Features length (test sequence)        : "+features.length);
				logger.error("Instances length (classification model): "+this.instances.numAttributes());
				
				//try to replace the instances
				try{
					FileReader reader = new FileReader(ca.getOutputdirectory()+ca.getFeatures_output_file());
					this.instances = new Instances(reader);
				}
				catch(Exception exc){
					return null;
				}
				if(features.length!=this.instances.numAttributes()){
					logger.error("Replacement did not succeed : "+features.length+" "+this.instances.numAttributes());
					return null;
				}
					
			}
			
			double [] newFeats = new double[features.length+1];
			System.arraycopy(features, 0, newFeats, 1, features.length);
			
			Instance instance = new Instance(1,newFeats);
			instance.setDataset(this.instances);
			instance.setMissing(0);
			
			double [] vals = smo.distributionForInstance2(instance);
			val = vals[0];
		}
		catch(Exception exc){
			exc.printStackTrace();
			logger.error(exc);
			
		}
		return val;
	}
	
	/**
	 * Method which produces a string that is similar to that of the Splicemachine
	 * program, according to the provided results. 
	 * @param classification_result The result of the classification, in string format.
	 * @param funsite The location of the functional site in the sequence. 
	 * @param increase An extra increae for the output (see documentation).
	 * @param classification_name The name for this type of functional site.
	 * @return The string in Splicemachine format.
	 */
	public String to_splice_machine_output(String classification_result, int funsite,int increase, String classification_name){
		if(classification_result==null||classification_result.indexOf("\t")==-1)
			return null;
		String[] split = classification_result.split("\t");
		double distance = Double.parseDouble(split[1]);
		double prob = 1/(1+Math.exp(sigmoid_A*distance+sigmoid_B));
		/*if(distance<0)
			prob=prob*0.1;*/
		int splice_final = funsite+increase;
		
		String res = splice_final+" "+classification_name+" "+prob;
		return res;		
	}
	
	
	/**
	 * Method which produces a string that is similar to the output provided by
	 * Splicemachine (with the provided results).
	 * @param distance A value (distance to hyperplane for SVM's) that is used to give a score to 
	 * a certain functional site. 
	 * @param funsite The location of the fuctional site in the sequence.
	 * @param increase An extra increase for the location of the functional site
	 * in the output (see documentation).
	 * @param classification_name The name for this type of evaluated functional site.
	 * @return The string in Splicemachine format.
	 */
	public String to_splice_machine_output(Double distance, int funsite,int increase, String classification_name){
		if(distance==null)
			return null;
		Double prob = 1/(1+Math.exp(sigmoid_A*distance+sigmoid_B));
		//prob = -2*Math.cos(prob*Math.PI)-1/Math.exp(prob);
		//prob = 4*prob - 2; //lineair dependency
		//prob = -7.5*prob*prob + 11.5*prob -2;	//mid at 0.2
		//prob = -3.8*prob*prob + 7.8*prob - 2;	//mid at 0.3
		
		//prob = -10.119*prob*prob*prob + 12.5*prob*prob + 1.619*prob-2;	//mid at 4.0,
		prob = -7.369*prob*prob*prob + 9.877*prob*prob +1.492*prob -2;		//mid at 4.5
		if(prob>2)prob = 2.0;
		int splice_final = funsite+increase;
		String res = splice_final+" "+classification_name+" "+format.format(prob);
		return res;	
	}
	
	/**
	 * Method which produces a string that can be used by the GenomeView program.
	 * @param id A unique id for the functional site in the sequence 
	 * @param distance  A value (distance to hyperplane for SVM's) that is used to give a score to 
	 * a certain functional site.
	 * @param funsite_start The start of the functional site in the sequence
	 * @param funsite_stop The stop of the functional site in the sequence
	 * @param classification_name The name for this type of evaluated functional site.
	 * @return The string in Splicemachine format.
	 */
	public String to_genomeview_output(int id, Double distance, int funsite_start, int funsite_stop, String classification_name){
		if(distance==null || classification_name == null)
			return null;
		Double prob = 1/(1+Math.exp(sigmoid_A*distance+sigmoid_B));
		StringBuffer buf = new StringBuffer();
		buf.append(id+"\t");
		buf.append(funsite_start+"\t");
		buf.append(funsite_stop+"\t");
		if(classification_name.endsWith("_rev"))
			buf.append(-1+"\t");
		else
			buf.append(1+"\t");
		buf.append(classification_name+"\t");
		buf.append("FunSiP"+"\t");
		buf.append(prob);
		return buf.toString();
	}
	
	/**
	 * Returns the amount of positive and negative examples in a trainingfile.
	 * @param file The trainingfile
	 * @return An array of size 2, with the first number being the amount of
	 * positive training examples and the second number the amount of negative
	 * training examples. 
	 */
	public int[] getPosNegExamplesInFile(File file){
		try{
			FileReader reader = new FileReader(file);
			Instances instances = new Instances(reader);
			instances.setClassIndex(0);
			int pos = 0;
			int neg = 0;
			for(int i=0;i<instances.numInstances();i++){
				Instance in = instances.instance(i);
				if(in.classValue()==1 ||in.classValue()==1.0)
					pos++;
				else
					neg++;
			}
			
			reader.close();
			//System.out.println("pos/neg : "+pos +" "+neg);
			int [] ret = {pos,neg};
			return ret;
		}
		catch(Exception exc){
			logger.error(exc);
			exc.printStackTrace();
			return null;
		}
	}
	
	/**
	 * @return The name of the file containg the model for the classifier.
	 */
	public String getModelFile(){
		return modelFile;
	}
	
	/**
	 * Changes the model for the classifier by changing the name of the modelfile.
	 * @param svmModelFile The name of the file containg the new model.
	 */
	public void setModelFile(String svmModelFile){
		this.modelFile = svmModelFile;
	}
	
	
	/**
	 * Changes the various options of this classifier.
	 * @param options The new set of options for this classifier.
	 */
	public void setOptions(ClassifierOptions options){
		if(options instanceof SVMOptions){
			this.svm_options = (SVMOptions)options;
		}
	}
	
	/**
	 * Returns the options set for this classifier
	 * @return The set of options for this classifier.
	 */
	public ClassifierOptions getOptions(){
		return this.svm_options;
	}
	
	/**
	 * Performs feature selection by evaluating different attributes.
	 * @param sort Whether to sort the resulting valposcombinations according to their values
	 * @param evaluator The evaluator used for performing the evaluation of the attributes
	 * @return A list with the values and the original positions of those values in order
	 * to be able to locate the classificationfeature this attribute belonged to.
	 */
	public List<ValPosCombination> performAttributeEvaluation(boolean sort,AttributeEvaluator evaluator){
		if(evaluator==null)
			return null;
		AbstractAttributeEvaluation aae = new AbstractAttributeEvaluation(logger,getTrainingFileInstances(),evaluator);
		aae.buildEvaluator();
		if(sort)
			return aae.returnSortedEvaluation();
		else
			return aae.returnUnsortedEvaluation();
	}
	
	
	/**
	 * Returns the the instances used for training
	 * the SMO. 
	 * @return The instances that are used for training the SMO. 
	 */
	public Instances getTrainingFileInstances(){
		Instances instances = null;
		try{
			String modelFeatureFile = ca.getOutputdirectory()+ca.getFeatures_output_file();
			FileReader reader = new FileReader(new File(modelFeatureFile));
			instances = new Instances(reader);
			instances.setClassIndex(0);
		}
		catch(Exception exc){
			logger.error(exc);
			exc.printStackTrace();
		}
		return instances;
	}
	
	/**
	 * After having used featureselection to get a filter, this filter can be used
	 * to change the featurefiles in order to optimize the svms.
	 * 
	 * @param attributeFilter The filter: this is an array with the numbers of the attributes
	 * that MUST be preserved.
	 * @param maxNumFeatures The maximum amount of features to be used by the classifier. 
	 * @param toBeFilteredFile The file containing the various features (set in a classifier dependend
	 * way) which should be filtered by the given attributefilter.
	 */
	public void applyAttributeFilter(List<Integer> attributeFilter, int maxNumFeatures, File toBeFilteredFile){
		try{
			logger.debug(this.getClass().getName()+" : Applying attribute filter");
			FileReader reader= new FileReader(toBeFilteredFile);
			Instances instances = new Instances(reader);
			instances.setClassIndex(0);
			reader.close();
			toBeFilteredFile.delete();
			BufferedWriter writer = new BufferedWriter(new FileWriter(toBeFilteredFile));
			
			//writeNewHeader
			createArffHeader(attributeFilter,writer);
			
			//writeNewData
			for(int i=0;i<instances.numInstances();i++){
				Instance inst = instances.instance(i);
				double[]features = inst.toDoubleArray();
				StringBuffer buffer = new StringBuffer();
				buffer.append("{");
				int clVal = -1;
				if(inst.classValue()==1.0)
					clVal = 1;
				buffer.append("0 "+clVal);
				for(int j=1;j<attributeFilter.size();j++){
					int index = attributeFilter.get(j);
					double val = features[index];
					if(val!=0.0 && val!=0){
						buffer.append(","+j+" "+val);
					}
				}
				buffer.append("}\n");
				writer.write(new String(buffer));
			}
			writer.close();
			
			reader = new FileReader(toBeFilteredFile);
			this.instances = new Instances(reader);
			reader.close();
		}
		catch(Exception exc){
			logger.error(exc);
			exc.printStackTrace();
		}
	}
	
	/**
	 * @return the sigmoid variable A (see documentation about restructuring the output
	 * by use of sigmoid curves)
	 */
	public double getSigmoid_A(){
		return sigmoid_A;
	}
	
	/**
	 * Changes the sigmoid variable A (see documentation about restructuring the output
	 * by use of sigmoid curves)
	 * @param sigmoid_A The new sigmoid variable A
	 */
	public void setSigmoid_A(double sigmoid_A){
		this.sigmoid_A = sigmoid_A;
	}
	
	/**
	 * @return the sigmoid variable B (see documentation about restructuring the output
	 * by use of sigmoid curves)
	 */
	public double getSigmoid_B(){
		return sigmoid_B;
	}
	
	/**
	 * Changes the sigmoid variable B (see documentation about restructuring the output
	 * by use of sigmoid curves)
	 * @param sigmoid_B The new sigmoid variable B
	 */
	public void setSigmoid_B(double sigmoid_B){
		this.sigmoid_B = sigmoid_B;
	}
	
}
