/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
package main;
import java.io.File;
import java.util.ArrayList;
import java.util.List;

import org.apache.log4j.ConsoleAppender;
import org.apache.log4j.Level;
import org.apache.log4j.Logger;
import org.apache.log4j.SimpleLayout;

import base.FunSiPWorkflow;
import base.ParameterFileParser;






/**
 * This class is a Java implementation of the splicemachine code, used for 
 * splice site prediction in the annotation pipeline. The original code was 
 * written in C, but we thought it might be a good opportunity to re-implement
 * the code in Java, so it can be used with other classifiers implemented in
 * (among others) the WEKA machine learning library.
 * 
 * The original SpliceMachine was written by Dr. Sven Degroeve
 * The new SPR (formerly known as JSpliceMachine) was written by Michiel Van Bel
 * 
 * We do not really think that it is necessary to make certain methods
 * private or protected, since this application is not really to be used
 * as a library type. Incorrect use of methods can however lead to 
 * inconsistency, since there are dependencies on the local class data.
 */

/*
 * Last update : 31/07/2007
 * -------------
 * Version History :
 * -----------------
 * v0.0.1 	: 	- start with the rewriting of the entire project.
 * v0.0.2 	:	- added better extraction of the possible features(rf/comp/pos) in a sequence
 * 				- redefined the commandline parameters so both acceptor and donor classifiers 
 * 				  can be build at the same time
 * 				- rewrote the fileparsers needed for the program 
 * 				  (there were some bugs in the old versions), mostly added checks
 * v0.0.3 	:	- Designed the new feature format representation: this has become a triple
 * 				  linked list of strings. This way we can easily traverse all the features.
 * v0.0.4	: 	- Rewrote all the internal data representation of features, so it is consistent
 * 				  with the feature representation declared in v0.0.3
 * 				- Creation of the outputmethods for these features. 
 * v0.0.5	:	- Creation of the necessary methods for loading the features from 
 * 				  a file, into an instances object.
 * v0.1.1	:	- various bug fixes : traversal and reading of the features and instances contained
 * 				  some serious flaws in non-general cases, which are (mostly ?) fixed.
 * v0.1.2	:	- succesfull creation of the different SVM's is possible. Problems 
 * 				  occur however when considering the size of the files that are written
 * 				  with standard object serialization. Only the weights should be needed.
 * v0.1.3	:	- written the methods necessary to extract data from different fasta files 
 * 				  that act as testfiles
 * v0.1.4	:	- total rewrite of the changes made in version 0.1.3, since the accumulated
 * 				  amount of data otherwise breaks the memory limits of Java. 
 * 				- Now every possible splicesite has its features extracted and is evaluated.
 * 				  Only the evaluation is kept in memory.
 * v0.1.5	:	- Some bugixes from v0.1.4, mainly to do with the introduced memory saving 
 * 				  techniques.  
 * v0.1.6	:	- Writing of javadoc code with all methods in the project 
 * v0.1.7	:	- Reorganization of the project so it should be more suitable to be implemented
 * 				  with different threads. Threading is however not yet implemented since it is
 * 				  difficult to make everything threadsafe. Besides, limiting factors are memory
 * 				  and hard disk access, not the threads themselves.
 * v0.1.8	:	- Completion of the project by allowing the reuse of constructed SVM's by loading
 * 				  the objects through zipfiles
 * v0.1.9	:	- Rework part of the classification process to assert for the usage of other
 * 				  possible classifiers and SVM's (most notably, the svm-light implementation).
 * v0.2.0	: 	- move to JASPr codebase
 * 				- simplification of the featureExtraction methods
 * 				- split of svm usage
 * v0.2.1	:	- cleanup and restored documentation
 * v0.2.2 	: 	- implementation of different kernel functions
 * v0.2.3	:	- addition of the JASPr experimenter
 * v0.2.4	:	- reimplementation of the JASPr experimenter to allow for more diverse functionality 
 * v0.3.0	: 	- implementation of feature selection functionality 
 * v0.3.1	:	- extension of the feature selection technology to
 * 		a) show results | b)use fs as a filter | c)write fs to file | d) load filter from file
 * v0.3.2	:	- implementation of a feature visualisation tool, used for displaying the importance
 * 				   of features and output to SVG.
 * v0.3.3	: 	- some debugging done in all areas of the application - mostly looking for nullpointers
 * v0.3.4	:	- finalization of the usage of testing files
 * v0.3.5	: 	- Normal user output has been converted to standard SpliceMachine output
 * v0.3.6	:	- Reverse strand searching added when classifying testfiles
 * v0.4.0	:	- Total cleanup and documentation of the project.
 * v0.6.0	:	- Again, reorganisation of the entire project, to make it more extensible and
 * 				  generic. This includes removal of the donor/acceptor principle, and moving on to
 *  			  a much more generic "ACTIONSITE"-based approach, which lets the user define its own
 *  			  special places in a sequence that need to be classified.
 *  			  Renaming of the project to SPR.
 * v0.6.1	:	- Several optimisations and precautions for dealing with large sequencefiles
 *  			  and the corresponding large datasets
 * v0.6.2	: 	- Again, several optimisations are implemented to speed up the gathering
 * 				  of the different features. The performance of these, is still one of 
 * 				  of the weaker spots in the project.
 * v0.6.3	:	- addition of a more userfriendly and shorter way of calling the program (with
 * 					a *- wildcard character).
 * v0.6.4	:	- addition of a general purpose optimizer for detection of the optimal
 * 					up/downstream ranges
 * v0.6.5a	:	- addition of newly configured types of secondary structure features
 * 				- addition of new computeFunctions class which attempts to find optimal binding functions
 * v0.6.5b	:	- rewrite of the configurationfile parser in order to make it more structured
 * 				- addition of preliminary support for multithreading actions
 * v0.6.5c	:	- removed some small bugs from the new configurationfile parser
 * 				- addition of comments and information to lots of code
 * 				- further work on the local multithreading problem
 * v0.6.6	: 	- finally added the correct loadClassifier / saveClassifier to the WekaSVM implementation.
 * 					This should be enough to prevent overzealous rebuilding of the models.
 * v0.6.7a	:	- addition of the ability to use cluster-multithreading ... a serious b%*ch to
 * 					make work for real. Some issues no doublty remain.
 * v0.6.7b	:	- fixed an issue of competing localized threads for data (induced by the cluster-mt)
 * v0.6.7c	:	- fixed an issue of incorrect configuration file creation.
 * v0.6.8a	:	- addition of the "Conversion" object into the workflow, allowing classificationfeatures
 * 					to translate the sequences they work on. 
 * v0.6.8b	: 	- some changes in API's and general behaviour to accomodate for the new Conversion type
 * v0.6.9a  : 	- Conversions are loaded dynamically from the util.conversions.implementations dir
 * v0.6.9b  : 	- ClassificationFeatures are loaded dynamically from both the util.classificationFeatures.implementations
 *                 dir and the util.classificationFeatures.implementations.test dir
 *                These updates allow that classes should just be placed in the directory, without changes 
 *                to anything else.
 * v0.7.0	: 	- Changes made so several distinct filters can be loaded at once, each paired with a classificationfeature
 * v0.7.1a	:	- Addition of the SPRGui program and the ./startup/gui package, which can be used to graphically
 * 					change the behavior of SPR and - if necessary - start SPR.
 * 				- Manual for SPR created
 * v0.7.1b	:	- Various bug-fixes in SPRGui and associated panels. 
 * 				- Further work on manual : addition of full featured configuration file options detailled
 * v0.7.1c	: 	- Again, further bug-fixes in SPRGui and associated panels. Fixed (among others) the incorrect
 * 					state of saved configuration files.
 * v0.7.2	:	- Major cleanup of unused/deprecated/replaced code (especially in ./util/ package). 
 * 					No recompile needed, as there are no changes to actual used code.
 * v0.8.0	:	- Major revamp of the program, additional support for running from jar-file (and webstart).
 * 				- Rename to FunSiP
 * 				- Addition of a funtional gui
 * v0.8.1a	: 	- fixed some bugs in the configuration file code generation 
 * v0.8.1b	:	- remapping of the output directories to the user directories, in case of launch from jar-file.
 */
public class FunSiP implements Runnable{

	/**
	 * Version of the program. 
	 */
	public static final String VERSION 			= "0.8.1b";	 			
	
	private Logger logger						= Logger.getLogger(this.getClass());
	private String[] args						= null;
	
	/**
	 * Main method for FunSiP.
	 * @param args Possible commandline parameters (at least one : the name/location
	 * of the configuration file).
	 */
	public static void main(String[] args){
		if(args.length==0){
			System.err.println("No parameterfile detected on commandline.");
			System.err.println("Please refer to the manual");
			System.exit(1);
		}
			
		FunSiP spr = new FunSiP(args);
		spr.run();	//no need to start different thread
	}
	
	/**
	 * The main launcher object for the project. This object just acts as an initializer
	 * for basic components.
	 * It is assumed that there is at least one commandline parameter (containing the
	 * name of the configuration file).
	 * @param args Array with commandline parameters.
	 */
	public FunSiP(String[] args){
		this.args = args;
		logger.setLevel((Level)Level.DEBUG);
		ConsoleAppender appender = new ConsoleAppender(new SimpleLayout());
		logger.addAppender(appender);
	}
	
	/**
	 * Method necessary to launch FunSiP as a different thread. Implements the 
	 * run-method from the Runnable-interface. 
	 * This method calls the configuration file parser, and then launches the 
	 * workflow that is built from this configuration file.
	 */
	public void run(){
		logger.info("FunSiP "+VERSION+" started");
		try{
			ParameterFileParser fileParser = new ParameterFileParser(logger);
			File[] fileNames = null;
			FunSiPWorkflow workflow = fileParser.parseFile(args[0]);
			if(workflow!=null){
				if(args.length>1){
					fileNames = extractFileNames(args);
					workflow.setExplicitFiles(fileNames);
					if(workflow.getClassificationActions().size()>=2)
						workflow.setMergeResults(true);
				}
				Thread thread = new Thread(workflow);
				thread.start();
				thread.join();
			}
			else{
				logger.fatal("Could not construct workflow, terminating FunSiP");
			}
		}
		catch(Exception exc){
			logger.fatal("Exception occured in class "+this.getClass()+" in method run()");
			logger.fatal(exc);
		}		
		logger.info("Exiting FunSiP...");
	}
	

	
	/*
	 * If there is more then 1 commandline parameter, it is assumed that the first one is 
	 * the configuration file (hence loop starting from 1) and the following ones are fasta
	 * files that need to be annotated.
	 * Care is taken to also allow the *-wildcard, although on most systems the underlying
	 * operating system or commandline shell take care of the translation by translating this
	 * *-wildcard into all the possibilities and concatenating them in a space-seperated
	 * string. 
	 * 
	 * The final result is that these files are supplied to the worflow object, which will
	 * take into account the subsequent merging of the resulting classification files.
	 */
	private File[] extractFileNames(String[] args){
		File [] res = null;
		List<File> files = new ArrayList<File>();
		
		for(int i=1;i<args.length;i++){
			String temp = args[i];
			File file = new File(temp);
			String name = file.getName();
			if(name.startsWith("*")){
				String extension = name.substring(name.lastIndexOf("."),name.length());
				File dir = file.getParentFile();
				if(dir==null) 
					break;
				File [] tempfiles = dir.listFiles();
				for(int j=0;j<tempfiles.length;j++){
					if(tempfiles[j].getName().endsWith(extension))
						files.add(tempfiles[j]);
				}
			}
			else{
				files.add(file);
			}
		}
		
		
		res = new File[files.size()];
		for(int i=0;i<files.size();i++){
			res[i] = files.get(i);
		}
		return res;
	}
	

	/**
	 * @return The logging object that is used by this instance of FunSiP.
	 */
	public Logger getLogger(){
		return logger;
	}
	
}

