/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
package main;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Dimension;
import java.awt.GridLayout;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.WindowEvent;
import java.awt.event.WindowListener;
import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.Writer;
import java.net.URL;
import java.util.List;

import javax.swing.BorderFactory;
import javax.swing.ImageIcon;
import javax.swing.JButton;
import javax.swing.JFileChooser;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JList;
import javax.swing.JMenu;
import javax.swing.JMenuBar;
import javax.swing.JMenuItem;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.ListSelectionModel;
import javax.swing.UIManager;

import org.apache.log4j.ConsoleAppender;
import org.apache.log4j.FileAppender;
import org.apache.log4j.Level;
import org.apache.log4j.Logger;
import org.apache.log4j.SimpleLayout;
import org.apache.log4j.WriterAppender;

import base.ClassificationAction;
import base.FunSiPWorkflow;
import base.ParameterFileParser;
import base.gui.OptionsPanel;
import base.gui.SprClassificationActionPanel;
import base.gui.SprOutputEditorWriter;
import base.gui.SprOutputPanel;
import base.gui.SprSplashScreen;
import base.gui.SprWorkflowPanel;
import base.gui.WizardFramework;



/**
 * This class consists of the building blocks for a graphical representation 
 * of the FunSiP program. Not only can FunSiPGui launch FunSiP from the GUI, it
 * can also be used to edit configuration files or create new configuration files.
 * This goal can be achieved by either giving all the values of a configuration
 * file directly, or by using a wizard which delivers most of the available options
 * in a nice and structured way.
 * 
 * WARNING 1: The source-code of FunSiPGui is a spaghetti-mess, this has yet to be cleaned
 * up, either by restructuring or by rewriting.
 * 
 * WARNING 2: Because of its very tight connection with the FunSiP program itself, 
 * we have opted to give FunSiPGui the same version-code as FunSiP.
 * 
 * @author Michiel Van Bel
 */
public class FunSiPGui implements ActionListener,WindowListener{
	
	
	/*
	 * A whole bunch of local variables, used for 
	 * both graphical stuff and workflow stuff.
	 */
	private JFrame frame					= null;
	private SprSplashScreen splash			= null;
	private JMenuItem newMenuItem 			= null;
	private JMenuItem loadMenuItem			= null;
	private JMenuItem saveMenuItem			= null;
	private JMenuItem saveAsMenuItem		= null;
	private JMenuItem exitMenuItem	 		= null;
	private JMenuItem optionsMenuItem		= null;
	private JMenuItem wizardMenuItem		= null;
	private JMenuItem sprOutputMenuItem		= null;
	private JMenuItem launchMenuItem 		= null;
	private JMenuItem showOutputPaneMenuItem= null;
	private JMenuItem helpMenuItem 			= null;
	private JMenuItem aboutMenuItem 		= null;
	
	private String currentConfigurationFileName	= null;
	private FunSiPWorkflow workflow				= null;
	
	private Exception lastException 		= null;
	private SprOutputPanel outputPanel 		= null;
	private SprOutputEditorWriter outputWriter = null;
	
	
	private JButton changeWorkflowButton 	= null;
	private boolean savedWorkflow			= true;
	private JLabel savedWorkflowLabel		= null;
	private JLabel currentWorkflowLabel		= null;
	
	private JButton addActionButton 		= null;
	private JList actionsList				= null;
	private JButton removeActionButton 		= null;
	private JButton editActionButton		= null;
	
	private Thread funsipThread			 	= null;
	private FunSiP spr						= null;
	private SprWorkflowPanel changeWorkflowPanel = null;
	
	private WizardFramework current_wizard 	= null;
	
	private boolean isFunsipRunning			= false;
	private static final String defaultConfigurationFilePath = "/config/ath.txt";
	
	/**
	 * Main method for FunSiPGui. Launches the program.
	 * @param args The supplied commandline parameters.
	 */
	public static void main(String[] args){
		new FunSiPGui();
	}
	
	
	/**
	 * This method tries to create a FunSiPWorkflow object (one of the basic
	 * building blocks of FunSiP) from a given configuration file. 
	 * @param fileName The filename of the configuration file containing the data from
	 * which the workflow-object should be built. Can be null, in which case a simple
	 * and empty FunSipWorkflow will be returned.
	 * @return The FunSiPWorkflow object that is built. Can be null of the parameter
	 * file parser was unable to extract the correct data from the supplied configuration
	 * file.
	 */
	public FunSiPWorkflow getSPRWorkflow(String fileName){
		Logger tempLogger = Logger.getLogger("tempLogger");
		tempLogger.removeAllAppenders();
		tempLogger.addAppender(new ConsoleAppender(new SimpleLayout()));
		if(fileName!=null){
			ParameterFileParser parser = new ParameterFileParser(tempLogger);
			FunSiPWorkflow work = parser.parseFile(fileName);
			if(work==null)
				tempLogger.error("FunSiPGui : workflow is null ->getWorkflow(String fileName)");
			return work;
		}
		else{
			FunSiPWorkflow work = new FunSiPWorkflow(tempLogger,null);
			return work;
		}
	}

	/**
	 * Copies a file from one location to another, by means of 
	 * copying every single line from the original file to the new file.
	 * @param from The location and name of the source file.
	 * @param to The location and name of the destiny file.
	 * @throws IOException In case something goes wrong while copying. 
	 */
	private void copyFile(String from,String to) throws IOException{
		InputStream is = getClass().getResourceAsStream(from);
		BufferedReader reader = new BufferedReader(new InputStreamReader(is));
		BufferedWriter writer = new BufferedWriter(new FileWriter(new File(to)));
		
		String s = reader.readLine();
		while(s!=null){
			writer.write(s+"\n");
			s = reader.readLine();
		}
		
		reader.close();
		writer.close();
	}
	
	/**
	 * Changes the location of the current configuration file to be saved.
	 * @param fileName The new location and name.
	 */
	private void redirectCurrentSave(String fileName){
		try{
			String newfileName = FunSiPJar.USER_DIR+fileName;
			//int index = newfileName.lastIndexOf("\\")>newfileName.lastIndexOf("/")?newfileName.lastIndexOf("\\"):newfileName.lastIndexOf("/");
			//String config_dir = newfileName.substring(0,index);
			File f = new File(newfileName);
			File p = f.getParentFile();
			p.mkdirs();
			//copy file
			copyFile(fileName,newfileName);
			
			File currentFile = new File(newfileName);
			currentConfigurationFileName = currentFile.getAbsolutePath();
			
			String outDir = new File(FunSiPJar.USER_DIR).getAbsolutePath();
			String message = "It is detected that the application is being run from a jar-archive,\nmaking it impossible to update configuration files.\n" +
					"Output redirected to :\n\n"+outDir+"\n\nThe FunSiP directory structure is simulated in this directory.\n" +
							"The directory can be changed in the extra->options menu.\n";
			JOptionPane.showMessageDialog(splash,message,"New location of files", JOptionPane.INFORMATION_MESSAGE);
			
		}
		catch(Exception exc){
			System.err.println(exc);
		}
	}
	
	
	/**
	 * Constructor for the graphical framework.
	 * Loads and configures all the necessary elements for a GUI.
	 */
	public FunSiPGui(){
		
		//load and display the splashscreen.
		splash = new SprSplashScreen();
		new Thread(splash).start();
		splash.updateMessage("Constructing gui-elements.");
		try{Thread.sleep(200);}catch(Exception e){}
		
		try {UIManager.setLookAndFeel(UIManager.getSystemLookAndFeelClassName());} 
		catch(Exception ex){}
		
		frame = new JFrame("FunSiP v"+FunSiP.VERSION+"   (c)2007  -  Michiel Van Bel   - Ghent University");
		currentConfigurationFileName = defaultConfigurationFilePath;
		
		/*
		 * Just because jar-files cannot be updated on the fly.
		 */
		if(FunSiPJar.RUN_FROM_JAR)
			redirectCurrentSave(currentConfigurationFileName);
			
			
		workflow = getSPRWorkflow(currentConfigurationFileName);
		frame.setDefaultCloseOperation(JFrame.DO_NOTHING_ON_CLOSE);
		frame.addWindowListener(this);
		
		JFrame.setDefaultLookAndFeelDecorated(true);
		splash.updateMessage("Constructing gui-elements..");
		try{Thread.sleep(200);}catch(Exception e){}
		
		frame.setMinimumSize(new Dimension(430,400));
		frame.setJMenuBar(createMenuBar());
		
		JPanel pan = createSPRGuiContent();
		frame.setContentPane(pan);

		splash.updateMessage("Constructing gui-elements...");
		try{Thread.sleep(200);}catch(Exception e){}
		
		frame.setLocationRelativeTo(null);
		frame.setResizable(false);
		splash.updateMessage("Constructing gui-elements....");
		try{Thread.sleep(200);}catch(Exception e){}
		outputPanel 		= new SprOutputPanel();
		outputWriter		= new SprOutputEditorWriter();
	
		splash.remove();
		splash = null;
		
		frame.pack();
		frame.setVisible(true);
	}
	
	/**
	 * Creation of the menubar
	 * @return JMenuBar object that is used by FunSiPGui.
	 */
	private JMenuBar createMenuBar(){
		JMenuBar menuBar = new JMenuBar();
		
		JMenu fileMenu = new JMenu("File");
		JMenu extraMenu = new JMenu("Extra");
		JMenu launchMenu = new JMenu("Launch");
		JMenu helpMenu = new JMenu("Help");
		
		newMenuItem 	= new JMenuItem("New",'N');
		loadMenuItem 	= new JMenuItem("Load",'L');
		saveMenuItem	= new JMenuItem("Save",'S');
		saveAsMenuItem	= new JMenuItem("Save As...");
		exitMenuItem 	= new JMenuItem("Exit",'X');
		optionsMenuItem	= new JMenuItem("Options",'O');
		wizardMenuItem	= new JMenuItem("Wizard",'W');
		launchMenuItem 	= new JMenuItem("Launch",'L');
		showOutputPaneMenuItem = new JMenuItem("Show output",'S');
		sprOutputMenuItem = new JMenuItem("Output options",'O');
		helpMenuItem	= new JMenuItem("Help",'H');
		aboutMenuItem	= new JMenuItem("About",'A');
		
		newMenuItem.addActionListener(this);
		loadMenuItem.addActionListener(this);
		saveMenuItem.addActionListener(this);
		saveAsMenuItem.addActionListener(this);
		exitMenuItem.addActionListener(this);
		optionsMenuItem.addActionListener(this);
		wizardMenuItem.addActionListener(this);
		launchMenuItem.addActionListener(this);
		showOutputPaneMenuItem.addActionListener(this);
		sprOutputMenuItem.addActionListener(this);
		helpMenuItem.addActionListener(this);
		aboutMenuItem.addActionListener(this);
		
		fileMenu.add(newMenuItem);
		fileMenu.add(loadMenuItem);
		fileMenu.addSeparator();
		fileMenu.add(saveMenuItem);
		fileMenu.add(saveAsMenuItem);
		fileMenu.addSeparator();
		fileMenu.add(exitMenuItem);
		
		extraMenu.add(optionsMenuItem);
		extraMenu.addSeparator();
		extraMenu.add(wizardMenuItem);
		
		launchMenu.add(launchMenuItem);
		launchMenu.add(showOutputPaneMenuItem);
		launchMenu.addSeparator();
		launchMenu.add(sprOutputMenuItem);
		
		helpMenu.add(helpMenuItem);
		helpMenu.addSeparator();
		helpMenu.add(aboutMenuItem);
		
		menuBar.add(fileMenu);
		menuBar.add(extraMenu);
		menuBar.add(launchMenu);
		menuBar.add(helpMenu);
		
		showOutputPaneMenuItem.setEnabled(false);
		
		return menuBar;
	}
	
	
	/**
	 * If a new workflow is loaded, then the list displaying the various classification
	 * actions defined in the workflow also need to be adjusted. 
	 * This is done by this method.
	 * @param work The newly loaded workflow object.
	 * @param li The JList, which must be adjusted to display the new values.
	 */
	private void fillActionsList(FunSiPWorkflow work,JList li){
		String [] initialvalues = null;
		if(workflow!=null){
			List<ClassificationAction> caList = workflow.getClassificationActions();
			initialvalues = createActionsArray(caList);
		}
		else
			initialvalues = new String[0];
		li.setListData(initialvalues);
	}
	
	/**
	 * Creates the main content panel and returns it.
	 * @return The main content JPanel.
	 */
	private JPanel createSPRGuiContent(){
		JPanel panel = new JPanel(new BorderLayout());
		JPanel workflowPanel = new JPanel(new BorderLayout());
		JPanel actionsPanel  = new JPanel();
				
		/*int h = frame.getHeight(); 
		int w = frame.getWidth(); 
		Dimension workflowDimension = new Dimension(w,h/4);
		Dimension actionDimension = new Dimension(w,3*h/4);
		Dimension ad1 = new Dimension(2*w/4,2*h/4);*/
		
		Dimension workflowDimension = new Dimension(430,100);
		Dimension actionDimension = new Dimension(430,300);
		Dimension ad1 = new Dimension(210,200);
		workflowPanel.setPreferredSize(workflowDimension);
		workflowPanel.setMaximumSize(workflowDimension);
		actionsPanel.setPreferredSize(actionDimension);
		actionsPanel.setMaximumSize(actionDimension);
		workflowPanel.setBorder(BorderFactory.createTitledBorder("Workflow options"));
		actionsPanel.setBorder(BorderFactory.createTitledBorder("Actions options"));
		panel.add(workflowPanel,BorderLayout.NORTH);
		panel.add(actionsPanel,BorderLayout.SOUTH);
		
		changeWorkflowButton 	= new JButton("Change workflow options");
		addActionButton 		= new JButton("Add action");
		removeActionButton 		= new JButton("Remove action");
		editActionButton		= new JButton("Edit action");
		
		/*String[] initialvalues = null;
		if(workflow!=null){
			List<ClassificationAction> caList = workflow.getClassificationActions();
			initialvalues = createActionsArray(caList);
		}
		else
			initialvalues = new String[0];
		actionsList				= new JList(initialvalues);*/
		actionsList = new JList();
		actionsList.setSelectionMode(ListSelectionModel.SINGLE_SELECTION);
		fillActionsList(workflow,actionsList);
		
		changeWorkflowButton.addActionListener(this);
		addActionButton.addActionListener(this);
		removeActionButton.addActionListener(this);
		editActionButton.addActionListener(this);
		//JLabel workflowLabel = new JLabel("Workflow : general options");
		savedWorkflowLabel		= new JLabel("");
		currentWorkflowLabel	= new JLabel("");
		updateLabels();
		JLabel actionsLabel  = new JLabel("Actions");
		
		JPanel wf1p = new JPanel(new BorderLayout());
		JPanel wf2p = new JPanel();
		wf1p.add(savedWorkflowLabel,BorderLayout.NORTH);
		wf1p.add(currentWorkflowLabel,BorderLayout.SOUTH);
		wf2p.add(changeWorkflowButton);
		workflowPanel.add(wf1p,BorderLayout.NORTH);
		workflowPanel.add(wf2p,BorderLayout.SOUTH);

		JPanel actionsListPanel = new JPanel(new BorderLayout());
		JPanel actionsButtonPanel = new JPanel(new GridLayout(5,1));
		actionsListPanel.add(actionsLabel,BorderLayout.NORTH);
		JScrollPane scroll = new JScrollPane(actionsList);
		//scroll.setMinimumSize(new Dimension(w/2,h/5));
		actionsListPanel.add(scroll,BorderLayout.CENTER);
		actionsButtonPanel.add(new JPanel().add(addActionButton));
		actionsButtonPanel.add(new JPanel().add(new JLabel(" ")));
		actionsButtonPanel.add(new JPanel().add(removeActionButton));
		actionsButtonPanel.add(new JPanel().add(new JLabel(" ")));
		actionsButtonPanel.add(new JPanel().add(editActionButton));
		
		
		actionsListPanel.setPreferredSize(ad1);
		actionsListPanel.setMaximumSize(ad1);
		actionsListPanel.setMinimumSize(ad1);
		
		actionsPanel.add(actionsListPanel);
		actionsPanel.add(actionsButtonPanel);
			
		return panel;
	}
	
	
	/**
	 * Creates the configuration file, with the currently loaded SPRWorkflow
	 * object as content. 
	 * @param fileName The name of the configuration file
	 * @return The name of the configuration file. Can be null if the creation 
	 * of the configuration file failed.
	 */
	public String createConfigurationFile(String fileName){
		try{
			String t = ParameterFileParser.getConfigurationFileString(workflow);
			BufferedWriter writer = new BufferedWriter(new FileWriter(new File(fileName)));
			writer.write(t+"\n");
			writer.close();
			return fileName;
		}
		catch(Exception exc){
			this.lastException = exc;
			return null;
		}
	}
	
	
	/**
	 * Changes the output of the supplied FunSiP program, according
	 * to the content of the output panel.
	 * @param spr The FunSiP program from which the output should be changed. Can be
	 * null, in which case nothing happens.
	 * @param output The content panel which has all the different options defined
	 * as how the output can be changed. Can be null, in which case nothing happens.
	 */
	public void setSprOutput(FunSiP spr, SprOutputPanel output){
		if(spr==null || output==null)
			return;
		SprOutputPanel.OUTPUT_FILTER of = output.getOf();
		SprOutputPanel.OUTPUT out = output.getOutput();
		
		Logger logger = spr.getLogger();

		switch(of){
		case DEBUG : logger.setLevel((Level)Level.DEBUG); break;
		case INFO : logger.setLevel((Level)Level.INFO);break;
		case WARN : logger.setLevel((Level)Level.WARN);break;
		case ERROR : logger.setLevel((Level)Level.ERROR); break;
		case FATAL : logger.setLevel((Level)Level.FATAL); break;
		}
	
		try{
			switch(out){
			case STANDARD :
				ConsoleAppender consoleAppender = new ConsoleAppender(new SimpleLayout());
				logger.removeAllAppenders();
				logger.addAppender(consoleAppender);
				break;
			case TO_FILE : 
				FileAppender fileAppender = new FileAppender(new SimpleLayout(),output.getOutputFile());
				logger.removeAllAppenders();
				logger.addAppender(fileAppender);
				break;
			case TO_FRAME_CONSOLE : 
				Writer w = outputWriter.getWriter();
				WriterAppender writerAppender = new WriterAppender(new SimpleLayout(),w);
				logger.removeAllAppenders();
				logger.addAppender(writerAppender);
				
				//start logging to frame
				Thread fr_thread = new Thread(outputWriter);
				fr_thread.start();
				
				showOutputPanel();
				break;
			}
		}
		catch(IOException exc){
			String message = "Cannot change logging system \n" +
					"into using the file-appender.";
			JOptionPane.showMessageDialog(frame, message,"Error logging",JOptionPane.ERROR_MESSAGE);
		}
	}
	
	/**
	 * Displays the frame to which graphical output of FunSiP is redirected.
	 */
	private void showOutputPanel(){
		new Thread(new Runnable(){
			public void run(){
				JOptionPane.showMessageDialog(frame,outputWriter);
			}
		}).start();
	}
	
	/**
	 * Updates all the labels in the current frame of the gui, according 
	 * to the new state in which it is positioned (e.g. after changing the 
	 * content of a configuration file, but not saving it, the gui will change).
	 */
	public void updateLabels(){
		this.currentWorkflowLabel.setText("Current file : "+currentConfigurationFileName);
		if(savedWorkflow){
			this.savedWorkflowLabel.setText("Workflow saved");
			this.savedWorkflowLabel.setForeground(new Color(50,200,100));
		}
		else{
			this.savedWorkflowLabel.setText("Workflow contains unsaved data!");
			this.savedWorkflowLabel.setForeground(Color.RED);
		}
		
		if(isFunsipRunning){
			launchMenuItem.setEnabled(false);
			showOutputPaneMenuItem.setEnabled(true);
		}
		else{
			launchMenuItem.setEnabled(true);
			showOutputPaneMenuItem.setEnabled(false);
		}
	}
	
	/**
	 * Displays the the options panel.
	 */
	private void showOptionsDialog(){
		OptionsPanel op = new OptionsPanel(this,frame);
		int res = JOptionPane.showConfirmDialog(frame,op, "Options", JOptionPane.OK_CANCEL_OPTION);
		if(res==JOptionPane.OK_OPTION){
			op.setOptions();
		}
	}
	
	/**
	 * Sets a new configuration file as standard loaded configuration
	 * file.
	 * @param newConfigurationFile The name and location of the new configuration file.
	 */
	public void setNewConfiguration(String newConfigurationFile){
		if(newConfigurationFile==null)
			return;
		FunSiPWorkflow wf = this.getSPRWorkflow(newConfigurationFile);
		if(wf==null)
			return;
		this.currentConfigurationFileName = newConfigurationFile;
		setNewWorkflow(wf);
		if(current_wizard!=null)
			current_wizard = null;
	}
	
	/**
	 * Changes the current workflow. This is part of setting the 
	 * new configuration file.
	 * @param wf The new workflow.
	 */
	private void setNewWorkflow(FunSiPWorkflow wf){
		this.workflow = wf;
		fillActionsList(workflow,actionsList);
		savedWorkflow = true;
		updateLabels();
	}
	
	
	public void actionPerformed(ActionEvent e){
		//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
		//file menu handlers
		if(e.getSource().equals(newMenuItem)){
			currentConfigurationFileName = "";
			setNewWorkflow(getSPRWorkflow(null));
			/*workflow = getSPRWorkflow(null);
			fillActionsList(workflow,actionsList);
			savedWorkflow = true;
			updateLabels();*/
		}
		else if(e.getSource().equals(loadMenuItem)){
			JFileChooser chooser = new JFileChooser("./");
			int returnVal = chooser.showOpenDialog(frame);
			if(returnVal==JFileChooser.APPROVE_OPTION){
				String file = chooser.getSelectedFile().getPath();
				currentConfigurationFileName = file;
				setNewWorkflow(getSPRWorkflow(currentConfigurationFileName));
				/*workflow = getSPRWorkflow(currentConfigurationFileName);
				fillActionsList(workflow,actionsList);
				savedWorkflow = true;
				updateLabels();*/
			}
		}
		else if(e.getSource().equals(saveMenuItem)){
			if(currentConfigurationFileName==null||currentConfigurationFileName.equals("")){
				JFileChooser chooser = new JFileChooser("./");
				int returnVal = chooser.showSaveDialog(frame);
				if(returnVal==JFileChooser.APPROVE_OPTION){
					String file = chooser.getSelectedFile().getPath();
					currentConfigurationFileName = file;
					createConfigurationFile(file);
					savedWorkflow = true;
					updateLabels();
				}
			}
			else{
				createConfigurationFile(currentConfigurationFileName);
				savedWorkflow = true;
				updateLabels();
			}
		}
		
		else if(e.getSource().equals(saveAsMenuItem)){
			JFileChooser chooser = new JFileChooser("./");
			int returnVal = chooser.showSaveDialog(frame);
			if(returnVal==JFileChooser.APPROVE_OPTION){
				String file = chooser.getSelectedFile().getPath();
				currentConfigurationFileName = file;
				createConfigurationFile(file);
				savedWorkflow = true;
				updateLabels();
			}
		}
		else if(e.getSource().equals(exitMenuItem)){
			showExitDialog();
		}
		
		
		
		//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
		// extra menu options
		else if(e.getSource().equals(optionsMenuItem)){
			showOptionsDialog();
		}
		else if(e.getSource().equals(wizardMenuItem)){
			current_wizard = new WizardFramework(this,frame);
		}
		
		
		// !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
		//launch optiosn
		else if(e.getSource().equals(showOutputPaneMenuItem)){
			showOutputPanel();
		}
		
		else if(e.getSource().equals(launchMenuItem)){
			//System.out.println(currentConfigurationFileName);
			
			String fi = createConfigurationFile(currentConfigurationFileName);
			if(fi==null){
				currentConfigurationFileName = "."+currentConfigurationFileName;
				fi = createConfigurationFile(currentConfigurationFileName);
			}
				
			outputWriter.setRuntimeFunSiP(this);
			//System.out.println(fi);
			if(fi==null){
				String message = "Error while generating configurationfile.\n" +
						"Exception : "+lastException+" was generated.";
				JOptionPane.showMessageDialog(frame,message,"Could not launch FunSiP!",JOptionPane.ERROR_MESSAGE);
			}
			else{
				String[] args = {fi};
				spr = new FunSiP(args);
				setSprOutput(spr,outputPanel);
				funsipThread = new Thread(spr);
				funsipThread.start();		//start in a different thread to avoid hanging of gui
				isFunsipRunning = true;
			}
		}
		else if(e.getSource().equals(sprOutputMenuItem)){
			
			JOptionPane.showMessageDialog(frame,outputPanel);
		}
		else if(e.getSource().equals(helpMenuItem)){
			String message = "FunSiP v"+FunSiP.VERSION+" help information. \n\n" +
					"Please refer to the documentation provided in the \\doc directory.\n" +
					"If this directory does not exist or appears to be empty, please visit :\n\n" +
					"http://bioinformatics.psb.ugent.be";
			JOptionPane.showMessageDialog(frame, message,"FunSiP Help",JOptionPane.INFORMATION_MESSAGE);
		}
		else if(e.getSource().equals(aboutMenuItem)){
			String message = "FunSiP v"+FunSiP.VERSION+"\n\n" +
					"(c) Michiel Van Bel, Ghent University\n" +
					"Contact : Michiel.vanbel@psb.ugent.be\n" +
					"                http://bioinformatics.psb.ugent.be \n\n"+
					"This program is subject to the GPL-license version 2 or higher.\n" +
					"Please refer to license.txt for legal information.\n\n";
			URL cu = this.getClass().getResource("/fig/about.jpg");
			if(cu!=null)
				JOptionPane.showMessageDialog(frame,message,"About FunSiP",JOptionPane.INFORMATION_MESSAGE,new ImageIcon(cu));
			else
				JOptionPane.showMessageDialog(frame,message,"About FunSiP",JOptionPane.INFORMATION_MESSAGE);
		}
		
		//normal buttons
		else if(e.getSource().equals(changeWorkflowButton)){
			if(changeWorkflowPanel==null)
				changeWorkflowPanel = new SprWorkflowPanel(workflow);
			JOptionPane.showMessageDialog(frame, changeWorkflowPanel,"Workflow options",JOptionPane.INFORMATION_MESSAGE);
			changeWorkflowPanel.updateWorkflow();
			savedWorkflow = false;
			updateLabels();
		}
		else if(e.getSource().equals(addActionButton)){
			ClassificationAction ca = new ClassificationAction(null,null);//can be null, its just for temporary storage purposes
			SprClassificationActionPanel sprCaPanel = new SprClassificationActionPanel(ca);
			int res = JOptionPane.showOptionDialog(frame, sprCaPanel,"Classification options : addition of new classification action", JOptionPane.OK_CANCEL_OPTION, 
					JOptionPane.INFORMATION_MESSAGE, null,null,null);
			if(res == JOptionPane.OK_OPTION && workflow != null){
				ClassificationAction action = sprCaPanel.getClassificationAction();
				if(action!=null){
					workflow.addClassificationAction(action);
					List<ClassificationAction> actions = workflow.getClassificationActions();
					actionsList.setListData(createActionsArray(actions));
					savedWorkflow = false;
					updateLabels();
				}
			}
		}
		else if(e.getSource().equals(removeActionButton)){
			int t = actionsList.getSelectedIndex();
			if(t>=0){
				List<ClassificationAction> actions = workflow.getClassificationActions();
				actions.remove(t);
				workflow.setClassificationActions(actions);
				actionsList.setListData(createActionsArray(actions));
				savedWorkflow = false;
				updateLabels();
			}
		}
		else if(e.getSource().equals(editActionButton)){
			int t = actionsList.getSelectedIndex();
			if(t>=0){
				ClassificationAction ca = workflow.getClassificationActions().get(t);
				SprClassificationActionPanel sprCaPanel = new SprClassificationActionPanel(ca);
				int res = JOptionPane.showOptionDialog(frame, sprCaPanel,"Classification options : edit classification action", JOptionPane.OK_CANCEL_OPTION, 
						JOptionPane.INFORMATION_MESSAGE, new ImageIcon(),null,null);
				if(res==JOptionPane.OK_OPTION){
					ClassificationAction action = sprCaPanel.getClassificationAction();
					if(action!=null){
						workflow.getClassificationActions().set(t,action);
						List<ClassificationAction> actions = workflow.getClassificationActions();
						actionsList.setListData(createActionsArray(actions));
						savedWorkflow = false;
						updateLabels();
					}
				}
			}
		}
		updateLabels();
	}
	
	
	/**
	 * Creates an array of classification action names from a list
	 * of classification actions. Necessary for displaying classification
	 * actions in a jlist.
	 * @param caList The list with classification actions.
	 * @return The array with the names of the classification actions. 
	 */
	private String[] createActionsArray(List<ClassificationAction> caList){
		String[] res = new String[caList.size()];
		for(int i=0;i<caList.size();i++){
			res[i]=caList.get(i).getName();
		}
		return res;
	}
	

	public void windowClosed(WindowEvent e){}
	public void windowClosing(WindowEvent e){
		showExitDialog();
	}
	public void windowActivated(WindowEvent e){}
	public void windowDeactivated(WindowEvent e){}
	public void windowDeiconified(WindowEvent e){}
	public void windowIconified(WindowEvent e){}
	public void windowOpened(WindowEvent e){}
	
	/**
	 * Shows the exit dialog. If yes is selected, the program is terminated without
	 * and error code.
	 */
	private void showExitDialog(){
		String message = "Exit (Y/N)? ";
		String title   = "Exit?";
		int res = JOptionPane.showConfirmDialog(frame,message,title,JOptionPane.YES_NO_OPTION );
		if(res==JOptionPane.YES_OPTION)
			System.exit(0);		
	}
	

}
