/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

package main;

import java.awt.Color;
import java.awt.Dimension;
import java.awt.Font;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.GridLayout;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.awt.event.MouseMotionListener;
import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.util.ArrayList;
import java.util.List;

import javax.swing.BorderFactory;
import javax.swing.JButton;
import javax.swing.JFileChooser;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JMenu;
import javax.swing.JMenuBar;
import javax.swing.JMenuItem;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JTextField;
import javax.swing.filechooser.FileFilter;

/**
 * This class is used for testing purposes of FunSiP only. 
 * It is currently only configured for donor/acceptor display of predicted
 * splice sites, and then only with certain file formats. 
 * 
 * It is just a program that displays all the resulting values of predictions, 
 * and also displays the associated gene annotations. This can be used
 * to see whether or not the predictions make sense.
 * 
 * @author Michiel Van Bel
 */
public class FunSiPResultGui implements ActionListener{

	private JFrame frame				= null;
	private JFrame extraFrame			= null;
	private double currentZoom 			= 1.0;
	private JMenuItem openMenuItem		= null;
	private JMenuItem zoomOutMenuItem 	= null;
	private JMenuItem zoomInMenuItem	= null;
	private JPanel pan					= null;
	private JTextField correctFileText	= null;
	private JButton findCorrectFileButton 		= null;
	private String correctFileName		= null;
	
	/**
	 * Launches the program
	 * @param args The commandline parameters.
	 */
	public static void main(String [] args){
		if(args.length==0){
			new FunSiPResultGui();
		}
		else{
			if(args.length!=3){
				System.out.println("Usage: java main.FunSiPResultGui <PredictionsFile> <AnnotedFile> <name>");
			}
			else{
				new FunSiPResultGui(args[0],args[1],args[2]);
			}
		}
	}
	
	
	/**
	 * The constructor for the program. Creates the framework 
	 * and calls the necessary fileparsers.
	 * @param fileName The name of the file which contains the predictions.
	 * @param coordsFileName The name of the file which contains the annoted genes.
	 */
	public FunSiPResultGui(String fileName,String coordsFileName,String name){
		
		File file = new File(fileName);
		List<ResultContent> content = getContent(file);
		
		File coordsFile = new File(coordsFileName);
		List<List<Integer>> sites = getCoords(coordsFile,file);
		System.out.println("# of genes : "+sites.size());
		
		frame = new JFrame("Visualization of results : "+name);
		pan = new VisualContentPanel(content,sites);
		
		Dimension d = new Dimension (800,450);
		//frame.setMinimumSize(d);	//in order to be able to resize, ignore these settings
		//frame.setMaximumSize(d);	//in order to be able to resize, ignore these settings
		frame.setPreferredSize(d);
		frame.setJMenuBar(createMenuBar());
		
		frame.setContentPane(new JScrollPane(pan));
		
		JFrame.setDefaultLookAndFeelDecorated(true);
		frame.setVisible(true);
		frame.pack();
		frame.setDefaultCloseOperation(JFrame.DISPOSE_ON_CLOSE);
		frame.setLocationRelativeTo(null);	
	}
	
	public FunSiPResultGui(){
		frame = new JFrame("Visualization of results ");
		Dimension d = new Dimension (800,450);
		frame.setMinimumSize(d);	//in order to be able to resize, ignore these settings
		//frame.setMaximumSize(d);	//in order to be able to resize, ignore these settings
		frame.setPreferredSize(d);
		JMenuBar menuBar = createMenuBar();
		frame.setJMenuBar(menuBar);
		frame.setContentPane(new JPanel());
		JFrame.setDefaultLookAndFeelDecorated(true);
		frame.pack();
		frame.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
		frame.setLocationRelativeTo(null);
		frame.setVisible(true);
		JOptionPane.showMessageDialog(frame, provideCorrectPredictionsFilePanel());
		correctFileName = correctFileText.getText();
	}
	
	private JPanel provideCorrectPredictionsFilePanel(){
		JPanel panel = new JPanel();
		panel.setMinimumSize(new Dimension(400,200));
		panel.setPreferredSize(new Dimension(400,200));
		panel.setLayout(new GridLayout(2,1));
		JPanel upPanel = new JPanel(new GridLayout(2,1));
		upPanel.add(new JLabel("<html>Please provide the file with the correct predictions.</html>"));
		upPanel.add(new JLabel("The most common file-extension is *.egn"));
		JPanel downPanel = new JPanel();
		correctFileText = new JTextField(20);
		findCorrectFileButton = new JButton("Find file");
		findCorrectFileButton.addActionListener(this);
		downPanel.add(correctFileText);
		downPanel.add(findCorrectFileButton);
		panel.add(upPanel);
		panel.add(downPanel);
		return panel;
	}
	
	
	private JMenuBar createMenuBar(){
		JMenuBar menubar 			= new JMenuBar();
		JMenu fileMenu 				= new JMenu("File");
		JMenu optionsMenu 			= new JMenu("Options");
		
		openMenuItem				= new JMenuItem("Open predictions file");
		zoomInMenuItem 	= new JMenuItem("Zoom in");
		zoomOutMenuItem	= new JMenuItem("Zoom out");
		
		openMenuItem.addActionListener(this);
		zoomInMenuItem.addActionListener(this);
		zoomOutMenuItem.addActionListener(this);
		
		optionsMenu.add(zoomInMenuItem);
		optionsMenu.add(zoomOutMenuItem);
		fileMenu.add(openMenuItem);
		
		menubar.add(fileMenu);
		menubar.add(optionsMenu);
		
		
		return menubar;
	}
	
	public void actionPerformed(ActionEvent e){
		if(e.getSource().equals(zoomOutMenuItem)){
			currentZoom = currentZoom/2;
			pan.repaint();
		}
		else if(e.getSource().equals(zoomInMenuItem)){
			currentZoom = currentZoom*2;
			pan.repaint();
		}
		else if(e.getSource().equals(findCorrectFileButton)){
			JFileChooser fc = new JFileChooser("./");
			int res = fc.showOpenDialog(frame);
			if(res==JFileChooser.APPROVE_OPTION){
				correctFileText.setText(fc.getSelectedFile().getPath());
			}
		}
		else if(e.getSource().equals(openMenuItem)){
			JFileChooser fc = new JFileChooser("./");
			fc.setFileFilter(new FunSiPFilter());
			int res = fc.showOpenDialog(frame);
			if(res==JFileChooser.APPROVE_OPTION){
				File file =fc.getSelectedFile();
				frame.setTitle("Visualization of results "+file);
				List<ResultContent> content = getContent(file);
				File coordsFile = new File(correctFileName);
				List<List<Integer>> sites = getCoords(coordsFile,file);
				//List<List<Integer>> doubleSites = sites.
				if(extraFrame!=null)
					extraFrame.dispose();
				showExtraFrame(content,sites,file.getName());
				pan = new VisualContentPanel(content,sites);
				frame.setContentPane(new JScrollPane(pan));
				
				frame.validate();
				pan.repaint();
				frame.repaint();
			}
		}
	}
	
	private void showExtraFrame(List<ResultContent>content,List<List<Integer>>sites,String fileName){
		//tries to display extra information concerning the loaded results.
		extraFrame = new JFrame("Statistics for the evaluation of "+fileName);
		Dimension d = new Dimension(450,600);
		extraFrame.setPreferredSize(d);
		extraFrame.setMinimumSize(d);
		extraFrame.setMaximumSize(d);
		
		JPanel contentPane = new JPanel(new GridLayout(sites.size(),1));
		int done = 0;
		for(int i=0;i<sites.size();i++){
			List<Integer> sites2 = sites.get(i);
//			System.out.println("inner "+i+" : "+sites2.size());
			if(sites2.size()>2){
				JPanel panelLocal = createStatPanel(content,sites2);
				contentPane.add(panelLocal);
				done++;
			}
		}
		//System.out.println("Num panels : "+done);
		
		JFrame.setDefaultLookAndFeelDecorated(true);
		extraFrame.setContentPane(new JScrollPane(contentPane));
		extraFrame.pack();
		extraFrame.setDefaultCloseOperation(JFrame.DO_NOTHING_ON_CLOSE);
		extraFrame.setVisible(true);
		extraFrame.validate();
	}
	
	private JPanel createStatPanel(List<ResultContent> content,List<Integer>sites){
		JPanel panel = new JPanel();
		Dimension d = new Dimension(340,250);
		panel.setPreferredSize(d);
		panel.setMinimumSize(d);
		panel.setMaximumSize(d);
		panel.setBorder(BorderFactory.createEtchedBorder());
		panel.setLayout(new GridLayout(9,1));
		
		double numDonors 			= 0;
		double numAcceptors 		= 0;
		double numFoundDonors 		= 0;
		double numFoundAcceptors	= 0;
		double numNoSiteDonors		= 0;
		double numNoSiteAcceptors	= 0;
		double numFalseDonors		= 0;
		double numFalseAcceptors	= 0;
		
		int test = sites.get(0);
				
		if(test>0){
			for(int i=1;i<sites.size()-1;i++){
				int don = sites.get(i);
				int acc = sites.get(++i);
				numDonors++;
				numAcceptors++;
				boolean foundDon = false;
				boolean foundAcc = false;
				for(int j=0;j<content.size();j++){
					ResultContent rc = content.get(j);
					String s = rc.type;
					int pos = rc.position;
					double val = rc.score;
					
					if(pos==(don+1) && s.equals("donor")){
						foundDon = true;
						if(val>0)
							numFoundDonors++;
					}
					if(pos==(acc-1) && s.equals("acceptor")){
						foundAcc = true;
						if(val>0)
							numFoundAcceptors++;
					}
					if(pos>don&&pos>acc)
						break;
				}
				if(!foundDon)
					numNoSiteDonors++;
				if(!foundAcc)
					numNoSiteAcceptors++;
			}
		}
		else{
			for(int i=sites.size()-2;i>=1;i--){
				int don = sites.get(i)*-1;
				int acc = sites.get(--i)*-1;
				numDonors++;
				numAcceptors++;
				boolean foundDon = false;
				boolean foundAcc = false;
				for(int j=content.size()-1;j>0;j--){
					ResultContent rc =content.get(j);
					String s = rc.type;
					int pos = rc.position;
					double val = rc.score;
					if(pos==(don-1) && s.equals("donor_rev")){
						foundDon = true;
						if(val>0)
							numFoundDonors++;
					}
					if(pos==(acc+1) && s.equals("acceptor_rev")){
						foundAcc = true;
						if(val>0)
							numFoundAcceptors++;
					}
					if(pos<don && pos<acc)
						break;
				}
				if(!foundDon)
					numNoSiteDonors++;
				if(!foundAcc)
					numNoSiteAcceptors++;
			}
			
		}
		if(numNoSiteDonors!=0)
			System.out.println("Wrong site donors : "+numNoSiteDonors);
		if(numNoSiteAcceptors!=0)
			System.out.println("Wrong site acceptors : "+numNoSiteAcceptors);
		
		if(test>0){
			int start = sites.get(0);
			int stop  = sites.get(sites.size()-1);

			for(int i=0;i<content.size();i++){
				ResultContent rc = content.get(i);
				double val = rc.score;
				String s = rc.type;
				int pos = rc.position;
				if(pos>start && pos<stop){
					
					if(s.equals("donor")&&val>0)
						numFalseDonors++;
					if(s.equals("acceptor")&&val>0)
						numFalseAcceptors++;
				}
			}
			numFalseDonors = numFalseDonors - numFoundDonors;
			numFalseAcceptors = numFalseAcceptors - numFoundAcceptors;
		}
		else{
			int start = sites.get(0)*-1;
			int stop = sites.get(sites.size()-1)*-1;
			System.out.println(start+" :: "+stop);
			if(start>stop){
				int temp = start;
				start = stop;
				stop = temp;
			}
			System.out.println(start+" :: "+stop);
			
			for(int i=0;i<content.size();i++){
				ResultContent rc = content.get(i);
				double val = rc.score;
				String s = rc.type;
				int pos = rc.position;
				if(pos>start && pos<stop){
					
					if(s.equals("donor_rev")&&val>0)
						numFalseDonors++;
					if(s.equals("acceptor_rev")&&val>0)
						numFalseAcceptors++;
				}
			}
			numFalseDonors = numFalseDonors - numFoundDonors;
			numFalseAcceptors = numFalseAcceptors - numFoundAcceptors;
			
		}
		
		panel.add(new JLabel("Gene from : "+sites.get(0)+" to "+sites.get(sites.size()-1)));
		JLabel tempDon = new JLabel();
		JLabel tempAcc = new JLabel();
		String td = "Donors : ";
		String ta = "Acceptors : ";
		if(test>0){
			for(int i=1;i<sites.size()-1;i++){
				td+=sites.get(i)+" ... ";
				ta+=sites.get(++i)+" ... ";
			}
		}
		else{
			for(int i=sites.size()-2;i>=1;i--){
				td+=sites.get(i)+" ... ";
				ta+=sites.get(--i)+" ... ";
			}
		}
		tempDon.setText(td);
		tempAcc.setText(ta);
		panel.add(tempDon);
		panel.add(tempAcc);
		
		double accRate = (numFoundAcceptors/numAcceptors)/(numFalseAcceptors+1);
		double donRate = (numFoundDonors/numDonors)/(numFalseDonors+1);
		
		panel.add(new JLabel("#true acceptors = "+(int)numAcceptors+" ||| #found acceptors = "+(int)numFoundAcceptors+" ||| #missed acceptors = "+(int)(numAcceptors-numFoundAcceptors)));
		panel.add(new JLabel("#true donors = "+(int)numDonors+" ||| #found donors = "+(int)numFoundDonors+" ||| #missed donors = "+(int)(numDonors-numFoundDonors)));
		panel.add(new JLabel("#false acceptors within gene range = "+(int)numFalseAcceptors));
		panel.add(new JLabel("#false donors within gene range = "+(int)numFalseDonors));
		
		panel.add(new JLabel("=>Acceptor rate = "+accRate));
		panel.add(new JLabel("=>Donor rate = "+donRate));
		
		return panel;
	}
	
	private class FunSiPFilter extends FileFilter{
		public FunSiPFilter(){
		}
		public boolean accept(File f){
			if(f.getName().endsWith("FunSiP_AD")||f.isDirectory())
				return true;
			else
				return false;
		}
		public String getDescription(){
			return "Funsip output files";
		}
	}
	
	/**
	 * Gets the coordinates of the annoted file. 
	 * @param file The file with the correct annotation
	 * @param f The file with the predicted values
	 * @return List with locations of the annoted gene splice sites.
	 */
	private List<List<Integer>> getCoords(File file,File f){
		String name = f.getName();
		name = name.substring(0,name.indexOf(".FunSiP_AD"));
		
		List<List<Integer>> list = new ArrayList<List<Integer>>();
		try{
			BufferedReader reader = new BufferedReader(new FileReader(file));
			String s = reader.readLine();
			while(s!=null){
				String[] split = s.split(" +");
				if(split[0].equals(name)){
					List<Integer> coords = new ArrayList<Integer>();
					for(int i=1;i<split.length;i++){
						int t = Integer.parseInt(split[i]);
						coords.add(t);
					}
					list.add(coords);
				}
				s = reader.readLine();
			}
		}
		catch(Exception exc){
			exc.printStackTrace();
			System.exit(1);
		}
		return list;
	}
	
	
	/**
	 * Get the content of the predictions file.
	 * @param file The file with the predictions.
	 * @return List which consists of objects that contain both the position and description
	 * of a potential splice site.
	 */
	private List<ResultContent> getContent(File file){
		List<ResultContent> res = new ArrayList<ResultContent>();
		try{
			BufferedReader reader = new BufferedReader(new FileReader(file));
			String s = reader.readLine();
			while(s!=null){
				String[] split = s.split(" ");
				ResultContent rc = new ResultContent();
				rc.position = Integer.parseInt(split[0]);
				rc.type = split[1];
				rc.score = Double.parseDouble(split[2]);
				res.add(rc);
				s = reader.readLine();
			}
			reader.close();
		}
		catch(Exception exc){
			exc.printStackTrace();
			System.exit(1);
		}
		return res;
	}
	
	/**
	 * private class, uses for visualisation.
	 * @author Michiel Van Bel
	 */
	private class VisualContentPanel extends JPanel implements MouseMotionListener, MouseListener{
		public static final long serialVersionUID = 14445654l;
		private List<ResultContent> list = null;
		private List<List<Integer>> coords = null;
		private JLabel dispLabel = null;
		
		public void mouseMoved(MouseEvent e){
			this.setToolTipText(e.getX()+"");
		}
		public void mouseDragged(MouseEvent e){
			
		}
		
		private int firstClickedX = 0;
		private boolean firstClicked = true;
		private String selectedFeature = null;
		public void mouseClicked(MouseEvent e){
			int x = e.getX();
			int y = e.getY();
			boolean accepted = false;
			String sel = null;
			for(int i=0;i<list.size();i++){
				ResultContent rc = list.get(i);
				int loc = rc.position;
				String type = rc.type;
				if(type.endsWith("_rev")){
					if(y>210&&x==loc){
						accepted = true;
						sel = type;
					}
				}
				else{
					if(y<210&&x==loc){
						
						sel = type;
						accepted = true;
					}
				}
			}
			if(accepted){
				if(firstClicked){
					selectedFeature = sel;
					firstClickedX = x;
					dispLabel.setText("Selected : "+selectedFeature+" on location "+x);
					dispLabel.repaint();
					firstClicked = !firstClicked;
				}
				else{
					if(selectedFeature.equals(sel)){
						int distance = (int)Math.abs(x-firstClickedX);
						dispLabel.setText("Distance between 2 "+sel+"'s : "+distance);
						dispLabel.repaint();
						firstClicked = !firstClicked;
					}
					else{
						
					}
				}
			}
				
			/*	
				
				if(x==loc){
					pos = loc;
					
				}
				if(firstClicked){
					selectedFeature = type;
					dispLabel.setText("Selected type : "+selectedFeature);
					dispLabel.repaint();
				}
			}
			if(firstClicked)
				firstClickedX = pos;
			else
				secondClickedX = pos;
			
			if(!firstClicked){
				
				JOptionPane.showMessageDialog(this, new JLabel("Distance : "+((int)Math.abs(secondClickedX-firstClickedX))),"Distance between locations",JOptionPane.INFORMATION_MESSAGE);
			}*/
		}
		
		public void mouseEntered(MouseEvent e){
		}
		public void mousePressed(MouseEvent e){
		}
		public void mouseReleased(MouseEvent e){
		}
		public void mouseExited(MouseEvent e){
		}
		
		private void createDisplayFrame(){
			JFrame frame = new JFrame("Options");
			frame.setVisible(true);
			JPanel dispPanel = new JPanel();
			Dimension d = new Dimension(250,150);
			frame.setPreferredSize(d);
			frame.setMaximumSize(d);
			frame.setMinimumSize(d);
			dispLabel = new JLabel();
			dispPanel.add(dispLabel);
			frame.setContentPane(dispPanel);
			frame.pack();
		}
		
		public VisualContentPanel(List<ResultContent> list,List<List<Integer>> coords){
			super();
			this.list = list;
			this.coords = coords;
			int max = list.get(list.size()-1).position;
			Dimension d = new Dimension(max,400);
			this.setMinimumSize(d);
			this.setMaximumSize(d);
			this.setPreferredSize(d);
			this.addMouseMotionListener(this);
			this.addMouseListener(this);
			createDisplayFrame();
		}
		
		public void paintComponent(Graphics g){
			super.paintComponent(g);
			Graphics2D g2d = (Graphics2D)g;
			g2d.scale(currentZoom,currentZoom);
			int max = list.get(list.size()-1).position;
			
			//Color acc_true = Color.RED;
			//Color don_true = Color.MAGENTA;
			Color acc_pos = Color.GREEN;
			Color don_pos = Color.BLUE;
			
			//draw the 3 dividing lines
			g2d.drawLine(0,100,max,100);
			g2d.drawLine(0,200,max,200);
			g2d.drawLine(0,210,max,210);
			g2d.drawLine(0,310,max,310);
			for(int i=0;i<max;i+=50){
				g2d.drawLine(i,200,i,190);
				g2d.drawLine(i,210,i,220);
				g2d.drawString(""+i,i-10, 210);
			}
			

			g2d.setColor(Color.black);
			
			g2d.setFont(new Font(null,Font.BOLD,16));
			g2d.drawString("0",0, 107);
			g2d.drawString("0",0, 317);
			g2d.drawString("-->", 0, 196);
			g2d.drawString("<--", 0, 220);
			g2d.setColor(acc_pos);g2d.drawString("Acceptor : green + red",0,30);
			g2d.setColor(don_pos);g2d.drawString("Donor : blue + black" , 0,50);
			g2d.setColor(Color.black);
			
			
			//iterate over the solution space
			for(int i=0;i<list.size();i++){
				ResultContent rc = list.get(i);
				int xpos = rc.position;
				String s = rc.type;
				double d = rc.score;
		
				String temp = null;
				int increase = 0;
				boolean rev = false;
				
				if(s.endsWith("_rev")){
					temp = s.substring(0,s.indexOf("_"));
					increase = 310;
					rev =true;
				}
				else{
					temp = s;
					increase = 100;
				}
				int ypos = increase - (int)(50*d);
				//int ypos = 100 - (int)(50*d);
				
				if(temp.equals("donor")){
					g2d.setColor(don_pos);
					
					if(rev){

						g2d.fillRect(xpos, ypos, 4, 410-ypos);
						
					}
					else{
						g2d.fillRect(xpos,ypos, 4, 200-ypos);

					}
				}
				else if(temp.equals("acceptor")){
					g2d.setColor(acc_pos);
					
					if(rev){
						g2d.fillRect(xpos,ypos,4,410-ypos);

					}
					else{
						g2d.fillRect(xpos,ypos, 4, 200-ypos);
					}
				}
				
			}
			
			/*
			 * True predictions
			 */
			for(int i=0;i<coords.size();i++){
				List<Integer>gene_coords = coords.get(i);
				if(gene_coords.size()<=2){
					//do nothing :)
				}
				else{
					int test = gene_coords.get(0);
					if(test>0){
						for(int j=1;j<gene_coords.size()-1;j++){
							int dp = gene_coords.get(j);
							int ap = gene_coords.get(++j);
						
							g2d.setColor(Color.MAGENTA);
							g2d.drawRect(dp,0, 4,200);
						//	g2d.setColor(Color.BLACK);
							g2d.fillRect(dp,0,4,20);
						
							g2d.setColor(Color.RED);
							g2d.drawRect(ap-2,0,4,200);
						//	g2d.setColor(Color.BLACK);
							g2d.fillRect(ap-2,0,4,20);
						}
					}
					else{
						for(int j=gene_coords.size()-2;j>=1;j--){
							int dp = gene_coords.get(j)*-1 -1;
							int ap = gene_coords.get(--j)*-1 +1;
							g2d.setColor(Color.MAGENTA);
							g2d.drawRect(dp,210, 4,200);
						//	g2d.setColor(Color.BLACK);
							g2d.fillRect(dp,210,4,20);
						
							g2d.setColor(Color.RED);
							g2d.drawRect(ap-2,210,4,200);
						//	g2d.setColor(Color.BLACK);
							g2d.fillRect(ap-2,210,4,20);
						}
					}
				}
			}
		}
		
		
	}
	

	
	/**
	 * private class, used for storing content of a prediction file.
	 * @author Michiel Van Bel
	 */
	private class ResultContent{
		public int position;
		public String type;
		public double score;
	}
	
}
