/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
package util;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.SortedSet;
import java.util.TreeSet;

/**
 * Class with different methods for organizing, re-organizing, extracting and normalizing
 * data from standard datastructures used in the FunSiP program package.
 * @author Michiel Van Bel
 */
public class Util {

		
	/**
	 * Bubblesort algorithm which sorts one list, and sorts (changing of indices) another
	 * list in exactly the same way.
	 * @param toSortList original list to be sorted.
	 * @param followList The list which will be sorted by the same indices of toSortList.
	 */
	public static void doubleSort(double[] toSortList, double[] followList){
		if(toSortList==null || followList==null || toSortList.length!=followList.length)
			return;
		for(int i=0;i<toSortList.length;i++){
			for(int j=i+1;j<toSortList.length;j++){
				double t1 = toSortList[i];
				double t2 = toSortList[j];
				if(t1>t2){
					toSortList[j] = t1;
					toSortList[i] = t2;
					double temp = followList[j];
					followList[j] = followList[i];
					followList[i] = temp;
				}
			}
		}
		
	}
	

	/**
	 * Normalizes the values of a map to the interval [0,1].
	 * It is assumed that the values in the original map are all positive, 
	 * and that the given max value is indeed the maximum.
	 * @param map The map with data to be normalized. Strings are the keys, ints are the values
	 * @param max The maximum int value in the map
	 * @return A map with normalized data, String are the keys and doubles are the values
	 */
	public static Map<String,Double> normalize(Map<String,Integer> map,int max){
		Map<String,Double> result = new HashMap<String,Double>();
		Set<String> s = map.keySet();
		Iterator<String> it = s.iterator();
		while(it.hasNext()){
			String string = it.next();
			double temp = map.get(string);
			double res = temp/max;
			res = (double)((int)(res*10000))/10000;
			result.put(string, res);
		}
		return result;
	}
	
	
	/**
	 * Merges 2 lists of integer data to one list of string data. 
	 * @param upData The data from upstream the splicesite
	 * @param downData The data from downstream the splicesite
	 * @return A merged datalist
	 */
	public static List<String> mergeUpDownInt(List<Integer> upData, List<Integer> downData){
		List<String> result = new ArrayList<String>();
		for(int i=0;i<upData.size();i++)
			result.add(upData.get(i)+"");
		for(int i=0;i<downData.size();i++)
			result.add(downData.get(i)+"");
		return result;
	}
	
	/**
	 * This method changes the strand of a sequence by both replacing the nucleotides
	 * by their complements and by inversing the order of the sequence.
	 * @param sequence The original (most likely reverse strand) sequence
	 * @return The new sequence on the (most likely forward) other strand
	 */
	public static String toForwardStrand(String sequence){
		String temp = sequence;
		temp = temp.toUpperCase();
		StringBuffer res = new StringBuffer();
		for(int i=0;i<temp.length();i++){
			char c = temp.charAt(temp.length()-1-i);
			switch(c){
			case 'A' : res.append('T'); break;
			case 'T' : res.append('A'); break;
			case 'C' : res.append('G'); break;
			case 'G' : res.append('C'); break;
			default : res.append('N'); break;
			}
		}
		String result = new String(res);
		return result;
	}
	
	/**
	 * Extracts a list of strings from the values of a map,
	 * but it keeps the normal alphabetical order between the keys of the map.
	 * @param map The map, with strings as keys and ints as values
	 * @return A list with the keys from the map being represented by strings
	 */
	public static List<String> extractFromMap(Map<String,?>map){
		List<String>result = new ArrayList<String>();
		SortedSet<String> set = new TreeSet<String>(map.keySet());
		Iterator<String> it = set.iterator();
		while(it.hasNext()){
			String t = it.next();
			Object temp = map.get(t);
			result.add(""+temp);
		}
		return result;
	}
}
