/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
package util.classificationFeatures.implementations;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import util.classificationFeatures.AttributeDescription;
import util.classificationFeatures.ClassificationFeature;
import util.classificationFeatures.ClassificationFeatureDrawing;
import util.classificationFeatures.ClassificationFeatureUtils;
import util.classificationFeatures.FeatureExtractionException;
import util.classificationFeatures.IncorrectClassificationFeatureException;
import util.classificationFeatures.SeconderyStructureClassificationFeature;
import util.conversions.Conversion;

public class CompositionalSecondaryStructureClassificationFeature implements SeconderyStructureClassificationFeature {
	private int length = 0;
	private int upstream_range = 0;
	private int downstream_range = 0;
	
	private String[] fullTypes = null;
	private String[] reducedTypes = null;

	public CompositionalSecondaryStructureClassificationFeature(){
	}
	
	/**
	 * This method asks for the up- and downstreamranges of the feature
	 * @return The up/down ranges
	 */
	public int[] getUpDownRanges(){
		int [] res = {upstream_range,downstream_range};
		return res;
	}
	
	/**
	 * This method asks the classificationfeature for its conversion-type.
	 * @return The used conversion type. Null of not appropriate.
	 */
	public Conversion getConversion(){
		return null;
	}
	
	
	/**
	 * This method sets the new upstream and downstream ranges  
	 * @param up The new upstream range
	 * @param down The new downstream range
	 */
	public void setUpDownRanges(int up, int down){
		this.upstream_range = up;
		this.downstream_range = down;
	}

	/**
	 * Asks whether the computer needs to precompute compositional types.
	 * When a conversion is taken with a large alphabet and/or a large length,
	 * this computation can take WAAAY to long (and take too much memory as well).
	 * Therefore, prior to this operation, this flag should be checked.
	 * @return Whether or not to supply compositional types of certain length.
	 */ 
	public boolean needCompositionalTypes(){
		return false;
	}
	
	
	/**
	 * This method extracts the features from a sequence (relative to the splicesite)
	 * and returns those features in a list of double-values (standardized on double
	 * values because of its generic but still small-sized nature).
	 * @param sequences The sequence from which the features should be extracted.
	 * @param splicesite The location of the splicesite (pseudo or real)
	 * @param types The compositional types (if requested, otherwise null).
	 * @param conversions A list with all the conversion types
	 * @param conversion_sequences  An array with all the converted sequences.
	 * @return A list with the features
	 * @throws FeatureExtractionException Thrown when something goes wrong when extracting the features.
	 */
	public List<Double> extractFeatures(String sequences[], int splicesite,String []types, List<Conversion> conversions,String[]conversion_sequences) throws FeatureExtractionException{
		if(sequences==null||sequences.length<3||sequences[0]==null||sequences[1]==null||sequences[2]==null)
			throw new FeatureExtractionException("Stringsequence(s) is/are null or could not be retrieved.");
		String prim_seq 	= sequences[0];
		String sec_struct	= sequences[1];
		if(prim_seq.length()!=sec_struct.length())
			throw new FeatureExtractionException("primary and secondary string lengths are not equal");
		
		//Double energy		= Double.parseDouble(sequences[2]);
		
		List<Double> result = new ArrayList<Double>();
		Map<String,Integer> upmap = new HashMap<String,Integer>();
		Map<String,Integer> downmap = new HashMap<String,Integer>();
		
		
		//were using reduced types, so convert string also
		String[] sec_types = this.getFullTypes();
		//String new_sec_struct = ClassificationFeatureUtils.convertFullSecToReducedSec(sec_struct);
		
		for(int i=0;i<sec_types.length;i++){
			upmap.put(sec_types[i],0);
			downmap.put(sec_types[i], 0);
		}
		
		int start = splicesite-upstream_range;
		int stop = splicesite+downstream_range-length;
		
		for(int i=start;i<splicesite-length;i++){
			String subseq = sec_struct.substring(i,i+length);
			Integer res = upmap.get(subseq);
			if(res==null || res == -1){}
			else{
				res++;
				upmap.put(subseq,res);
			}
		}
		for(int i=splicesite;i<stop;i++){
			String subseq = sec_struct.substring(i,i+length);
			Integer res = downmap.get(subseq);
			if(res==null||res==-1){}
			else{
				res++;
				downmap.put(subseq, res);
			}
		}
		
		for(int i=0;i<sec_types.length;i++){
			double d = upmap.get(sec_types[i]);
			result.add(d);
		}
		for(int i=0;i<sec_types.length;i++){
			double d = downmap.get(sec_types[i]);
			result.add(d);
		}
		
		return result;
	}
	
	/**
	 * Returns the string representation of the classificationfeature,
	 * but with all possible spaces and tabs replaced by underscores.
	 * @return Alternative string representation
	 */
	public String toStringUnderscore(){
		return "COMPOSITIONAL_SS_"+upstream_range+" "+downstream_range;
	}
	
	
	/**
	 * String representation of the classificationFeature
	 * @return String representation of the classificationfeature
	 */
	public String toString(){
		return "COMPOSITIONAL SS "+upstream_range+" "+downstream_range;
	}
	
	/**
	 * String representation of the classificationFeature that contains
	 * the core elements, and discards the upstream/downstream range
	 * elements.
	 * @return Shortened version of the string representation of the classificationfeature
	 */
	public String toNoRangeString(){
		return "COMPOSITIONAL SS "+length;
	}
	
	/**
	 * Returns a standardized description of the location that is indicated.
	 * e.g. 5 bp upstream of splicesite. In case the location is outside the range
	 * of the feature, an error string is returned.
	 * @param loc The location
	 * @return The description
	 */
	public AttributeDescription getLocationDescription(int loc){
		AttributeDescription description = new AttributeDescription();
		description.type = ClassificationFeatureUtils.CLASSIFICATIONFEATURE_TYPES.SEC_POSITIONAL;
		return description;
	}
	
	/**
	 * Returns the number of attributes that are created by this classificationfeature,
	 * according to the internal parameters (e.g. up- and downstreamranges). When the 
	 * internal parameters are not set, or they are in an incorrect state, this method will
	 * return -1 as value
	 * @return The number of attributes created by this classificationfeature
	 */
	public int getNumberOfAttributes(){
		return 0;
	}
	
	/**
	 * This method sets the necessary parameters for the classificationfeature. 
	 * The parameters are all generically passed as an array of strings, which of course
	 * need to be decoded by the classificationfeature itself.
	 * @param args The array with the parameters in string format
	 * @throws IncorrectClassificationFeatureException Can be thrown when the parameters
	 * are incorrect (outside possible ranges) and/or the parameters could not be parsed.
	 */
	public void setParameters(String[] args) throws IncorrectClassificationFeatureException{
		if(args.length!=4)
			throw new IncorrectClassificationFeatureException("Incorrect number of parameters for sec struct compositional feature");
		String name = args[0].toUpperCase();
		if(!name.equals("CSS"))
			throw new IncorrectClassificationFeatureException("Incorrect naming for sec struct compositional feature");
		try{
			length = Integer.parseInt(args[1]);
			upstream_range = Integer.parseInt(args[2]);
			downstream_range = Integer.parseInt(args[3]);
		}
		catch(NumberFormatException exc){
			throw new IncorrectClassificationFeatureException("Integer parameters for sec struct compositional feature cannot be parsed correctly");
		}
	}
	
	
	
	
	public void setFullTypes(String[] fullTypes){
		this.fullTypes = fullTypes;
	}
	public String[] getFullTypes(){
		if(fullTypes!=null)
			return fullTypes;
		else{
			fullTypes =ClassificationFeatureUtils.generateFullSecStructCompositionalTypes(length); 
			return fullTypes;
		}
	}
	
	public void setReducedTypes(String[] reducedTypes){
		this.reducedTypes = reducedTypes;
	}
	
	public String[] getReducedTypes(){
		if(reducedTypes!=null)
			return reducedTypes;
		else{
			reducedTypes = ClassificationFeatureUtils.generateReducedSecStructCompositionalTypes(length);
			return reducedTypes;
		}
	}
	
	/**
	 * Because all classificationfeatures need different algorithms to draw them,
	 * according to the specifications, these utilities are also moved to the 
	 * implementations of the ClassificationFeatures. However, as not to clutter them,
	 * we have chosen to place this data in a seperate interface. The programmer thus
	 * has a choice of using one or two classes to implement both interfaces.
	 * 
	 * @return The object which can draw the classification features.
	 */
	public ClassificationFeatureDrawing getClassificationFeatureDrawing(){
		return new CompositionalSecondaryStructureClassificationFeatureDrawing(this);
	}
	
	public int getDownstream_range() {
		return downstream_range;
	}

	public int getUpstream_range() {
		return upstream_range;
	}
	public int getLength(){
		return length;
	}
	
	
	public void setDownstream_range(int downstream_range) {
		this.downstream_range = downstream_range;
	}


	public void setLength(int length) {
		this.length = length;
	}


	public void setUpstream_range(int upstream_range) {
		this.upstream_range = upstream_range;
	}
	
	/**
	 * Makes a copy of this particular classificationfeature. 
	 * The cloneable interface cannot be used directly because it is only applicable
	 * on classes, not interfaces.
	 * @return A copy of the classificationfeature
	 */
	public ClassificationFeature make_copy(){
		ClassificationFeature res = null;
		try{
			res = (ClassificationFeature)this.clone();
		}
		catch(Exception exc){}
		return res;
	}

}
