/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

package util.classificationFeatures.implementations.test;

import java.util.ArrayList;
import java.util.List;

import util.classificationFeatures.AttributeDescription;
import util.classificationFeatures.ClassificationFeature;
import util.classificationFeatures.ClassificationFeatureDrawing;
import util.classificationFeatures.FeatureExtractionException;
import util.classificationFeatures.IncorrectClassificationFeatureException;
import util.conversions.Conversion;

public class DeterminantClassificationFeature implements ClassificationFeature{

	
	public DeterminantClassificationFeature(){	
	}
	
	
	/**
	 * This method extracts the features from a sequence (relative to the splicesite)
	 * and returns those features in a list of double-values (standardized on double
	 * values because of its generic but still small-sized nature).
	 * @param sequences The sequence from which the features should be extracted.
	 * @param splicesite The location of the splicesite (pseudo or real)
	 * @param types The compositional types (if requested, otherwise null).
	 * @param conversions A list with all the conversion types
	 * @param conversion_sequences  An array with all the converted sequences.
	 * @return A list with the features
	 * @throws FeatureExtractionException Thrown when something goes wrong when extracting the features.
	 */
	public List<Double> extractFeatures(String sequences[], int splicesite, String[] types, List<Conversion> conversions,String[]conversion_sequences) throws FeatureExtractionException{
		List<Double> res = new ArrayList<Double>();
		String seq= sequences[0];
		int funSite = splicesite;
		
		//check max-lengths, being 4
		if((splicesite-2*4)<0){
			seq = "NNNNNNNN"+seq;
			funSite+=8;
		}
		if((splicesite+2*4)>seq.length()){
			seq = seq + "NNNNNNNN";
			//no change to funsite, appendage is to the end.
		}
		
		List<Double> t2 = getFeats(seq,funSite,2);
		List<Double> t3 = getFeats(seq,funSite,3);
		List<Double> t4 = getFeats(seq,funSite,4);
		
		res.addAll(t2);
		res.addAll(t3);
		res.addAll(t4);
	
		return res;
	}
	
	private List<Double> getFeats(String sequence,int site, int length){
		List<Double> result = new ArrayList<Double>();
		
		//we assume that the sequences are now of sufficient length
		String A = sequence.substring(site-2*length,site-length);
		String B = sequence.substring(site-length,site);
		String C = sequence.substring(site,site+length);
		String D = sequence.substring(site+length,site+2*length);
		
		double AC = getDeterminant(A,C,length);
		double AD = getDeterminant(A,D,length);
		double BC = getDeterminant(B,C,length);
		double BD = getDeterminant(B,D,length);
		
		result.add(AC);
		result.add(AD);
		result.add(BC);
		result.add(BD);
		return result;
	}
	
	private double getDeterminant(String s1,String s2,int l){
		double[][] res = new double[l][l];
		s1 = inverse(s1);
		for(int i=0;i<s1.length();i++){
			for(int j=0;j<s2.length();j++){
				char c1 = s1.charAt(i);
				char c2 = s2.charAt(j);
				double x = getBPScore(c1,c2);
				res[i][j] = x;
			}
		}
		
		double det = Determinant(res);
		return det;		
	}
	
	
	private String inverse(String a){
		StringBuffer res = new StringBuffer();
		for(int i=a.length()-1;i>=0;i--)
			res.append(a.charAt(i));
		return res.toString();
	}

	
	public static final double[][] BASEPAIRMATRIX = 
	{{1,1,4,6},
	{1,2,7,1},
	{4,7,2,5},
	{6,1,5,1}};
	
	/**
	 * Returns the place wihtin the basepairbinding matrix for
	 * a nucleotide
	 * @param c The nucleotide 
	 * @return The place within matrix BASEPAIRMATRIX
	 */
	public int getBPVal(char c){
		int res = 0;
		switch(c){
		case 'A': case 'a' : res = 0; break;
		case 'C': case 'c' : res = 1; break;
		case 'G': case 'g' : res = 2; break;
		case 'T': case 't' : case 'U' : case 'u' : res = 3; break;
		case 'N': res = -1; break;
		}
		return res;
	}
	
	public double getBPScore(char a,char b){
		int t1 = getBPVal(a);
		int t2 = getBPVal(b);
		if(t1<0||t2<0)
			return 0;
		else
			//return atomicMassRNA[t1]/atomicMassRNA[t2];
			return BASEPAIRMATRIX[t1][t2];
	}
	
	private int iDF = 0;
	public double[][] UpperTriangle(double[][] m){

	    double f1 = 0;   double temp = 0;
	    int tms = m.length;  // get This Matrix Size (could be smaller than global)
	    int v=1;
	    iDF=1;
	    for (int col=0; col < tms-1; col++)
	    {
	        for (int row=col+1; row < tms; row++)
	        {
	            v=1;
	    
	            outahere:
	            while(m[col][col] == 0)             // check if 0 in diagonal
	            {                                   // if so switch until not
	                if (col+v >= tms)               // check if switched all rows
	                {   iDF=0;
	                    break outahere; 
	                }
	                else
	                {
	                    for(int c=0; c < tms; c++)
	                    {  temp = m[col][c];
	                       m[col][c]=m[col+v][c];       // switch rows
	                       m[col+v][c] = temp;   
	                    }
	                    v++;                            // count row switchs
	                    iDF = iDF * -1;                 // each switch changes determinant factor
	                }
	            }

	            if ( m[col][col] != 0 ) {

	            	try {
	            		f1 = (-1) * m[row][col] / m[col][col];
	            		for (int i=col; i < tms; i++) { 
	            			m[row][i] = f1*m[col][i] + m[row][i]; 
	            		}
	            	}
	            	catch(Exception e) { e.printStackTrace(); }

	            }
	            
	        }
	    }

	    return m;
	}

//	--------------------------------------------------------------

	public double Determinant(double[][] matrix)
	{
	    int tms = matrix.length;
	    
	    double det=1;

	    matrix = UpperTriangle(matrix);
	    for (int i=0; i < tms; i++)
	    {  det = det * matrix[i][i]; }      // multiply down diagonal

	    det = det * iDF;                    // adjust w/ determinant factor

	    return det;
	}

	
	public boolean needCompositionalTypes(){
		return false;
	}
	
	/**
	 * This method asks the classificationfeature for its conversion-type.
	 * @return The used conversion type. Null of not appropriate.
	 */
	public Conversion getConversion(){
		return null;
	}
	
	
	
	
	public ClassificationFeature make_copy(){
		ClassificationFeature res = null;
		try{
			res = (ClassificationFeature)this.clone();
		}
		catch(Exception exc){}
		return res;
	}
	
	public String toStringUnderscore(){
		return "DET_CF";
	}
	
	public String toString(){
		return "DET_CF";
	}
	public String toNoRangeString(){
		return "DET_CF";
	}
	
	public int getNumberOfAttributes(){
		return 12;
	}
	
	public void setParameters(String[] args) throws IncorrectClassificationFeatureException{
		if(!args[0].equals("DET_CF")||args.length!=1){
			throw new IncorrectClassificationFeatureException("No DET_CF featre");
		}
	}
	
	public ClassificationFeatureDrawing getClassificationFeatureDrawing(){
		return null;
	}
	
	public int[] getUpDownRanges(){
		return null;
	}
	
	public void setUpDownRanges(int up, int down){
	}
	
	public int getLength(){
		return 0;
	}
	public AttributeDescription getLocationDescription(int loc){
		return null;
	}
}
