/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
package util.classificationFeatures.implementations.test;

import java.util.ArrayList;
import java.util.List;

import util.classificationFeatures.AttributeDescription;
import util.classificationFeatures.ClassificationFeature;
import util.classificationFeatures.ClassificationFeatureDrawing;
import util.classificationFeatures.FeatureExtractionException;
import util.classificationFeatures.IncorrectClassificationFeatureException;
import util.conversions.Conversion;

public class NearestEqualClassificationFeature implements ClassificationFeature{

	private int length = 0;
	private int upstream_range = 0;
	private int downstream_range = 0;
	
	public final static int POSSIBLE_SOLUTIONS = 3;
	
	public NearestEqualClassificationFeature(){
	}
	
	/**
	 * Asks whether the computer needs to precompute compositional types.
	 * When a conversion is taken with a large alphabet and/or a large length,
	 * this computation can take WAAAY to long (and take too much memory as well).
	 * Therefore, prior to this operation, this flag should be checked.
	 * @return Whether or not to supply compositional types of certain length.
	 */
	public boolean needCompositionalTypes(){
		return false;
	}
	
	/**
	 * This method extracts the features from a sequence (relative to the splicesite)
	 * and returns those features in a list of double-values (standardized on double
	 * values because of its generic but still small-sized nature).
	 * @param sequences The sequence from which the features should be extracted.
	 * @param splicesite The location of the splicesite (pseudo or real)
	 * @param types The compositional types (if requested, otherwise null).
	 * @param conversions A list with all the conversion types
	 * @param conversion_sequences  An array with all the converted sequences.
	 * @return A list with the features
	 * @throws FeatureExtractionException Thrown when something goes wrong when extracting the features.
	 */
	public List<Double> extractFeatures(String sequences[], int splicesite, String[] types, List<Conversion> conversions,String[]conversion_sequences) throws FeatureExtractionException{
		if(sequences==null||sequences.length<1||sequences[0]==null)
			throw new FeatureExtractionException("Stringsequence is null or could not be retrieved.");
		String sequence = sequences[0].toUpperCase();
		if(splicesite<0 || splicesite>sequence.length())
			throw new FeatureExtractionException("Splicesite is outside permitted range for positional feature");
		List<Double> result = new ArrayList<Double>();
		
		int finalUpRange = 0;
		int finalDownRange = 0;
		if(splicesite-upstream_range<0)
			finalUpRange = 0;
		else
			finalUpRange = splicesite-upstream_range;
		
		if(splicesite+downstream_range>sequence.length())
			finalDownRange = sequence.length();
		else
			finalDownRange = splicesite+downstream_range;
		
		//have 6 possible binding locations: 3 up, 3 down
		
		for(int i=finalUpRange;i<finalDownRange-length;i++){
			
			double[] upstream = new double[POSSIBLE_SOLUTIONS];
			double[] downstream = new double[POSSIBLE_SOLUTIONS];
			
			String base = sequence.substring(i,i+length);
			
			int countUpstreamMatches = 0;
			int countDownstreamMatches = 0;
			//find the upstream matches
			for(int j=(i-length);j>=finalUpRange;j--){ //go from closest to farther away
				String match = sequence.substring(j,j+length);
				boolean doesMatch = makesMatch(base,match);
				if(doesMatch&&countUpstreamMatches<POSSIBLE_SOLUTIONS){
					double distance = (i-j);
					upstream[countUpstreamMatches] = 1.0/distance;
					countUpstreamMatches++;
				}
			}
			//find the downstream matches
			for(int j=i;j<finalDownRange-length;j++){   //go from closest to farther away
				String match = sequence.substring(j,j+length);
				boolean doesMatch = makesMatch(base,match);
				if(doesMatch&&countDownstreamMatches<POSSIBLE_SOLUTIONS){
					double distance = (j-i);
					downstream[countDownstreamMatches] = 1.0/distance;
					countDownstreamMatches++;
				}
			}
			
			for(int j=0;j<POSSIBLE_SOLUTIONS;j++)
				result.add(upstream[j]);
			for(int j=0;j<POSSIBLE_SOLUTIONS;j++)
				result.add(downstream[j]);
		}

		return result;
	}
	
	public boolean makesMatch(String base, String match){
		String rev = toInverseString(match);
		rev = rotateString(rev);
		
		if(base.equals(rev))
			return true;
		else 
			return false;
	}
	

	
	public String toInverseString(String s){
		StringBuffer sb = new StringBuffer();
		for(int i=0;i<s.length();i++){
			char c = s.charAt(i);
			switch(c){
			case 'A' : sb.append('T'); break;
			case 'C' : sb.append('G'); break;
			case 'G' : sb.append('C'); break;
			case 'T' : sb.append('A'); break;
			case 'U' : sb.append('A'); break;
			default : sb.append('N'); break;
			}
		}
		return sb.toString();
	}
	
	public String rotateString(String s){
		StringBuffer sb = new StringBuffer();
		for(int i=s.length()-1;i>=0;i--){
			sb.append(s.charAt(i));
		}
		return sb.toString();
	}
	
	/**
	 * This method asks the classificationfeature for its conversion-type.
	 * @return The used conversion type. Null of not appropriate.
	 */
	public Conversion getConversion(){
		return null;
	}
	
	
	
	/**
	 * Returns the string representation of the classificationfeature,
	 * but with all possible spaces and tabs replaced by underscores.
	 * @return Alternative string representation
	 */
	public String toStringUnderscore(){
		return "NEARESTEQUAL_"+upstream_range+"_"+downstream_range;
	}
	
	
	/**
	 * String representation of the classificationFeature
	 * @return String representation of the classificationfeature
	 */
	public String toString(){
		return "NEARESTEQUAL "+upstream_range+" "+downstream_range; 
	}
	
	/**
	 * String representation of the classificationFeature that contains
	 * the core elements, and discards the upstream/downstream range
	 * elements.
	 * @return Shortened version of the string representation of the classificationfeature
	 */
	public String toNoRangeString(){
		return "NEARESTEQUAL";
	}
	
	/**
	 * Returns a standardized description of the location that is indicated.
	 * e.g. 5 bp upstream of splicesite. In case the location is outside the range
	 * of the feature, an error string is returned.
	 * @param loc The location
	 * @return The description
	 */
	public AttributeDescription getLocationDescription(int loc){
		return null;
	}
	
	/**
	 * Returns the number of attributes that are created by this classificationfeature,
	 * according to the internal parameters (e.g. up- and downstreamranges). When the 
	 * internal parameters are not set, or they are in an incorrect state, this method will
	 * return -1 as value
	 * @return The number of attributes created by this classificationfeature
	 */
	public int getNumberOfAttributes(){
		int temp = upstream_range+downstream_range;
		return 2*temp;
	}
	
	/**
	 * This method sets the necessary parameters for the classificationfeature. 
	 * The parameters are all generically passed as an array of strings, which of course
	 * need to be decoded by the classificationfeature itself.
	 * @param args The array with the parameters in string format
	 * @throws IncorrectClassificationFeatureException Can be thrown when the parameters
	 * are incorrect (outside possible ranges) and/or the parameters could not be parsed.
	 */
	public void setParameters(String[] args) throws IncorrectClassificationFeatureException{
		if(args.length!=4)
			throw new IncorrectClassificationFeatureException("Incorrect number of parameters for NearestEqual feature");
		String name = args[0].toUpperCase();
		if(!name.equals("NEAREST")&&!name.equals("NE"))
			throw new IncorrectClassificationFeatureException("Incorrect naming for NearestEqual feature");
		try{
			length = Integer.parseInt(args[1]);
			upstream_range = Integer.parseInt(args[2]);
			downstream_range = Integer.parseInt(args[3]);
		}
		catch(NumberFormatException exc){
			throw new IncorrectClassificationFeatureException("Integer parameters for NearestEqual feature cannot be parsed correctly");
		}
		
		if(length<1 || upstream_range<0 || downstream_range<0)
			throw new IncorrectClassificationFeatureException("Indicated ranges for NearestEqual features are not permitted");
	}
	
	/**
	 * Because all classificationfeatures need different algorithms to draw them,
	 * according to the specifications, these utilities are also moved to the 
	 * implementations of the ClassificationFeatures. However, as not to clutter them,
	 * we have chosen to place this data in a seperate interface. The programmer thus
	 * has a choice of using one or two classes to implement both interfaces.
	 * 
	 * @return The object which can draw the classification features.
	 */
	public ClassificationFeatureDrawing getClassificationFeatureDrawing(){
		return new NearestEqualClassificationFeatureDrawing(this);
	}
	
	
	/**
	 * This method asks for the up- and downstreamranges of the feature
	 * @return The up/down ranges
	 */
	public int[] getUpDownRanges(){
		int [] res = {upstream_range,downstream_range};
		return res;
	}
	
	/**
	 * This method sets the new upstream and downstream ranges  
	 * @param up The new upstream range
	 * @param down The new downstream range
	 */
	public void setUpDownRanges(int up, int down){
		this.upstream_range = up;
		this.downstream_range = down;
	}
	
	/**
	 * This method returns the length of the feature. If not applicable,
	 * (such as in case of branchpointfeatures), return 0;
	 * @return length of the feature
	 */
	public int getLength(){
		return length;
	}

	public int getDownstream_range() {
		return downstream_range;
	}

	public void setDownstream_range(int downstream_range) {
		this.downstream_range = downstream_range;
	}

	public int getUpstream_range() {
		return upstream_range;
	}

	public void setUpstream_range(int upstream_range) {
		this.upstream_range = upstream_range;
	}

	public void setLength(int length) {
		this.length = length;
	}
	
	/**
	 * Makes a copy of this particular classificationfeature. 
	 * The cloneable interface cannot be used directly because it is only applicable
	 * on classes, not interfaces.
	 * @return A copy of the classificationfeature
	 */
	public ClassificationFeature make_copy(){
		ClassificationFeature res = null;
		try{
			res = (ClassificationFeature)this.clone();
		}
		catch(Exception exc){}
		return res;
	}
}
