/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
package util.classificationFeatures.implementations.test;

import java.util.ArrayList;
import java.util.List;

import util.classificationFeatures.AttributeDescription;
import util.classificationFeatures.ClassificationFeature;
import util.classificationFeatures.ClassificationFeatureDrawing;
import util.classificationFeatures.FeatureExtractionException;
import util.classificationFeatures.IncorrectClassificationFeatureException;
import util.conversions.Conversion;

public class NonBindingClassificationFeature implements ClassificationFeature{
	
	private Conversion conversion = null;
	private int length = 0;

	
	public NonBindingClassificationFeature(){
	}
	
	/**
	 * This method asks the classificationfeature for its conversion-type.
	 * @return The used conversion type. Null of not appropriate.
	 */
	public Conversion getConversion(){
		return conversion;
	}
	
	/**
	 * Makes a copy of this particular classificationfeature. 
	 * The cloneable interface cannot be used directly because it is only applicable
	 * on classes, not interfaces.
	 * @return A copy of the classificationfeature
	 */
	public ClassificationFeature make_copy(){
		ClassificationFeature res = null;
		try{
			res = (ClassificationFeature)this.clone();
		}
		catch(Exception exc){}
		return res;
	}
	
	/**
	 * Asks whether the computer needs to precompute compositional types.
	 * When a conversion is taken with a large alphabet and/or a large length,
	 * this computation can take WAAAY to long (and take too much memory as well).
	 * Therefore, prior to this operation, this flag should be checked.
	 * @return Whether or not to supply compositional types of certain length.
	 */
	public boolean needCompositionalTypes(){
		return false;
	}
	
	/**
	 * This method extracts the features from a sequence (relative to the splicesite)
	 * and returns those features in a list of double-values (standardized on double
	 * values because of its generic but still small-sized nature).
	 * @param sequences The sequence from which the features should be extracted.
	 * @param splicesite The location of the splicesite (pseudo or real)
	 * @param types The compositional types (if requested, otherwise null).
	 * @param conversions A list with all the conversion types
	 * @param conversion_sequences  An array with all the converted sequences.
	 * @return A list with the features
	 * @throws FeatureExtractionException Thrown when something goes wrong when extracting the features.
	 */
	public List<Double> extractFeatures(String sequences[], int splicesite, String[] types, List<Conversion> conversions,String[]conversion_sequences) throws FeatureExtractionException{
		//doesnt work with conversion, just on original string...
		List<Double> result = new ArrayList<Double>();
		String sequence = sequences[0];
		String substring = sequence.substring(splicesite-length/2-length%2,splicesite+length/2);
		String reverse_substring1 = reverseString(substring);
		String reverse_substring2 = reverseString(inverseString(substring));
		
		double count1 = 0;
		double count2 = 0;
		double count3 = 0;
		double count4 = 0;
		for(int i=0;i<sequence.length()-length;i++){
			/*
			 * First literal checks without g-wobble pairs
			 */
			String t = sequence.substring(i,i+length);
			if(t.equals(reverse_substring1))
				count1++;
			if(t.equals(reverse_substring2))
				count2++;
			
			/*
			 * second check with g-wobble pair
			 */
			StringBuffer buf1 = new StringBuffer();
			StringBuffer buf2 = new StringBuffer();
			for(int j=0;j<t.length();j++){
				if(t.charAt(j)=='G'){	//WOBBLE PAIR possible!!
					char f1 = mainBindingChar(reverse_substring1.charAt(j));
					char f2 = mainBindingChar(reverse_substring2.charAt(j));
					buf1.append(f1);
					buf2.append(f2);
				}
				else{
					buf1.append(t.charAt(j));
					buf2.append(t.charAt(j));
				}
			}
			String b1 = buf1.toString();
			String b2 = buf2.toString();
			
			if(b1.equals(reverse_substring1))
				count3++;
			if(b2.equals(reverse_substring2))
				count4++;
			
		}
		result.add(count1);
		result.add(count2);
		result.add(count3);
		result.add(count4);
		double end = count1+count2+count3+count4;
		result.add(end);		
		return result;
	}
	
	private char mainBindingChar(char c){
		char r = 'N';
		switch(c){
		case 'A' : r = 'T'; break;
		case 'C' : r = 'G'; break;
		case 'G' : r = 'C'; break;
		case 'T' : case 'U' : r = 'A'; break;
		default : r = 'N'; break;
		}
		return r;
	}
	
	private String inverseString(String t){
		StringBuffer f = new StringBuffer();
		for(int i=t.length()-1;i>=0;i--){
			f.append(t.charAt(i));
		}
		return f.toString();
	}
	
	private String reverseString(String t){
		StringBuffer f = new StringBuffer();
		for(int i=0;i<t.length();i++){
			char c = t.charAt(i);
			char r = 'N';
			switch(c){
			case 'A' : r = 'T'; break;
			case 'C' : r = 'G'; break;
			case 'G' : r = 'C'; break;
			case 'T' : case 'U' : r = 'A'; break;
			default : r = 'N'; break;
			}
			f.append(r);
		}
		return f.toString();
	}
	
	/**
	 * Returns the string representation of the classificationfeature,
	 * but with all possible spaces and tabs replaced by underscores.
	 * @return Alternative string representation
	 */
	public String toStringUnderscore(){
		return "NOBINDING_"+length;
	}
	
	
	/**
	 * String representation of the classificationFeature
	 * @return String representation of the classificationfeature
	 */
	public String toString(){
		return "NOBINDING "+length;
	}
	
	
	/**
	 * String representation of the classificationFeature that contains
	 * the core elements, and discards the upstream/downstream range
	 * elements.
	 * @return Shortened version of the string representation of the classificationfeature
	 */
	public String toNoRangeString(){
		return "NOBINDING "+length;
	}
	
	/**
	 * Returns the number of attributes that are created by this classificationfeature,
	 * according to the internal parameters (e.g. up- and downstreamranges). When the 
	 * internal parameters are not set, or they are in an incorrect state, this method will
	 * return -1 as value
	 * @return The number of attributes created by this classificationfeature
	 */
	public int getNumberOfAttributes(){
		return 4; //just two numbers, indicating number of associated binding sites
	}
	
	/**
	 * This method sets the necessary parameters for the classificationfeature. 
	 * The parameters are all generically passed as an array of strings, which of course
	 * need to be decoded by the classificationfeature itself.
	 * @param args The array with the parameters in string format
	 * @throws IncorrectClassificationFeatureException Can be thrown when the parameters
	 * are incorrect (outside possible ranges) and/or the parameters could not be parsed.
	 */
	public void setParameters(String[] args) throws IncorrectClassificationFeatureException{
		if(args.length!=2)
			throw new IncorrectClassificationFeatureException("Incorrect number of parameters for nonbinding feature");
		String name = args[0].toUpperCase();
		if(!name.equals("NONBINDING")&&!name.equals("NB"))
			throw new IncorrectClassificationFeatureException("Incorrect naming for nonbinding feature");
		try{
			length = Integer.parseInt(args[1]);
		}
		catch(NumberFormatException exc){
			throw new IncorrectClassificationFeatureException("Integer parameters for nonbinding feature cannot be parsed correctly");
		}
		
		if(length<1)
			throw new IncorrectClassificationFeatureException("Indicated ranges for nonbinding features are not permitted");
	}
	
	/**
	 * Because all classificationfeatures need different algorithms to draw them,
	 * according to the specifications, these utilities are also moved to the 
	 * implementations of the ClassificationFeatures. However, as not to clutter them,
	 * we have chosen to place this data in a seperate interface. The programmer thus
	 * has a choice of using one or two classes to implement both interfaces.
	 * 
	 * @return The object which can draw the classification features.
	 */
	public ClassificationFeatureDrawing getClassificationFeatureDrawing(){
		return null;
	}
	
	
	/**
	 * This method asks for the up- and downstreamranges of the feature
	 * @return The up/down ranges
	 */
	public int[] getUpDownRanges(){
		int [] res = {0,0};
		return res;
	}
	
	
	/**
	 * This method sets the new upstream and downstream ranges  
	 * @param up The new upstream range
	 * @param down The new downstream range
	 */
	public void setUpDownRanges(int up, int down){
	}
	
	
	/**
	 * This method returns the length of the feature. If not applicable,
	 * (such as in case of branchpointfeatures), return 0;
	 * @return length of the feature
	 */
	public int getLength(){
		return length;
	}
	
	
	/**
	 * Returns a standardized description of the location that is indicated.
	 * e.g. 5 bp upstream of splicesite. In case the location is outside the range
	 * of the feature, an error string is returned.
	 * @param loc The location
	 * @return The description
	 */
	public AttributeDescription getLocationDescription(int loc){
		return null;
	}

}
