package util.classificationFeatures.implementations.test;

import java.util.ArrayList;
import java.util.List;

import util.classificationFeatures.AttributeDescription;
import util.classificationFeatures.ClassificationFeature;
import util.classificationFeatures.ClassificationFeatureDrawing;
import util.classificationFeatures.FeatureExtractionException;
import util.classificationFeatures.IncorrectClassificationFeatureException;
import util.conversions.Conversion;

public class PercentageClassificationFeature implements ClassificationFeature{

	
	private int upstream_range = 0;
	private int downstream_range = 0;
	
	
	public PercentageClassificationFeature(){
		
	}
	
	/**
	 * This method extracts the features from a sequence (relative to the splicesite)
	 * and returns those features in a list of double-values (standardized on double
	 * values because of its generic but still small-sized nature).
	 * @param sequences The sequence from which the features should be extracted.
	 * @param splicesite The location of the splicesite (pseudo or real)
	 * @param types The compositional types (if requested, otherwise null).
	 * @param conversions A list with all the conversion types
	 * @param conversion_sequences  An array with all the converted sequences.
	 * @return A list with the features
	 * @throws FeatureExtractionException Thrown when something goes wrong when extracting the features.
	 */
	public List<Double> extractFeatures(String[] sequences, int splicesite, String[] types, List<Conversion> conversions, String[] conversion_sequences) throws FeatureExtractionException {
		List<Double> result = new ArrayList<Double>();
		//some basic error checking.
		if(sequences==null || sequences[0]==null)
			throw new FeatureExtractionException("Stringsequence is null or could not be retrieved.");
		String sequence = sequences[0].toUpperCase();
		if(splicesite<0 || splicesite>sequence.length())
			throw new FeatureExtractionException("Splicesite is outside permitted range for percentage feature -> splicesite : "+splicesite+" | sequence length : "+sequence.length());
		
		List<Double> singlePercentagesUp = new ArrayList<Double>();
		List<Double> multiPercentagesUp = new ArrayList<Double>();
		List<Double> singlePercentagesDown = new ArrayList<Double>();
		List<Double> multiPercentagesDown = new ArrayList<Double>();
		
		String upSequence = null;
		String downSequence = null;
		
		//get the correct upstream string
		if(splicesite-upstream_range<0){
			upSequence = sequence.substring(0,splicesite);
			String nString = getNString(upstream_range-splicesite);
			upSequence = nString+upSequence;
		}
		else{
			upSequence = sequence.substring(splicesite-upstream_range,splicesite);
		}
		
		//get the correct downstream string
		if(splicesite+downstream_range>sequence.length()){
			downSequence = sequence.substring(splicesite,sequence.length());
			String nString = getNString(splicesite+downstream_range-sequence.length());
			downSequence = downSequence+nString;
		}
		else{
			downSequence = sequence.substring(splicesite,splicesite+downstream_range);
		}
		
		singlePercentagesUp.add(getPercentage(upSequence,'A'));
		singlePercentagesUp.add(getPercentage(upSequence,'C'));
		singlePercentagesUp.add(getPercentage(upSequence,'G'));
		singlePercentagesUp.add(getPercentage(upSequence,'T'));
		
		singlePercentagesDown.add(getPercentage(downSequence,'A'));
		singlePercentagesDown.add(getPercentage(downSequence,'C'));
		singlePercentagesDown.add(getPercentage(downSequence,'G'));
		singlePercentagesDown.add(getPercentage(downSequence,'T'));
		
		char[] ac = {'A','C'};
		char[] at = {'A','T'};
		char[] ag = {'A','G'};
		char[] ct = {'C','T'};
		char[] cg = {'C','G'};
		char[] gt = {'G','T'};
		
		multiPercentagesUp.add(getPercentage(upSequence,ac));
		multiPercentagesUp.add(getPercentage(upSequence,at));
		multiPercentagesUp.add(getPercentage(upSequence,ag));
		multiPercentagesUp.add(getPercentage(upSequence,ct));
		multiPercentagesUp.add(getPercentage(upSequence,cg));
		multiPercentagesUp.add(getPercentage(upSequence,gt));
		
		multiPercentagesDown.add(getPercentage(downSequence,ac));
		multiPercentagesDown.add(getPercentage(downSequence,at));
		multiPercentagesDown.add(getPercentage(downSequence,ag));
		multiPercentagesDown.add(getPercentage(downSequence,ct));
		multiPercentagesDown.add(getPercentage(downSequence,cg));
		multiPercentagesDown.add(getPercentage(downSequence,gt));
		
		result.addAll(singlePercentagesUp);
		result.addAll(singlePercentagesDown);
		result.addAll(multiPercentagesUp);
		result.addAll(multiPercentagesDown);
		
		return result;
	}
	
	private double getPercentage(String s,char c){
		if(s==null || s.length()==0)
			return 0.0;
		double count = 0;
		double total = s.length();
		
		for(int i=0;i<s.length();i++){
			if(s.charAt(i)==c)
				count++;
		}
		count = count/total;
		return count;
	}
	
	private double getPercentage(String s,char[]ch){
		if(s==null || s.length()==0 || ch==null||ch.length==0)
			return 0.0;
		double count = 0;
		for(int i=0;i<ch.length;i++){
			char c = ch[i];
			count+= getPercentage(s,c);
		}
		
		return count;
	}
	
	private String getNString(int t){
		StringBuffer buf = new StringBuffer();
		for(int i=0;i<t;i++)
			buf.append('N');
		return buf.toString();
	}

	/**
	 * Because all classificationfeatures need different algorithms to draw them,
	 * according to the specifications, these utilities are also moved to the 
	 * implementations of the ClassificationFeatures. However, as not to clutter them,
	 * we have chosen to place this data in a seperate interface. The programmer thus
	 * has a choice of using one or two classes to implement both interfaces.
	 * 
	 * @return The object which can draw the classification features.
	 */
	public ClassificationFeatureDrawing getClassificationFeatureDrawing() {
		return null;
	}

	/**
	 * This method asks the classificationfeature for its conversion-type.
	 * @return The used conversion type. Null of not appropriate.
	 */
	public Conversion getConversion() {
		return null;
	}

	/**
	 * This method returns the length of the feature. If not applicable,
	 * (such as in case of branchpointfeatures), return 0;
	 * @return length of the feature
	 */
	public int getLength() {
		return 0;
	}

	/**
	 * Returns a standardized description of the location that is indicated.
	 * e.g. 5 bp upstream of splicesite. In case the location is outside the range
	 * of the feature, an error string is returned.
	 * @param loc The location
	 * @return The description
	 */
	public AttributeDescription getLocationDescription(int loc) {
		return null;
	}

	/**
	 * Returns the number of attributes that are created by this classificationfeature,
	 * according to the internal parameters (e.g. up- and downstreamranges). When the 
	 * internal parameters are not set, or they are in an incorrect state, this method will
	 * return -1 as value
	 * @return The number of attributes created by this classificationfeature
	 */
	public int getNumberOfAttributes() {
		//return ;
		return 0;
	}

	/**
	 * This method asks for the up- and downstreamranges of the feature
	 * @return The up/down ranges
	 */
	public int[] getUpDownRanges() {
		int [] res = {upstream_range,downstream_range};
		return res;
	}

	/**
	 * Makes a copy of this particular classificationfeature. 
	 * The cloneable interface cannot be used directly because it is only applicable
	 * on classes, not interfaces.
	 * @return A copy of the classificationfeature
	 */
	public ClassificationFeature make_copy() {
		ClassificationFeature res = null;
		try{
			res = (ClassificationFeature)this.clone();
		}
		catch(Exception exc){}
		return res;
	}

	/**
	 * Asks whether the computer needs to precompute compositional types.
	 * When a conversion is taken with a large alphabet and/or a large length,
	 * this computation can take WAAAY to long (and take too much memory as well).
	 * Therefore, prior to this operation, this flag should be checked.
	 * @return Whether or not to supply compositional types of certain length.
	 */
	public boolean needCompositionalTypes() {
		return false;
	}

	/**
	 * This method sets the necessary parameters for the classificationfeature. 
	 * The parameters are all generically passed as an array of strings, which of course
	 * need to be decoded by the classificationfeature itself.
	 * @param args The array with the parameters in string format
	 * @throws IncorrectClassificationFeatureException Can be thrown when the parameters
	 * are incorrect (outside possible ranges) and/or the parameters could not be parsed.
	 */
	public void setParameters(String[] args) throws IncorrectClassificationFeatureException {
		if(args==null || args.length !=3){
			throw new IncorrectClassificationFeatureException("Incorrect number of arguments for percentage feature");
		}
		String name = args[0].toUpperCase();
		if(!name.equals("PERCENTAGE"))
			throw new IncorrectClassificationFeatureException("Incorrect naming for percentage feature");
		upstream_range = Integer.parseInt(args[1]);
		downstream_range = Integer.parseInt(args[2]);		
	}

	/**
	 * This method sets the new upstream and downstream ranges  
	 * @param up The new upstream range
	 * @param down The new downstream range
	 */
	public void setUpDownRanges(int up, int down) {
		this.upstream_range = up;
		this.downstream_range = down;
	}

	/**
	 * String representation of the classificationFeature that contains
	 * the core elements, and discards the upstream/downstream range
	 * elements.
	 * @return Shortened version of the string representation of the classificationfeature
	 */
	public String toNoRangeString() {
		return "PERCENTAGE";
	}

	/**
	 * Returns the string representation of the classificationfeature,
	 * but with all possible spaces and tabs replaced by underscores.
	 * @return Alternative string representation
	 */
	public String toStringUnderscore() {
		return "PERCENTAGE_"+upstream_range+"_"+downstream_range;
	}

	/**
	 * String representation of the classificationFeature
	 * @return String representation of the classificationfeature
	 */
	public String toString(){
		return "PERCENTAGE "+upstream_range+" "+downstream_range;
	}

}
