package util.classificationFeatures.implementations.test;

import java.util.ArrayList;
import java.util.List;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import util.classificationFeatures.AttributeDescription;
import util.classificationFeatures.ClassificationFeature;
import util.classificationFeatures.ClassificationFeatureDrawing;
import util.classificationFeatures.FeatureExtractionException;
import util.classificationFeatures.IncorrectClassificationFeatureException;
import util.conversions.Conversion;

public class PreComputedComboClassificationFeature implements ClassificationFeature{

	private String type = "LEFTMOST";
	private int length  = 3;
	
	private List<String[]> correctTypes = new ArrayList<String[]>(); 
	
	public PreComputedComboClassificationFeature(){
		
	}
	
	/**
	 * Asks whether the computer needs to precompute compositional types.
	 * When a conversion is taken with a large alphabet and/or a large length,
	 * this computation can take WAAAY to long (and take too much memory as well).
	 * Therefore, prior to this operation, this flag should be checked.
	 * @return Whether or not to supply compositional types of certain length.
	 */
	public boolean needCompositionalTypes(){
		return false;
	}
	
	/**
	 * This method asks the classificationfeature for its conversion-type.
	 * @return The used conversion type. Null of not appropriate.
	 */
	public Conversion getConversion(){
		return null;
	}
	
	/**
	 * Makes a copy of this particular classificationfeature. 
	 * The cloneable interface cannot be used directly because it is only applicable
	 * on classes, not interfaces.
	 * @return A copy of the classificationfeature
	 */
	public ClassificationFeature make_copy(){
		ClassificationFeature res = null;
		try{
			res = (ClassificationFeature)this.clone();
		}
		catch(Exception exc){}
		return res;
	}
	
	/**
	 * This method extracts the features from a sequence (relative to the splicesite)
	 * and returns those features in a list of double-values (standardized on double
	 * values because of its generic but still small-sized nature).
	 * @param sequences The sequence from which the features should be extracted.
	 * @param splicesite The location of the splicesite (pseudo or real)
	 * @param types The compositional types (if requested, otherwise null).
	 * @param conversions A list with all the conversion types
	 * @param conversion_sequences  An array with all the converted sequences.
	 * @return A list with the features
	 * @throws FeatureExtractionException Thrown when something goes wrong when extracting the features.
	 */
	public List<Double> extractFeatures(String sequences[], int splicesite, String[] types, List<Conversion> conversions,String[]conversion_sequences) throws FeatureExtractionException{
		if(sequences==null||sequences.length<1||sequences[0]==null)
			throw new FeatureExtractionException("Stringsequence is null or could not be retrieved.");
		String sequence = sequences[0];
		if(splicesite<0 || splicesite>sequence.length())
			throw new FeatureExtractionException("Splicesite is outside permitted range for positional feature -> splicesite : "+splicesite+" | sequence length : "+sequence.length());
		List<Double> result = new ArrayList<Double>();
		
		//get all combinations
		String in_1 = null;
		String in_2 = null;
		String ou_1 = null;
		boolean outOfBounds = false;
		try{
			if(type.equals("LEFTMOST")){
				in_1 = sequence.substring(splicesite-(2*(length+1)),splicesite-(length+1));
				in_2 = sequence.substring(splicesite-(length+1),splicesite);
				ou_1 = sequence.substring(splicesite,splicesite+(length+1));
			}
			else if(type.equals("RIGHTMOST")){
				ou_1 = sequence.substring(splicesite-(length+1),splicesite);
				in_1 = sequence.substring(splicesite,splicesite+(length+1));
				in_2 = sequence.substring(splicesite+(length+1),splicesite+2*(length+1));
			}
		}
		catch(IndexOutOfBoundsException exc){
			outOfBounds = true;
		}
		
		if(outOfBounds){
			int num = 0;
			if(length==3)
				num = 448;
			else if(length ==4)
				num = 875;
			else
				num = 0;
			
			for(int i=0;i<num;i++)
				result.add(0.0);
		}
		else{
			List<String[]> combinations = getCombinations(in_1,in_2,ou_1);
		
			Pattern pattern_1  = Pattern.compile("(.{1})\\1");
			Pattern pattern_2  = Pattern.compile("(.{2})\\1");
			Pattern pattern_3  = Pattern.compile("(.{3})\\1");
		
			for(int i=0;i<combinations.size();i++){
				String[] te = combinations.get(i);
				//find repeats.
				double f1 = getNumDoubles(te,1,pattern_1);
				double f2 = getNumDoubles(te,2,pattern_2);
				double f3 = getNumDoubles(te,3,pattern_3);
				result.add(f1);
				result.add(f2);
				result.add(f3);
				
				//find sames
				double y1 = getSames(te);
				result.add(y1);
				
				//find binding possibilities
				double[] b1 = getBindingPossibilities(te);
				result.add(b1[0]);
				result.add(b1[1]);
				result.add(b1[2]);
			}
		}
		
		return result;
	}
	
	private double[] getBindingPossibilities(String[] te){
		double[] result = new double[3];
		
		if(canBind(te[0],te[1]))
			result[0] = 1.0;
		if(canBind(te[0],te[2]))
			result[1] = 1.0;
		if(canBind(te[1],te[2]))
			result[2] = 1.0; 
		return result;
	}
	
	private String toReverseStringBinding(String a){
		StringBuffer sb = new StringBuffer();
		for(int i=a.length()-1;i>=0;i--){
			char c = a.charAt(i);
			switch(c){
			case 'A' : sb.append('T'); break;
			case 'T' : sb.append('A'); break;
			case 'C' : sb.append('G'); break;
			case 'G' : sb.append('C'); break;
			default : sb.append('N'); break;
			}
		}
		return sb.toString();
	}
	
	private boolean canBind(String a,String b){
		boolean res = false;
		if(toReverseStringBinding(a).equals(b))
			res = true;
		return res;
	}
		                             
	
	private double getSames(String[]te){
		double res = 0.0;
		if(te[0].equals(te[1]))
			res+=0.3;
		if(te[0].equals(te[2]))
			res+=0.3;
		if(te[1].equals(te[2]))
			res+=0.3;
		return res;
	}
	
	private double getNumDoubles(String[] combo, int f,Pattern p){
		double res = 0;
		String total = null;
		if(type.equals("LEFTMOST"))
			total = combo[0]+combo[1]+combo[2];
		else
			total = combo[2]+combo[0]+combo[1];
		int count = 0;
		Matcher matcher = p.matcher(total);
		while(matcher.find())
			count++;
		res = count;
		return res;
	}
	
	
	private List<String[]> getCombinations(String s1,String s2, String t1){
		List<String[]> result= new ArrayList<String[]>();
		List<String> s1_list = getCombos(s1);
		List<String> s2_list = getCombos(s2);
		List<String> t1_list = getCombos(t1);
		
		for(int i=0;i<s1_list.size();i++){
			String po1 = s1_list.get(i);
			for(int j=0;j<s2_list.size();j++){
				String po2 = s2_list.get(j);
				for(int k=0;k<t1_list.size();k++){
					String po3 = t1_list.get(k);
					String[] res = {po1,po2,po3};
					result.add(res);
				}
			}
		}
		
		return result;
	}
	
	private List<String> getCombos(String s){
		List<String> res=  new ArrayList<String>();
		if(length==3&&s.length()==4){
			res.add(s.substring(0,length));
			res.add(s.substring(1,length+1));
			
			String split1 = ""+ s.charAt(0)+ s.charAt(2)+ s.charAt(3);
			String split2 = ""+ s.charAt(0)+ s.charAt(1)+ s.charAt(3);
			res.add(split1);
			res.add(split2);
		}
		else if(length==4&&s.length()==5){
			res.add(s.substring(0,length));
			res.add(s.substring(1,length+1));
			
			String split1 = ""+s.charAt(0)+s.charAt(1)+s.charAt(2)+s.charAt(4);
			String split2 = ""+s.charAt(0)+s.charAt(2)+s.charAt(3)+s.charAt(4);
			String split3 = ""+s.charAt(0)+s.charAt(1)+s.charAt(3)+s.charAt(4);
			
			res.add(split1);
			res.add(split2);
			res.add(split3);
		}
		return res;
	}
	
	
	/**
	 * Returns the string representation of the classificationfeature,
	 * but with all possible spaces and tabs replaced by underscores.
	 * @return Alternative string representation
	 */
	public String toStringUnderscore(){
		return "PCC_"+length+"_"+type;
	}
	
	
	/**
	 * String representation of the classificationFeature
	 * @return String representation of the classificationfeature
	 */
	public String toString(){
		return "PCC "+length+" "+type;
	}
	
	
	/**
	 * String representation of the classificationFeature that contains
	 * the core elements, and discards the upstream/downstream range
	 * elements.
	 * @return Shortened version of the string representation of the classificationfeature
	 */
	public String toNoRangeString(){
		return "PCC "+length+" "+type;
	}
	
	
	/**
	 * Returns the number of attributes that are created by this classificationfeature,
	 * according to the internal parameters (e.g. up- and downstreamranges). When the 
	 * internal parameters are not set, or they are in an incorrect state, this method will
	 * return -1 as value
	 * @return The number of attributes created by this classificationfeature
	 */
	public int getNumberOfAttributes(){
		return correctTypes.size();
	}
	
	/**
	 * This method sets the necessary parameters for the classificationfeature. 
	 * The parameters are all generically passed as an array of strings, which of course
	 * need to be decoded by the classificationfeature itself.
	 * @param args The array with the parameters in string format
	 * @throws IncorrectClassificationFeatureException Can be thrown when the parameters
	 * are incorrect (outside possible ranges) and/or the parameters could not be parsed.
	 */
	public void setParameters(String[] args) throws IncorrectClassificationFeatureException{
		if(args.length!=3)
			throw new IncorrectClassificationFeatureException("Incorrect number of parameters for PCC feature");
		String name = args[0].toUpperCase();
		if(!name.equals("PCC"))
			throw new IncorrectClassificationFeatureException("Incorrect naming for PCC feature");
		try{
			length = Integer.parseInt(args[1]);
			if(length<3 || length>4)
				throw new IncorrectClassificationFeatureException("Only supported are lengths 3 and 4");
			String temp = args[2].toUpperCase();
			if(temp.equals("RIGHTMOST"))
				type = temp;
			else if(temp.equals("LEFTMOST"))
				type = temp;
			else
				throw new IncorrectClassificationFeatureException("Incorrect type for PCC feature");
			
			//loadCorrectTypes(args[3]);
			
		}
		catch(Exception exc){
			throw new IncorrectClassificationFeatureException("Parameters for pcc feature cannot be parsed correctly "+exc);
		}
		
	}
	
	/**
	 * Because all classificationfeatures need different algorithms to draw them,
	 * according to the specifications, these utilities are also moved to the 
	 * implementations of the ClassificationFeatures. However, as not to clutter them,
	 * we have chosen to place this data in a seperate interface. The programmer thus
	 * has a choice of using one or two classes to implement both interfaces.
	 * 
	 * @return The object which can draw the classification features.
	 */
	public ClassificationFeatureDrawing getClassificationFeatureDrawing(){
		return null;
	}
	
	
	/**
	 * This method asks for the up- and downstreamranges of the feature
	 * @return The up/down ranges
	 */
	public int[] getUpDownRanges(){
		int [] res = new int[2];
		if(type.equals("LEFTMOST")){
			res[0] = 2*(length+1);
			res[1] = (length+1);
		}
		else if(type.equals("RIGHTMOST")){
			res[0] = (length+1);
			res[1] = 2*(length+1);
		}
		return res;
	}
	
	
	/**
	 * This method sets the new upstream and downstream ranges  
	 * @param up The new upstream range
	 * @param down The new downstream range
	 */
	public void setUpDownRanges(int up, int down){
		//do nothing
	}
	
	
	/**
	 * This method returns the length of the feature. If not applicable,
	 * (such as in case of branchpointfeatures), return 0;
	 * @return length of the feature
	 */
	public int getLength(){
		return length;
	}
	

	
	/**
	 * Returns a standardized description of the location that is indicated.
	 * e.g. 5 bp upstream of splicesite. In case the location is outside the range
	 * of the feature, an error string is returned.
	 * @param loc The location
	 * @return The description
	 */
	public AttributeDescription getLocationDescription(int loc){
		return null;
	}

	
}
