/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
package util.featureSelection;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

import main.FunSiPJar;

import org.apache.log4j.Logger;

import util.classificationFeatures.ClassificationFeature;
import weka.attributeSelection.AttributeEvaluator;
import weka.attributeSelection.BestFirst;
import weka.attributeSelection.CfsSubsetEval;
import weka.attributeSelection.ChiSquaredAttributeEval;
import weka.attributeSelection.FCBFSearch;
import weka.attributeSelection.GainRatioAttributeEval;
import weka.attributeSelection.InfoGainAttributeEval;
import weka.attributeSelection.ReliefFAttributeEval;
import weka.attributeSelection.SubsetEvaluator;
import weka.attributeSelection.SymmetricalUncertAttributeEval;
import weka.attributeSelection.SymmetricalUncertAttributeSetEval;
import weka.core.Instances;
import base.ClassificationAction;
import classifier.Classifier;

public class FeatureSelectionStub implements Cloneable {

	public enum FEATURE_SELECTION_TYPE {NO_FEATURE_SELECTION,PRECOMPUTED_FILTER,SYMMETRICAL_UNCERTAINTY,CFS,INFORMATION_GAIN,CHI_SQUARED,
				GAIN_RATIO,RELIEF,FCBF};
	public enum FEATURE_SELECTION_PURPOSE{USE_FILTER,WRITE_MINIMAL_FILTER,WRITE_FULL_FILTER,LOAD_FILTER,USE_AND_WRITE_FULL_FILTER};
	
	private ClassificationAction ca 							= null;
	private Logger logger										= null;
	private FEATURE_SELECTION_TYPE feature_selection_type 		= FEATURE_SELECTION_TYPE.NO_FEATURE_SELECTION;
	private FEATURE_SELECTION_PURPOSE feature_selection_purpose = FEATURE_SELECTION_PURPOSE.USE_FILTER;
	private String feature_selection_input_file					= null;
	private String feature_selection_output_file				= null;
	private int maxNumFeatures									= -1;
	private List<ValPosCombination> featSelection 				= null;
	private List<Integer> filter								= null;
	
	
	public Object clone() throws CloneNotSupportedException{
		FeatureSelectionStub stub = new FeatureSelectionStub(ca,logger);
		stub.setFeature_selection_input_file(this.feature_selection_input_file);
		stub.setFeature_selection_output_file(this.feature_selection_output_file);
		stub.setFeature_selection_purpose(this.feature_selection_purpose);
		stub.setFeature_selection_type(this.feature_selection_type);
		return stub;
	}

	public FeatureSelectionStub(ClassificationAction ca, Logger logger){
		this.ca = ca;
		this.logger = logger;
	}
	
	public void setLocalClassificationAction(ClassificationAction ca){
		this.ca = ca;
	}
	
	public void applyFeatureSelectionToTraining(Classifier classifier, boolean sort){
		logger.debug("Performing feature selection "+feature_selection_type+" for "+ca.getName());
		try{
			featSelection = null;
			boolean filterCreated = false;
			boolean skipExtraction = false;
			switch(feature_selection_type){
			case NO_FEATURE_SELECTION : return;
			case PRECOMPUTED_FILTER : 
				skipExtraction = true;
				filterCreated  = true;
				break;
			case SYMMETRICAL_UNCERTAINTY :
				//System.out.println("doing symm unc");
				AttributeEvaluator su_ae = new SymmetricalUncertAttributeEval();
				featSelection = classifier.performAttributeEvaluation(sort, su_ae);
				//System.out.println("featselection done : "+featSelection.size());
				break;
			case CFS :
				Instances cfs_inst = classifier.getTrainingFileInstances();
				SubsetEvaluator cfs_se = new CfsSubsetEval();
				cfs_se.buildEvaluator(cfs_inst);
				BestFirst cfs_search = new BestFirst();
				int[] cfs_retainted = cfs_search.search(cfs_se, cfs_inst);
				Arrays.sort(cfs_retainted);
				filter  = new ArrayList<Integer>();
				for(int i=0;i<cfs_retainted.length;i++)
					filter.add(cfs_retainted[i]);
				filterCreated = true;
				break;
			case FCBF: 
				Instances fcbf_inst = classifier.getTrainingFileInstances();
				FCBFSearch fcbf_search = new FCBFSearch();
				SymmetricalUncertAttributeSetEval fcbf_evaluator = new SymmetricalUncertAttributeSetEval();
				fcbf_evaluator.buildEvaluator(fcbf_inst);
				int[] fcbf_retainted = fcbf_search.search(fcbf_evaluator, fcbf_inst);
				Arrays.sort(fcbf_retainted);
				filter = new ArrayList<Integer>();
				for(int i=0;i<fcbf_retainted.length;i++)
					filter.add(fcbf_retainted[i]);
				filterCreated =true;
				break;
			case INFORMATION_GAIN:
				AttributeEvaluator ig_ae = new InfoGainAttributeEval();
				featSelection = classifier.performAttributeEvaluation(sort, ig_ae);
				break;
			case CHI_SQUARED :
				AttributeEvaluator cs_ae = new ChiSquaredAttributeEval();
				featSelection = classifier.performAttributeEvaluation(sort,cs_ae);
				break;
			case GAIN_RATIO : 
				AttributeEvaluator gr_ae = new GainRatioAttributeEval();
				featSelection = classifier.performAttributeEvaluation(sort,gr_ae);
				break;
			case RELIEF  : 
				AttributeEvaluator r_ae = new ReliefFAttributeEval();
				featSelection = classifier.performAttributeEvaluation(sort,r_ae);
				break;
			default : break;
			}
			
			if(featSelection==null&& !filterCreated)
				return;
			
			if(!filterCreated)
				filter = createFilter(featSelection);
			filterCreated = true;
			
			if(!skipExtraction)
				logger.debug("Feature filter size : "+filter.size());
			//System.out.println("Size filter : "+filter.size());
			
			File file = new File(ca.getOutputdirectory()+ca.getFeatures_output_file());
		//	System.out.println("TEST ::: "+file.getPath());
			switch(feature_selection_purpose){
			case USE_FILTER : 
				classifier.applyAttributeFilter(filter, maxNumFeatures,file);
				logger.debug("Filter used");
				break;
			case WRITE_MINIMAL_FILTER :
				writeFeatureSelectionFile(ca.getOutputdirectory()+feature_selection_output_file,filter);
				logger.debug("Minimal filter created");
				break;
			case WRITE_FULL_FILTER :
				writeFeatureSelectionFileFull(ca.getOutputdirectory()+feature_selection_output_file,featSelection);
				logger.debug("Full filter created");
				break;
			case LOAD_FILTER:
				filter = loadFSFile(feature_selection_input_file);
				classifier.applyAttributeFilter(filter, maxNumFeatures, file);
				logger.debug("Filter loaded and applied"); break;
			case USE_AND_WRITE_FULL_FILTER:
				classifier.applyAttributeFilter(filter,maxNumFeatures,file);
				writeFeatureSelectionFileFull(ca.getOutputdirectory()+feature_selection_output_file,featSelection);
				logger.debug("Filter used and written");
				break;
			default : break;
			}
		}
		catch(Exception exc){
			logger.error(exc);
			exc.printStackTrace();
		}
		
		logger.debug("Feature selection for "+ca.getName()+" completed.");
	}

	
	public void printOutDir(ClassificationAction ac){
		System.out.println("FeatureSelection test1 (first copy) : "+this.ca.getOutputdirectory());
		System.out.println("FeatureSelection test2 (second copy) : "+ac.getOutputdirectory());
		System.out.println("Equality : "+ac.equals(ca));
	}
	
	public List<Double> applyFeatureSelectionToTest(List<Integer> filter, List<Double> features){
		if(filter==null)
			return features;
		List<Double> result = new ArrayList<Double>();
		/*
		 * Filter contains an extra first element for the classindex.
		 * This should not be included (hence int i=1). Also, every element in
		 * the filter should be shifted on to the left. (hence filt=filter.get(i)-1).
		 */
		for(int i=1;i<filter.size();i++){
			int filt = filter.get(i)-1;	
			if(filt>=features.size()){
				//should not happen
				logger.error("filter outside range : "+features.size()+" "+filt);	
			}
			else
				result.add(features.get(filt));
		}
		return result;
	}
	
	public List<Integer> getStandardClassifierFilter(){
		if(filter==null){
			System.out.println("WARNING : CREATING FILTER FROM SCRATCH");
			switch(feature_selection_purpose){
			case USE_FILTER : return createFilter(featSelection);
			case LOAD_FILTER : return loadFSFile(feature_selection_input_file);
			default : return null;
			}
		}
		else
			return filter;
	}
	
	public List<Integer> createFilter(List<ValPosCombination> list){
		if(list==null)
			return null;
		List<Integer> result = new ArrayList<Integer>();
		for(int i=0;i<list.size();i++){
			ValPosCombination vpc = list.get(i);
			int position = vpc.getPosition();
			double value = vpc.getValue();
			if(value!=0.0)
				result.add(position);
		}
		
		return result;
	}
	
	
	public void writeFeatureSelectionFile(String fileName,List<Integer> list){
		try{
			File file = new File(fileName);
			BufferedWriter writer = new BufferedWriter(new FileWriter(file));
			writer.write(ca.getClassification_name().toUpperCase()+"\n");
			List<ClassificationFeature> cfList = ca.getClassificationFeatures();
			for(int i=0;i<cfList.size();i++){
				writer.write("CLASSIFICATION_FEATURE\t"+cfList.get(i).toString()+"\n");
			}
			for(int i=0;i<list.size();i++){
				writer.write(list.get(i)+"\n");
			}
			writer.close();
		}
		catch(Exception exc){
			logger.error(exc);
		}	
	}
	
	public void writeFeatureSelectionFileFull(String fileName,List<ValPosCombination> list){
		try{
			File file = new File(fileName);
			BufferedWriter writer = new BufferedWriter(new FileWriter(file));
			//writer.write(ca.getClassification_name().toUpperCase()+"\n");
			List<ClassificationFeature> cfList = ca.getClassificationFeatures();
			for(int i=0;i<cfList.size();i++){
				writer.write("CLASSIFICATION_FEATURE\t"+cfList.get(i)+"\n");
			}
			for(int i=0;i<list.size();i++){
				ValPosCombination vpc = list.get(i);
				writer.write(vpc.getPosition()+"\t"+vpc.getValue()+"\n");
			}
			writer.close();
		}
		catch(Exception exc){
			logger.error(exc);
		}
	}
	
	
	public List<Integer> loadFSFile(String path){
		List<Integer> list = new ArrayList<Integer>();
		try{
			if(path==null){
				logger.error("Unable to load filter : featureselection file is null");
				return null;
			}
			path = FunSiPJar.setCorrectFilePath(path);
			InputStream stream  = getClass().getResourceAsStream(path);
			InputStreamReader read = null;
			if(stream!=null){
				read = new InputStreamReader(stream);
			}
			else{
				read = new FileReader(new File(path));
			}
			BufferedReader reader = new BufferedReader(read);
			String s = reader.readLine(); //read line by line.
			while(s!=null){
				if(s.startsWith("CLASSIFICATION_FEATURE")){} //header
				else{
					String[] split = s.split("\t");
					if(split.length==1){	//type 1 of fs output
						int temp = Integer.parseInt(s);
						list.add(temp);
					}
					else{					//type 2 of fs output
						double value = Double.parseDouble(split[1]);
						int temp = Integer.parseInt(split[0]);
						if(value!=0.0 && value!=0)
							list.add(temp);
					}
				}
				s = reader.readLine();
			}
			reader.close();
		}
		catch(Exception exc){
			logger.error(exc);
			exc.printStackTrace();
		}
		return list;
	}
	
	
	public void setFilter(List<Integer> newFilter){
		this.filter = newFilter;
	}
	public FEATURE_SELECTION_PURPOSE getFeature_selection_purpose() {
		return feature_selection_purpose;
	}
	public void setFeature_selection_purpose(
			FEATURE_SELECTION_PURPOSE feature_selection_purpose) {
		this.feature_selection_purpose = feature_selection_purpose;
	}
	public boolean setFeature_selection_purpose(String s){
		for(FEATURE_SELECTION_PURPOSE fs : FEATURE_SELECTION_PURPOSE.values()){
			if(s.toUpperCase().equals(fs.toString())){
				setFeature_selection_purpose(fs);
				return true;
			}
		}
		return false;
	}
	public FEATURE_SELECTION_TYPE getFeature_selection_type() {
		return feature_selection_type;
	}
	public void setFeature_selection_type(
			FEATURE_SELECTION_TYPE feature_selection_type) {
		this.feature_selection_type = feature_selection_type;
	}
	public boolean setFeature_selection_type(String s){
		for(FEATURE_SELECTION_TYPE fs : FEATURE_SELECTION_TYPE.values()){
			if(fs.toString().equals(s.toUpperCase())){
				setFeature_selection_type(fs);
				return true;
			}
		}
		return false;
	}
	public String getFeature_selection_input_file() {
		return feature_selection_input_file;
	}
	public void setFeature_selection_input_file(String feature_selection_input_file) {
		this.feature_selection_input_file = feature_selection_input_file;
	}
	public String getFeature_selection_output_file() {
		return feature_selection_output_file;
	}
	public void setFeature_selection_output_file(
			String feature_selection_output_file) {
		this.feature_selection_output_file = feature_selection_output_file;
	}
	public int getMaxNumFeatures() {
		return maxNumFeatures;
	}
	public void setMaxNumFeatures(int maxNumFeatures) {
		this.maxNumFeatures = maxNumFeatures;
	}	
	public boolean doFeatureSelection(){
		if(feature_selection_type==FEATURE_SELECTION_TYPE.NO_FEATURE_SELECTION){
			if(feature_selection_purpose==FEATURE_SELECTION_PURPOSE.LOAD_FILTER)
				return true;
			else
				return false;
		}
		else{
			return true;
		}
	}
	
}
