/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
package util.secondaryStructure;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileOutputStream;
import java.io.FileWriter;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.OutputStream;
import java.io.PrintWriter;
import java.util.ArrayList;
import java.util.List;

import org.apache.log4j.ConsoleAppender;
import org.apache.log4j.Logger;
import org.apache.log4j.SimpleLayout;

import util.io.FileParser;

/**
 * This class is primarily used for building the secondary structure 
 * of non-training files (although it can also be used for this purpose, there
 * are better ways of doing this). This class allows a link the the native
 * RNAfold method (which is assumed to be installed locally).
 * @author Michiel Van Bel, VIB, Ghent University, Michiel.vanbel\@psb.ugent.be
 */
public class SecondaryStructureExtraction {

	
	private Logger logger = null;
	
	/**
	 * Constructor of the secondary structure builder class
	 * @param logger The logging facility for the project
	 */
	public SecondaryStructureExtraction(Logger logger){
		this.logger = logger;
	}
	
	
	/**
	 * This method extracts the actual secondary structures from the sequence, given by the 
	 * list of possible splicesites. A subsequemce of certain length (according to the maxUp/maxDown)
	 * regulations is then extracted from the sequence around the splicesite. This is done
	 * because RNAfold works faster with smaller sequencelengths (results are not as
	 * good, but we have to choose between these trade-offs). 
	 * @param sequence The sequence that contains the splicesites and from which the sec structs will be extracted
	 * @param locations The different (pseudo-) splice sites within the sequence
	 * @param outputDirectory The outputdirectory for temporary files
	 * @param maxUp maximum upstream range from the splicesites
	 * @param maxDown maximum downstream range from the splicesite
	 * @return Data structure containing the secondary structures, the associated primary
	 * sequences of the same length and the associated free energies.
	 */
	public SecondaryStructureData extractStructures(String sequence,List<Integer> locations, String outputDirectory, int maxUp, int maxDown){
		SecondaryStructureData dat = null;		
		try{
			File tempInFile = new File(outputDirectory+"sec_struct_in.txt");
			File tempOutFile = new File(outputDirectory+"sec_struct_out.txt");
			BufferedWriter writer = new BufferedWriter(new FileWriter(tempInFile));
			
			List<String> sequences = new ArrayList<String>();
			//extracting all DNA-seqs
			for(int i=0;i<locations.size();i++){
				int location = locations.get(i);
				String seq = extractSequenceString(sequence,location,maxUp,maxDown);
				writer.write(seq+"\n");
				sequences.add(seq);
			}
			writer.close();
			
			FileOutputStream fos = new FileOutputStream(tempOutFile);
			
			//executing RNAfold
			logger.info("Extracting the secondary structures for current sequence");
			String[] command = {"RNAfold"};
			
			Runtime rt = Runtime.getRuntime();
			Process proc = rt.exec(command);
			
			InputStream out = proc.getInputStream();
			OutputStream in = proc.getOutputStream();
			ProcessReader procReader = new ProcessReader(logger,out,fos);
			ProcessWriter procWriter = new ProcessWriter(logger,in,sequences);
			procReader.start();
			procWriter.start();
			
			proc.waitFor();
			Thread.sleep(1000);
			fos.flush();
			fos.close();
			
			//placing the results in the output
			FileParser parser = new FileParser();
			dat = parser.extractSecondaryStructuresFromFile(tempOutFile.getPath());
			dat.setSequences(sequences);
			
			//deleting the temporary files
			tempInFile.delete();
			tempOutFile.delete();
		}
		catch(Exception exc){
			logger.error(exc);
			exc.printStackTrace();
		}
		
		return dat;
	}
	
	
	/**
	 * This method extracts a string of the same length for all possible maximum upstream and
	 * downstream lengths, relative to the splicesite location. If the boundaries fall outside
	 * the normal lengths of the string, 'N' nucleotide characters are added. 
	 * @param sequence The sequence 
	 * @param location The location of the splicesite
	 * @param maxUp The maximum upstream location
	 * @param maxDown The maximum downstream location
	 * @return The substring of the right size
	 */
	public String extractSequenceString(String sequence,int location,int maxUp,int maxDown){
		String result = null;
		
		if(maxUp<location && (location+maxDown)<sequence.length())
			result = sequence.substring(location-maxUp,location+maxDown);
		
		else if(maxUp<location && (location+maxDown)>sequence.length()){
			result = sequence.substring(location-maxUp,sequence.length());
			StringBuffer buffer = new StringBuffer();
			for(int i=0;i<(location+maxDown-sequence.length());i++)
				buffer.append('N');
			result+=buffer.toString();
		}
		
		else if(maxUp>location && (location+maxDown)<sequence.length()){
			StringBuffer buffer = new StringBuffer();
			for(int i=0;i<(maxUp-location);i++)
				buffer.append('N');
			result = buffer.toString()+sequence.substring(0,location+maxDown);
		}
		else{
			StringBuffer buffer = new StringBuffer();
			for(int i=0;i<(maxUp-location);i++)
				buffer.append('N');
			result = buffer.toString()+sequence.substring(0,sequence.length());
			buffer = new StringBuffer();
			for(int i=0;i<(location+maxDown-sequence.length());i++)
				buffer.append('N');
			result+=buffer.toString();
		}
		
		return result;
	}
	
	//for testing purposes only
	public static void main(String[] args){
		Logger logger = Logger.getLogger("TEST");
		ConsoleAppender appender = new ConsoleAppender(new SimpleLayout());
		logger.addAppender(appender);

		String sequence = "AACGAGAGAGAGCGAGGAGGCGGAGCGCGCGCGCGAGCGGATATCGATCTCTCTCTAGCATGC";
		SecondaryStructureExtraction ex = new SecondaryStructureExtraction(logger);
		List<Integer> locations = new ArrayList<Integer>();
		locations.add(1);locations.add(5);locations.add(20); locations.add(34); locations.add(60);
		
		SecondaryStructureData dat = ex.extractStructures(sequence, locations, "./output/", 30, 10);
		List<String> seqs = dat.getSequences();
		List<String> structs = dat.getStructures();
		List<Double> energies = dat.getEnergies();
		
		if(seqs.size()!=structs.size() || structs.size()!=energies.size())
			System.out.println("incompatible sizes : "+seqs.size()+"\t"+structs.size()+"\t"+energies.size());
		else{
			for(int i=0;i<seqs.size();i++){
				System.out.println(seqs.get(i)+"\n"+structs.get(i)+"\n"+energies.get(i));
				System.out.println("++++++++++++++++++++++++++++++++");
			}
		}
		
		
		
	}
	
	
	private class ProcessWriter extends Thread{
		private Logger logger = null;
		private OutputStream out = null;
		private List<String> data = null;

		public ProcessWriter(Logger logger,OutputStream out,List<String> list){
			this.logger = logger;
			this.out = out;
			this.data = list;
		}

		public void run(){
			try{
				PrintWriter writer = new PrintWriter(out);
				for(int i=0;i<data.size();i++){
					Thread.sleep(1);
					writer.println(data.get(i));
				}
				writer.println("@");
				writer.flush();
			}	
			catch(Exception exc){
				logger.error(exc);
				exc.printStackTrace();
			}
		}
	}
	
	/**
	 * This private class functions simply as a stub, because there has to 
	 * be a way for the native process to tunnel its output to. If this 
	 * is not done properly, the native process stops and blocks the entire
	 * workflow.  
	 * @author Michiel Van Bel, VIB, Ghent University, Michiel.vanbel@psb.ugent.be
	 */
	private class ProcessReader extends Thread{
		
		private Logger logger = null;
		private InputStream in = null;
		private OutputStream os = null;
		
		public ProcessReader(Logger logger, InputStream in,OutputStream toFile){
			this.logger = logger;
			this.in = in;
			this.os = toFile;
		}
				
		public void run(){
			try{
				PrintWriter pw = new PrintWriter(os);
				BufferedReader reader = new BufferedReader(new InputStreamReader(in));
				String line = null;
				while((line = reader.readLine()) != null){
					pw.println(line);
				}
				pw.flush();
				pw.close();
			}
			catch(Exception exc){
				logger.error(exc);
			}
		}
	}
}
