/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
package util.conversions.implementations;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.util.HashMap;
import java.util.Map;

import util.conversions.Conversion;
import util.conversions.ConversionUtils;


/**
 * This class is designed to take advantage of the normal RNA(DNA) to Secondary
 * structure conversion, by designing a new alphabet around them.
 * This way, more information is captured then by considering
 * the two seperately (RNA and secondary structure extracted features seperate).
 * 
 * Research towards this end is done by (among others) Donald Patterson and Barelli.
 *  
 * @author Michiel Van Bel
 */
public class DNA_SS_Extension_Conversion implements Conversion{

	private int length = 3;
	private String GT_COMMAND = "WITH_WOBBLE";
	private Conversion conv = new DNA_SS2_Conversion();
	
	private char[] dna_array = {'A','C','G','T'};
	private char[] full_array = {'A','B','C','D','E','F','G','H','I',
			'J','K','L','M','N','O','P','Q','R','S','T','U',
			'V','W','X','Y','Z'};
	
	private Map<String,Character> mapping = new HashMap<String,Character>();
	
	public static void main(String[] args){
		try{
			BufferedReader reader = new BufferedReader(new FileReader(new File(args[0])));
			String[] normalConversion = new String("DNA_SS 5 WITH_PN NO_WOBBLE WITH_SNRNP").split(" ");
			Conversion ss_conversion = ConversionUtils.parseConversionFromString(normalConversion);
			String s = reader.readLine();
			while(s!=null){
				String temp = ss_conversion.makeConversion(s);
				if(temp==null || temp.length()!=s.length()){
					System.out.println(s);
					System.out.println(temp);
				}
				s  = reader.readLine();
			}
		}
		catch(Exception exc){
			exc.printStackTrace();
		}
	}
	
	public DNA_SS_Extension_Conversion(){
		
	}
		
	
	/**
	 * Returns the name of conversion. This name should be unique
	 * among all other conversionnames, active or not (just in case).
	 * @return The unique name of the conversion
	 */
	public String getName(){
		return "DNA_SS_EXTENSION_CONVERSION";
	}
	
	/**
	 * Returns the parameter name of the conversion. This is usefull for
	 * compatibility with gui's and other programs which rely on the
	 * parameter for identification.
	 * @return String with parametername (unique).
	 */
	public String getParamName(){
		return "DNA_SS_X";
	}
	
	
	/**
	 * This method performs the actual conversion from one string
	 * to another. This can for example be the translation from RNA
	 * to pyrimidine/purine sequence, or the translation from the
	 * primary structure to the secondary structure.
	 * @param s The original sequence
	 * @return The translated string
	 */
	public String makeConversion(String s){
		if(s==null)
			return null;
		
		//String[] normalConversion = new String("DNA_SS2 "+length+" WITH_PN "+GT_COMMAND+" NO_SNRNP").split(" ");
		//Conversion ss_conversion = ConversionUtils.parseConversionFromString(normalConversion);
		String ss_string = conv.makeConversion(s);
		
		if(s.length()!=ss_string.length())
			return null;	//something b0rked
		
		
		StringBuffer buff = new StringBuffer();
		for(int i=0;i<s.length();i++){
			char s_c = s.charAt(i);
			char ss_c = ss_string.charAt(i);
			
			String t = s_c+""+ss_c;
			Character res = mapping.get(t);
			if(res!=null)
				buff.append(res);
			else{
				buff.append('#');	//character not in dataset
			}
		}
		return buff.toString();
	}
	
	
		
	/**
	 * This method sets the possible options for this conversion.
	 * These options are of course dependend on the type of conversion,
	 * so this method is as general as possible.
	 * @param args The possible options (name+value)
	 * @return  Whether setting the options succeeded (true) or not (false).
	 */
	public boolean setOptions(String[] args){
		if(args==null)
			return false;
		if(args.length!=3)
			return false;
		if(!args[0].equals("DNA_SS_X"))
			return false;
		try{
			int le = Integer.parseInt(args[1]);
			this.length = le;
			
			String t = args[2].toUpperCase().trim();
			if(t.equals("WITH_WOBBLE"))
				GT_COMMAND = t;
			else if(t.equals("NO_WOBBLE"))
				GT_COMMAND = t;
			else{
				return false;
			}
		}
		catch(Exception exc){
			return false;
		}
		
		//now, construct the conversion with the right options
		String[] normalConversion = new String("DNA_SS2 "+length+" WITH_PN "+GT_COMMAND+" NO_SNRNP").split(" ");
		conv = ConversionUtils.parseConversionFromString(normalConversion);
		
		//no errors occured, now construct the correct alphabet.
		
		char[] ss_array = conv.getDifferentCharacters();		
		
		mapping = new HashMap<String,Character>();
		for(int i=0;i<dna_array.length;i++){
			for(int j=0;j<ss_array.length;j++){
				String t = dna_array[i]+""+ss_array[j];
				mapping.put(t, full_array[i*ss_array.length+j]);
			}
		}
		return true;
	}
	

	/**
	 * This method returns the number of different characters present in
	 * the final representation after a conversion on a sequence has 
	 * occured (excluding the 'N' character).
	 * @return The number of different characters
	 */
	public int getNumberOfDifferentCharacters(){
		return mapping.keySet().size();
	}
	
	
	/**
	 * This method returns all possible different characters that occur
	 * after the conversion of a sequence. This is of course a description
	 * of the alphabet used by the conversion. The 'N' character is 
	 * excluded.
	 * @return All posible characters used by the conversion 
	 */
	public char[] getDifferentCharacters(){
		int length1 = dna_array.length;
		int length2 = conv.getNumberOfDifferentCharacters();
		int total = length1*length2;
		char[] result = new char[total];
		for(int i=0;i<total;i++)
			result[i] = full_array[i];
		return result;
	}

}
